#!/usr/bin/perl -w

die "Change this script to use a version of Perl greater than or equal to 5."
    if $PERL_VERSION < 5;

use strict;
use English;
use Getopt::Long;

use DBI;
use ba_mysql_shared;

use ba_harness_shared;

use constant false => 0;
use constant true => 1;

my $options_use_db = true;
my $options_simulate = false;
my $options_stress = false;
my %options_arg_ranges = (random_seed => "1-1;1",
			  req_sched => "list:FCFS",
			  dis_sched => "list:n/a",
			  num_users => "10-10;1",
			  virtual_clock_terminate => "259200-259200;1.0",
			  max_req_dis_tasks_per_user => "0-0;1",
			  user_arrivals => "poisson:100.0-100.0;1",
			  task_set_change_range =>
			  "uniform:0.0-0.0;0.1,0.3-0.3;0.1",
			  tasks_per_task_set => "exponential:10.0-10.0;1",
			  service_time => "exponential:20.0-20.0;1",
			  think_time => "exponential:20.0-20.0;1");
my $options_db_host = $db_default_loc;
my $options_use_traces = false;
my $options_tracedir = undef;
my $options_id_file = undef;
my $options_id_type = "output";
my $options_skip_runs = false;

my @settable_args = qw( req_sched dis_sched num_users
			virtual_clock_terminate
		        max_req_dis_tasks_per_user user_arrivals
		        task_set_change_range tasks_per_task_set service_time
		        think_time random_seed );

my %arg_letters = (req_sched => "y",
		   dis_sched => "z",
		   num_users => "n",
		   virtual_clock_terminate => "v",
		   max_req_dis_tasks_per_user => "m",
		   user_arrivals => "a",
		   task_set_change_range => "b",
		   tasks_per_task_set => "c",
		   service_time => "d",
		   think_time => "e",

		   trace_output => "t",
		   trace_input => "u",

		   any_random_seed => "R",
		   random_seed => "r");

my %arg_ranges;

my @write_ids;

my @multi_args_list;

######################################################################

&parse_options;

&setup;

if($options_stress) {
    &stress;
} elsif($options_simulate) {
    &simulate;
} else {
    die "huh?";
}

&write_id_file($options_id_file, \@write_ids);

&teardown;

exit 0;

sub setup
{
    my $seed_str;
    my $seed;
    my $str;

    $str = &show_settings_header;
    print "$str";
    print "\n";

    if($options_use_db) {
	&db_ba_connect($ba_harness_ver, $options_db_host, "dpetrou",
		       $db_dpetrou_pw);
    }

    # srand(time() ^ ($$ + ($$ << 15)));

    open RANDFILE, "</dev/urandom" or die "couldn't open random device";
    # 4 = the size of an int
    read RANDFILE, $seed_str, 4 or die "couldn't read";
    close RANDFILE or die "couldn't close file";

    $seed = unpack "L", $seed_str;
    srand($seed);
}

sub teardown
{
    if($options_use_db) {
	&db_disconnect;
    }
}

######################################################################

sub stress
{
    &multiple_ba_sim(\&single_ba_sim_stress);
}

sub single_ba_sim_stress
{
    my $args = shift;
    my $ba_output;

    $args->{any_random_seed} = true;
    $ba_output = &execute_ba_sim($args);
    delete $args->{any_random_seed};

    print $ba_output;
}

######################################################################

sub simulate
{
    my $num_runs;

    if($options_use_traces) {
	if(&test_dir($options_tracedir) == false) {
	    &make_dir($options_tracedir);
	}
    }

    $num_runs = &get_num_runs;
    print "number of runs asked to do: $num_runs\n";

    $| = 1; # so single chars come out unbuffered...
    &multiple_ba_sim(\&single_ba_sim_simulate);
    $| = 0;

    print "\n";
}

sub get_num_runs
{
    my $num_runs;

    $num_runs = 1;
    foreach my $key (keys %arg_ranges) {
	my ($arg, $ranges, $num_elems);
	
	$arg = $arg_ranges{$key};

	if(defined($arg->{list})) {
	    my $list = $arg->{list};

	    $num_elems = @$list;
	    $num_runs *= $num_elems;

	    next;
	}

	$ranges = $arg->{ranges};
	$num_elems = &get_max_num_samples($ranges);
	$num_runs *= $num_elems;
    }

    return $num_runs;
}

sub single_ba_sim_simulate
{
    my $args = shift;
    my ($trace_input_id, $output_id);

#    &debug_show_args($args);

    ($trace_input_id, $output_id) = &does_output_exist($args);

    if(defined($trace_input_id) && defined($output_id)) {
	print ".";
	goto insert_id;
    }

    if($options_skip_runs) {
	print "/";
	return;
    }

    if($options_use_traces) {
	($trace_input_id, $output_id) = &when_using_traces($args);
    } else {
	($trace_input_id, $output_id) = &when_not_using_traces($args);
    }

  insert_id:
    if($options_id_type eq "output") {
	push @write_ids, $output_id;
    } elsif($options_id_type eq "trace") {
	push @write_ids, $trace_input_id;
    } else {
	die;
    }
}

sub when_not_using_traces
{
    my $args = shift;
    my ($trace_input_id, $output_id);
    my ($ba_output, $parsed_output);
    my $parsed_trace_db;
    my $need_to_insert_behavior;

    $need_to_insert_behavior = false;

    $ba_output = &execute_ba_sim($args);
    $parsed_output = &parse_output($ba_output);

    # doing the lock here because we call find_trace_input_id with a
    # valid third arg.  (see more detailed comment when i do this in
    # `when_using_traces.')
    &db_lock_tables;
    $trace_input_id = &find_trace_input_id($args, false, \$parsed_trace_db);
    &db_unlock_tables;

    if(defined($trace_input_id)) {
	&assert_total_trace($parsed_output, "current run with args `" .
			    $parsed_output->{args} . "'",
			    $parsed_trace_db, "db id " . $trace_input_id);
    } else {
	$need_to_insert_behavior = true;
    }

    if(defined($trace_input_id) && $need_to_insert_behavior == false) {
	print "o";
    } elsif(!defined($trace_input_id) && $need_to_insert_behavior) {
	print "O";
    } else {
	die;
    }

    if($need_to_insert_behavior) {
	&db_lock_tables;
	$trace_input_id = &find_trace_input_id($args, false, undef);
	if(defined($trace_input_id)) {
	    # no need to assert the db trace, since the
	    # find_trace_input_id would only have succeeded if it was good
	    # (that func checks just about everything...)
	    &db_update_trace($trace_input_id, $parsed_output);
	} else {
	    $trace_input_id = &db_insert_trace($parsed_output);
	}
	&db_unlock_tables;
    }

    &db_lock_tables;
    $output_id = &find_output_id($args, $trace_input_id);
    if(defined($output_id)) {
	&assert_total_output($parsed_output, "current run with args `" .
			     $parsed_output->{args} . "'",
			     &db_select_output_id($output_id),
			     "db id " . $output_id);
	&db_update_output($output_id, $parsed_output);
    } else {
	$output_id = &db_insert_output($parsed_output, $trace_input_id);
    }
    &db_unlock_tables;

    return ($trace_input_id, $output_id);
}

sub when_using_traces
{
    my $args = shift;
    my $need_to_create_local_trace;
    my ($trace_input_id, $output_id);
    my $trace_output;
    my $parsed_trace_db;

    $need_to_create_local_trace = false;

    # doing the lock here because we call find_trace_input_id with a
    # valid third arg.  and in doing so, that func will then need to
    # do two db ops atomically.  i don't put the locks within
    # find_trace_input_id, because sometimes that func is called when
    # locks are already held.  (breaks if i do.  getting new locks
    # releases old ones.  haha.)
    &db_lock_tables;
    $trace_input_id = &find_trace_input_id($args, true, \$parsed_trace_db);
    &db_unlock_tables;

    if(defined($trace_input_id)) {
	if(&confirm_local_trace_output($trace_input_id, $parsed_trace_db)
	   == false) {
	    $need_to_create_local_trace = true;
	}
	$trace_output = $parsed_trace_db->{trace_output};
    } else {
	$need_to_create_local_trace = true;
    }

    if(defined($trace_input_id) && $need_to_create_local_trace == false) {
	print "o";
    } elsif(defined($trace_input_id) && $need_to_create_local_trace) {
	print "X";
    } elsif(!defined($trace_input_id) && $need_to_create_local_trace) {
	print "O";
    } else {
	die;
    }

    if($need_to_create_local_trace) {
	($trace_input_id, $trace_output) = &create_local_trace($args);
    }

    die "shouldn't have happened" if(!defined($trace_input_id));

    $output_id = &do_simulate_run_when_using_traces($args, $trace_output,
						    $trace_input_id);

    return ($trace_input_id, $output_id);
}

sub do_simulate_run_when_using_traces
{
    my ($args, $trace_output, $trace_input_id) = @_;
    my $ba_output;
    my $parsed_output;
    my $output_id;

    $args->{trace_input} = $trace_output;
    $ba_output = &execute_ba_sim($args);
    delete $args->{trace_input};
    $parsed_output = &parse_output($ba_output);

    &db_lock_tables;
    $output_id = &find_output_id($args, $trace_input_id);
    if(defined($output_id)) {
	&assert_total_output($parsed_output, "current run with args `" .
			     $parsed_output->{args} . "'",
			     &db_select_output_id($output_id),
			     "db id " . $output_id);
	&db_update_output($output_id, $parsed_output);
    } else {
	$output_id = &db_insert_output($parsed_output, $trace_input_id);
    }
    &db_unlock_tables;

    return $output_id;
}

sub does_output_exist
{
    my $args = shift;
    my $trace_input_id;
    my $output_id;

    $trace_input_id = &find_trace_input_id($args, false, undef);
    if(!defined($trace_input_id)) {
	return undef;
    }

    $output_id = &find_output_id($args, $trace_input_id);
    if(!defined($output_id)) {
	return undef;
    }

    return ($trace_input_id, $output_id);
}

sub confirm_local_trace_output
{
    my $trace_input_id = shift;
    my $parsed_trace_db = shift;
    my $trace_file;
    my $trace_info;
    my $parsed_trace_file;
    my $trace_output;

    $trace_output = $parsed_trace_db->{trace_output};
    $trace_info = &get_trace_info($trace_output);
    if(!defined($trace_info)) {
	# trace file didn't exist!
	return false;
    }

    $parsed_trace_file = &parse_trace($trace_info);
    &assert_total_trace($parsed_trace_db, "db id " . $trace_input_id,
			$parsed_trace_file, $trace_output);

    return true;
}

sub create_local_trace
{
    my $args = shift;
    my $trace_input_id;
    my $randval;
    my $trace_output;
    my $trace_info;
    my $parsed_trace;

    $randval = rand;
    $trace_output = $options_tracedir . "/" . $randval;

    $args->{trace_output} = $trace_output;
    &execute_ba_sim($args);
    delete $args->{trace_output};

    $trace_info = &get_trace_info($trace_output);
    die "couldn't find the trace we just wrote out" if(!defined($trace_info));
    $parsed_trace = &parse_trace($trace_info);

    &db_lock_tables;
    $trace_input_id = &find_trace_input_id($args, false, undef);
    if(defined($trace_input_id)) {
	# no need to assert the db trace, since the
	# find_trace_input_id would only have succeeded if it was good
	# (that func checks just about everything...)
	&db_update_trace($trace_input_id, $parsed_trace);
    } else {
	$trace_input_id = &db_insert_trace($parsed_trace);
    }
    &db_unlock_tables;

    return ($trace_input_id, $trace_output);
}

sub find_trace_input_id
{
    my $args = shift;
    my $locally = shift;
    my $parsed_trace_db_out = shift;
    my $trace_input_id;

    $trace_input_id = &db_find_trace_id($locally,
					$args->{num_users},
					$args->{virtual_clock_terminate},
				$args->{max_req_dis_tasks_per_user},
					$args->{user_arrivals},
					$args->{task_set_change_range},
					$args->{tasks_per_task_set},
					$args->{service_time},
					$args->{think_time});

    if(defined($trace_input_id)) {
	if(defined($parsed_trace_db_out)) {
	    $$parsed_trace_db_out = &db_select_trace_id($trace_input_id);
	}
    }

    return $trace_input_id;
}

sub find_output_id
{
    my ($args, $trace_input_id) = @_;
    my $output_id;

    $output_id = &db_find_output_id($trace_input_id, $args->{random_seed},
				    $args->{req_sched}, $args->{dis_sched});

    return $output_id;
}

######################################################################

sub assert_total_output
{
    my $a = shift;
    my $a_key = shift;
    my $b = shift;
    my $b_key = shift;

    &assert_total_shared($a, $b, $a_key, $b_key);

    &assert_helper([$a->{random_seed}], [$b->{random_seed}],
		   "random_seed", $a_key, $b_key);

    # already verified that treat_req_dis_diff was the same, can't hurt...
    &assert_helper([$a->{treat_req_dis_diff}],
		   [$b->{treat_req_dis_diff}],
		   "treat_req_dis_diff", $a_key, $b_key);
    # already verified that req_sched was the same, can't hurt, though...
    &assert_helper([$a->{req_sched}],
		   [$b->{req_sched}],
		   "req_sched", $a_key, $b_key);
    # already verified that dis_sched was the same, can't hurt, though...
    &assert_helper([$a->{dis_sched}],
		   [$b->{dis_sched}],
		   "dis_sched", $a_key, $b_key);
    
    &assert_helper([$a->{virtual_clock}], [$b->{virtual_clock}],
		   "virtual_clock", $a_key, $b_key);
    &assert_helper([$a->{all_cand_queue}], [$b->{all_cand_queue}],
		   "all_cand_queue", $a_key, $b_key);
    &assert_helper([$a->{all_cand_queue_weighed}],
		   [$b->{all_cand_queue_weighed}],
		   "all_cand_queue_weighed", $a_key, $b_key);
    &assert_helper([$a->{req_cand_queue}], [$b->{req_cand_queue}],
		   "req_cand_queue", $a_key, $b_key);
    &assert_helper([$a->{req_cand_queue_weighed}],
		   [$b->{req_cand_queue_weighed}],
		   "req_cand_queue_weighed", $a_key, $b_key);
    &assert_helper([$a->{dis_cand_queue}], [$b->{dis_cand_queue}],
		   "dis_cand_queue", $a_key, $b_key);
    &assert_helper([$a->{dis_cand_queue_weighed}],
		   [$b->{dis_cand_queue_weighed}],
		   "dis_cand_queue_weighed", $a_key, $b_key);
    
    &assert_helper($a->{all_res_usage},
		   $b->{all_res_usage},
		   "all_res_usage", $a_key, $b_key);
    &assert_helper($a->{req_res_usage},
		   $b->{req_res_usage},
		   "req_res_usage", $a_key, $b_key);
    &assert_helper($a->{dis_res_usage},
		   $b->{dis_res_usage},
		   "dis_res_usage", $a_key, $b_key);
    &assert_helper($a->{scaled_res_usage},
		   $b->{scaled_res_usage},
		   "scaled_res_usage", $a_key, $b_key);
    &assert_helper($a->{billable_resources}, $b->{billable_resources},
		   "billable_resources", $a_key, $b_key);
    &assert_helper([$a->{res_busy_range_time}],
		   [$b->{res_busy_range_time}],
		   "res_busy_range_time", $a_key, $b_key);
    &assert_helper([$a->{res_busy_time}],
		   [$b->{res_busy_time}],
		   "res_busy_time", $a_key, $b_key);
    &assert_helper([$a->{res_idle_time}],
		   [$b->{res_idle_time}],
		   "res_idle_time", $a_key, $b_key);
    &assert_helper([$a->{non_work_cons_time}],
		   [$b->{non_work_cons_time}],
		   "non_work_cons_time", $a_key, $b_key);
    &assert_helper([$a->{load}],
		   [$b->{load}],
		   "load", $a_key, $b_key);
    &assert_helper([$a->{req_load}],
		   [$b->{req_load}],
		   "req_load", $a_key, $b_key);
    &assert_helper([$a->{req_load_ignoring_idle}],
		   [$b->{req_load_ignoring_idle}],
		   "req_load_ignoring_idle", $a_key, $b_key);
    
    &assert_helper([$a->{user_count}],
		   [$b->{user_count}],
		   "user_count", $a_key, $b_key);
    &assert_helper([$a->{num_hit_req_dis_limit}],
		   [$b->{num_hit_req_dis_limit}],
		   "num_hit_req_dis_limit", $a_key, $b_key);
    &assert_helper($a->{task_sets},
		   $b->{task_sets},
		   "task_sets", $a_key, $b_key);
    &assert_helper([$a->{task_sets_alive_time}],
		   [$b->{task_sets_alive_time}],
		   "task_sets_alive_time", $a_key, $b_key);
    &assert_helper([$a->{no_task_sets_alive_time}],
		   [$b->{no_task_sets_alive_time}],
		   "no_task_sets_alive_time", $a_key, $b_key);
    &assert_helper([$a->{tasks_ready_time}],
		   [$b->{tasks_ready_time}],
		   "tasks_ready_time", $a_key, $b_key);
    &assert_helper($a->{disclosed_tasks},
		   $b->{disclosed_tasks},
		   "disclosed_tasks", $a_key, $b_key);
    &assert_helper($a->{disclosed_only_tasks},
		   $b->{disclosed_only_tasks},
		   "disclosed_only_tasks", $a_key, $b_key);
    &assert_helper($a->{executed_tasks},
		   $b->{executed_tasks},
		   "executed_tasks", $a_key, $b_key);
    &assert_helper($a->{requested_tasks},
		   $b->{requested_tasks},
		   "requested_tasks", $a_key, $b_key);
    &assert_helper($a->{finished_tasks},
		   $b->{finished_tasks},
		   "finished_tasks", $a_key, $b_key);
    &assert_helper($a->{canceled_tasks},
		   $b->{canceled_tasks},
		   "canceled_tasks", $a_key, $b_key);
    
    &assert_helper($a->{vrt},
		   $b->{vrt},
		   "vrt", $a_key, $b_key);
    &assert_helper($a->{vsd},
		   $b->{vsd},
		   "vsd", $a_key, $b_key);
    &assert_helper($a->{user_vrt},
		   $b->{user_vrt},
		   "user_vrt", $a_key, $b_key);
    &assert_helper($a->{user_vsd},
		   $b->{user_vsd},
		   "user_vsd", $a_key, $b_key);
    &assert_helper($a->{deadlines_met},
		   $b->{deadlines_met},
		   "deadlines_met", $a_key, $b_key);
}

sub assert_same_trace_outputs
{
    my $parsed_outputs = shift;
    my $master_output;
    my $master_key;
    my $first;

    $first = true;
    foreach my $key (keys %$parsed_outputs) {
	my $iter_output;

	if($first) {
	    $first = false;
	    $master_output = $parsed_outputs->{$key};
	    $master_key = $key;
	    next;
	}

	$iter_output = $parsed_outputs->{$key};

	&assert_helper($master_output->{req_res_usage},
		       $iter_output->{req_res_usage},
		       "requested resource usage",
		       $master_key, $key);
	&assert_helper($master_output->{task_sets},
		       $iter_output->{task_sets},
		       "task sets", $master_key, $key);
	&assert_helper($master_output->{disclosed_tasks},
		       $iter_output->{disclosed_tasks},
		       "disclosed tasks", $master_key, $key);
	&assert_helper($master_output->{disclosed_only_tasks},
		       $iter_output->{disclosed_only_tasks},
		       "disclosed only tasks", $master_key, $key);
	&assert_helper($master_output->{requested_tasks},
		       $iter_output->{requested_tasks},
		       "requested tasks", $master_key, $key);
	&assert_helper($master_output->{finished_tasks},
		       $iter_output->{finished_tasks},
		       "finished tasks", $master_key, $key);
	&assert_helper($master_output->{canceled_tasks},
		       $iter_output->{canceled_tasks},
		       "canceled tasks", $master_key, $key);
    }
}

sub assert_helper
{
    my $val1_ref = shift;
    my $val2_ref = shift;
    my $name = shift;
    my $key1 = shift;
    my $key2 = shift;
    my @val1;
    my @val2;

    @val1 = @$val1_ref;
    @val2 = @$val2_ref;

    die "huh?" if(@val1 == 0 || @val2 == 0);

    if(@val1 != @val2) {
	print "failure: val1 had @val1 elems and val2 had @val2 elems";
	&assert_die_helper($name, $key1, $key2);
    }

    for(my $i = 0; $i < @val1; $i++) {
	my $good;

	$good = true;
	if($val1[$i] =~ /[a-zA-Z]/) {
	    if($val1[$i] ne $val2[$i]) {
		$good = false;
	    }
	} else {
	    if($val1[$i] != $val2[$i]) {
		$good = false;
	    }
	}

	if($good == false) {
	    print "val1: ";
	    &print_output_stats($val1_ref);
	    print "\n";
	    print "val2: ";
	    &print_output_stats($val2_ref);
	    print "\n";
	    &assert_die_helper($name, $key1, $key2);
	}
    }
}

sub assert_die_helper
{
    my $name = shift;
    my $key1 = shift;
    my $key2 = shift;

    die "assert outputs failure on `$name': val1 ($key1) != " .
	"val2 ($key2)";
}

sub assert_total_trace
{
    my $a = shift;
    my $a_key = shift;
    my $b = shift;
    my $b_key = shift;

    &assert_total_shared($a, $b, $a_key, $b_key);

    &assert_helper([$a->{num_users}],
		   [$b->{num_users}],
		   "num_users", $a_key, $b_key);
    &assert_helper([$a->{virtual_clock_terminate}],
		   [$b->{virtual_clock_terminate}],
		   "virtual_clock_terminate", $a_key, $b_key);
    &assert_helper([$a->{max_req_dis_tasks_per_user}],
		   [$b->{max_req_dis_tasks_per_user}],
		   "max_req_dis_tasks_per_user", $a_key, $b_key);
    &assert_helper([$a->{user_arrivals}],
		   [$b->{user_arrivals}],
		   "user_arrivals", $a_key, $b_key);
    &assert_helper([$a->{task_set_change_range}],
		   [$b->{task_set_change_range}],
		   "task_set_change_range", $a_key, $b_key);
    &assert_helper([$a->{tasks_per_task_set}],
		   [$b->{tasks_per_task_set}],
		   "tasks_per_task_set", $a_key, $b_key);
    &assert_helper([$a->{service_time}],
		   [$b->{service_time}],
		   "service_time", $a_key, $b_key);
    &assert_helper([$a->{think_time}],
		   [$b->{think_time}],
		   "think_time", $a_key, $b_key);
}

sub assert_total_shared
{
    my ($a, $b, $a_key, $b_key) = @_;

    # some of the following may already have been verified to be the
    # same (specifically, when asserting outputs).  it doesn't hurt to
    # compare them again, though.  they are: program, random_seed, and
    # trace_input.

    &assert_helper([$a->{program}], [$b->{program}],
		   "program", $a_key, $b_key);
    # don't care about args being the same
    # don't care about run_date being the same

    # don't care about random_seed being the same for traces, so we
    # only check that in assert_total_output.  this is assuming no
    # local tracing!! [?] XXX

    # don't care about trace_input or trace_output being the same
}

######################################################################

sub parse_options
{
    my ( %opt_hash, @letters_for_getopt, $num_actions );

    $Getopt::Long::autoabbrev = false;
    $Getopt::Long::order = $REQUIRE_ORDER;
    $Getopt::Long::ignorecase = false;

    $num_actions = 0;

    @letters_for_getopt = &set_letters_for_getopt;

    if(GetOptions(\%opt_hash,
		  "use-traces",
		  "tracedir=s",
		  "stress",
		  "show-settings",
		  "db-host=s",
		  "id-file=s",
		  "id-type=s",
		  "skip-runs",
		  @letters_for_getopt) == false) {
	die "error in GetOptions()";
    }

    if(defined($opt_hash{"db-host"})) {
	$options_db_host = $opt_hash{"db-host"};
    }
    if(defined($opt_hash{"stress"})) {
	$num_actions++;
	$options_stress = true;
	$options_use_db = false;
    }
    if(defined($opt_hash{"use-traces"})) {
	die "haven't tested tracing in a while";
	$options_use_traces = true;
    }
    if(defined($opt_hash{"tracedir"})) {
	die "haven't tested tracing in a while";
	$options_tracedir = $opt_hash{"tracedir"};
    }
    foreach my $arg (@settable_args) {
	if(defined($opt_hash{$arg_letters{$arg}})) {
	    $options_arg_ranges{$arg} = $opt_hash{$arg_letters{$arg}};
	}
    }
    if(defined($opt_hash{"id-file"})) {
	$options_id_file = $opt_hash{"id-file"};
    }
    if(defined($opt_hash{"id-type"})) {
	$options_id_type = $opt_hash{"id-type"};
    }
    if(defined($opt_hash{"skip-runs"})) {
	$options_skip_runs = true;
    }

    if($num_actions == 0) {
	$options_simulate = true;
    }
    if($options_simulate) {
	if(!defined($options_tracedir)) {
	    if(defined($ENV{BAHARNESSTRACES})) {
		$options_tracedir = $ENV{BAHARNESSTRACES};
	    }
	}
    }
    &parse_arg_ranges;
    &show_settings if(defined($opt_hash{"show-settings"}));

    if(@ARGV >= 1) {
	foreach my $foo (@ARGV) {
	    print "$foo\n";
	}
	die "left over args";
    }
    die "too many actions" if($num_actions > 1);

    if($options_simulate) {
	die "need to use the db when simulating"
	    if($options_use_db == false);
	if($options_use_traces) {
	    die "couldn't figure where to place traces"
		if(!defined($options_tracedir));
	}
    }
    die "unknown id write type" if($options_id_type ne "output" &&
				   $options_id_type ne "trace");
}

sub set_letters_for_getopt
{
    my @letters_for_getopt;

    foreach my $arg (@settable_args) {
	push @letters_for_getopt, $arg_letters{$arg} . "=s";
    }

    return @letters_for_getopt;
}

sub parse_arg_ranges
{
    foreach my $key (keys %options_arg_ranges) {
	&parse_arg_range($key);
    }
}

sub parse_arg_range
{
    my $key = shift;
    my $dist_name;
    my $ranges_str;
    my @ranges_list;
    my @ranges;
    my %arg;
    my $max_num_samples;

    # not checking as thoroughly as ba_sim...
    if($options_arg_ranges{$key} =~ /^(.+):(.*)$/) {
	if($1 eq "list") {
	    &parse_list($key, $2);
	    return;
	}
    }

    if($options_arg_ranges{$key} =~ /^(.+):(.+)$/) {
	$dist_name = $1;
	die "unknown distribution `$dist_name'"
	    if(!defined($dists{$dist_name}));
	$ranges_str = $2;
    } else {
	$ranges_str = $options_arg_ranges{$key};
    }
    
    if(defined($dist_name)) {
	$arg{dist} = $dists{$dist_name};
    }

    @ranges_list = split ',', $ranges_str;
    @ranges = ( );
    foreach my $range_str (@ranges_list) {
	push @ranges, &parse_range($range_str);
    }
    die if @ranges == 0;

    $max_num_samples = &get_max_num_samples(\@ranges);
    foreach my $range (@ranges) {
	my $samples = $range->{samples};
	my $num_samples = @$samples;
	if($num_samples != 1) {
	    die "bad num samples" if($num_samples != $max_num_samples);
	}
    }

    $arg{ranges} = \@ranges;
    $arg_ranges{$key} = \%arg;
}

sub get_max_num_samples
{
    my $ranges = shift;
    my $max_num_samples;
    my $first;

#    print "entering get_max_num_samples\n";

    $first = true;
    foreach my $range (@$ranges) {
	my $samples = $range->{samples};
	my $num_samples = @$samples;

#	print "a range------\n";
#	foreach my $foo (@$samples) {
#	    print "a sample: $foo\n";
#	}

	if($first) {
	    $first = false;
	    $max_num_samples = $num_samples;
	    next;
	}
	$max_num_samples = $num_samples if($num_samples > $max_num_samples);
    }

#    print "returning $max_num_samples\n";

    return $max_num_samples;
}

sub parse_range
{
    my $range_str = shift;
    my ($lower, $upper, $inc, $mult);
    my %range;

    if($range_str =~ /^([^-;]+)-([^-;]+);([^-;]+)$/) {
	$lower = $1;
	$upper = $2;
	$inc = $3;
    } else {
	die "can't break down range into lower, upper, and inc";
    }

    if($inc =~ /\*$/) {
	chop $inc;
	$mult = true;
    } else {
	$mult = false;
    }

    die "upper range can't be below lower" if $upper < $lower;
    die "increment can't be 0" if $inc == 0;
    if($mult) {
	die "increment can't be less than or equal to 1 when mult is on"
	    if $inc <= 1;
    }

    $range{lower} = $lower;
    $range{upper} = $upper;
    $range{inc} = $inc;
    $range{mult} = $mult;
    $range{samples} = &compute_samples($lower, $upper, $inc, $mult);

    return \%range;
}

sub parse_list
{
    my $key = shift;
    my $list_str = shift;
    my @list;
    my %arg;

    @list = split ',', $list_str;

    $arg{list} = \@list;

    $arg_ranges{$key} = \%arg;
}

sub compute_samples
{
    my ($lower, $upper, $inc, $mult) = @_;
    my @samples;

    @samples = ( );

# need that 0.000001 because of rounding stuffage XXX
    for(my $i = $lower; $i <= $upper + 0.000001;
	$mult ? ($i *= $inc) : ($i += $inc)) {
	push @samples, $i;

#	print "compute_samples: $lower-$upper;$inc == $i\n";

    }

    return \@samples;
}

sub show_settings
{
    my $str;

    $str = "";
    $str .= &show_settings_header;
    $str .= "\n";
    $str .= &show_settings_args;

    print $str;
}

sub show_settings_header
{
    my $str;

    $str = "";

    if($options_stress) {
	$str .= "stressing...\n";
    }
    if($options_simulate) {
	$str .= "simulating...\n";
    }
    $str .= "db-host: $options_db_host\n";
    if($options_use_traces) {
	$str .= "use-traces: true\n";
	$str .= "tracedir: $options_tracedir\n";
    } else {
	$str .= "use-traces: false\n";
	$str .= "tracedir: (n/a)\n";
    }
    if(defined($options_id_file)) {
	$str .= "id-file: $options_id_file\n";
	$str .= "id-type: $options_id_type\n";
    } else {
	$str .= "id-file: (none)\n";
	$str .= "id-type: (n/a)\n";
    }
    if($options_skip_runs) {
	$str .= "skip-runs: true\n";
    } else {
	$str .= "skip-runs: false\n";
    }

    return $str;
}

sub show_settings_args
{
    my $str;

    $str = "";
    foreach my $key (keys %options_arg_ranges) {
	$str .= "$key: " . &ranges_make_str($arg_ranges{$key}) . "\n";
    }

    return $str;
}

sub ranges_make_str
{
    my $arg = shift;
    my $str;
    my $first;
    my $ranges;

    die if(!defined($arg));

    $str = "";

    $first = true;
    if(defined($arg->{list})) {
	my $list = $arg->{list};

	foreach my $elem (@$list) {
	    $first ? $first = false : ($str .= ",");
	    $str .= $elem;
	}
	return $str;
    }

    if(defined($arg->{dist})) {
	$str .= $arg->{dist} . ":";
    }

    $ranges = $arg->{ranges};
    $first = true;
    foreach my $range (@$ranges) {
	$first ? $first = false : ($str .= ",");
	$str .= $range->{lower} . "-" . $range->{upper} . ";" . $range->{inc};
	$str .= "*" if($range->{mult});
    }

    return $str;
}

######################################################################

sub multiple_ba_sim
{
    my $func = shift;
    my %recurse_ranges;

    %recurse_ranges = %arg_ranges;

    &reset_ranges(\%recurse_ranges);

    &multiple_ba_sim_helper(\%recurse_ranges, [keys %recurse_ranges], $func);

    &fisher_yates_shuffle(\@multi_args_list);
    foreach my $args (@multi_args_list) {
	&$func($args);	
    }
}

sub fisher_yates_shuffle
{
    my $array = shift;
    my $i;

    for($i = @$array; --$i; ) {
	my $j = int rand ($i + 1);
	next if $i == $j;
	@$array[$i, $j] = @$array[$j, $i];
    }
}

sub reset_ranges
{
    my $ranges = shift;

    foreach my $key (keys %$ranges) {
	my $arg;

	$arg = $ranges->{$key};
	&reset_currents($arg);
    }
}

sub reset_currents
{
    my $arg = shift;
    my $ranges;

    $arg->{list_index} = 0;

    $ranges = $arg->{ranges};
    foreach my $range (@$ranges) {
	$range->{current} = $range->{lower};
    }
}

sub multiple_ba_sim_helper
{
    my $recurse_ranges_in = shift;
    my $recurse_keys_in = shift;
    my $func = shift; # not used in here anymore
    my $recurse_ranges;
    my @recurse_keys;
    my $recurse_key;
    my $end;

    $recurse_ranges = &copy_ranges($recurse_ranges_in);
    @recurse_keys = @$recurse_keys_in;

    if(@recurse_keys == 0) {
	my %args;

	foreach my $key (keys %options_arg_ranges) {
	    $args{$key} = &arg_current_str($recurse_ranges->{$key});
	}

#	&$func(\%args); [? so where is it used now?]

	push @multi_args_list, \%args;

	return;
    }

    $recurse_key = shift @recurse_keys;
    $end = false;
    while($end == false) {
	&multiple_ba_sim_helper($recurse_ranges, \@recurse_keys, $func);
	$end = &arg_increment($recurse_ranges->{$recurse_key});
    }
}

sub copy_ranges
{
    my $ranges_in = shift;
    my %ranges;

    foreach my $key (keys %$ranges_in) {
	$ranges{$key} = &copy_arg($ranges_in->{$key});
    }

    return \%ranges;
}

sub copy_arg
{
    my $arg_in = shift;
    my $ranges_in;
    my %arg;
    my @ranges;

    if(defined($arg_in->{list})) {
	$arg{list_index} = $arg_in->{list_index};
	$arg{list} = $arg_in->{list};
	return \%arg;
    }

    if(defined($arg_in->{dist})) {
	$arg{dist} = $arg_in->{dist};
    }

    $ranges_in = $arg_in->{ranges};
    @ranges = ( );
    foreach my $range_in (@$ranges_in) {
	push @ranges, &copy_range($range_in);
    }
    $arg{ranges} = \@ranges;

    return \%arg;
}

sub copy_range
{
    my $range_in = shift;
    my %range;

    $range{current} = $range_in->{current};

    $range{lower} = $range_in->{lower};
    $range{upper} = $range_in->{upper};
    $range{inc} = $range_in->{inc};
    $range{mult} = $range_in->{mult};
    $range{samples} = $range_in->{samples};

    return \%range;
}

sub arg_current_str
{
    my $arg = shift;
    my $ranges;
    my $str;
    my $first;

    if(defined($arg->{list})) {
	$str = &list_current_str($arg);
	return $str;
    }

    $str = "";
    if(defined($arg->{dist})) {
	$str .= $arg->{dist} . ":";
    }

    $ranges = $arg->{ranges};
    $first = true;
    foreach my $range (@$ranges) {
	$first ? $first = false : ($str .= ",");
	$str .= &range_current_str($range, defined($arg->{dist}));
    }

    return $str;
}

sub range_current_str
{
    my $range = shift;
    my $is_dist = shift;
    my $str;

    if($is_dist) {
	# not necessary any more to sprintf this for 6 dec places...
	$str = sprintf("%.6f", $range->{current});
    } else {
	$str = $range->{current};
    }
	
    return $str;
}

sub list_current_str
{
    my $arg = shift;
    my $list;
    my $elem;
    my $index;

    $list = $arg->{list};
    $index = $arg->{list_index};

    $elem = @$list[$index];

    return $elem;
}

sub arg_increment
{
    my $arg = shift;
    my $ranges;
    my $list;
    my $max_num_samples;

    if(defined($arg->{list})) {
	$list = $arg->{list};
	if($arg->{list_index} == @$list - 1) {
	    return true;
	} else {
	    $arg->{list_index}++;
	    return false;
	}
    }

    $ranges = $arg->{ranges};
    $max_num_samples = &get_max_num_samples($ranges);
    foreach my $range (@$ranges) {
	if(&range_increment($range, $max_num_samples)) {
	    return true;
	}
    }

    return false;
}

sub range_increment
{
    my ($range, $max_num_samples) = @_;
    my $samples = $range->{samples};
    my $num_samples = @$samples;
    my $tmp;

    if($max_num_samples > $num_samples) {
	die if($num_samples != 1);
	return false;
    }

    if($range->{mult}) {
	$tmp = $range->{current} * $range->{inc};
    } else {
	$tmp = $range->{current} + $range->{inc};
    }

# need that 0.000001 because of rounding stuffage XXX
    if($tmp > $range->{upper} + 0.000001 ) {
	return true;
    }
    $range->{current} = $tmp;

    return false;
}

sub debug_show_args
{
    my $args = shift;

    print "debug args:\n";
    foreach my $key (keys %$args) {
	my $val = $args->{$key};
	print "    $key: $val\n";
    }
    print "\n";
}

sub execute_ba_sim
{
    my $args = shift;
    my $output;
    my $args_str;
    my $ba_sim;

    $ba_sim = "./ba_sim";

    $args_str = &make_args_str($args);

#    print "executing with: $args_str\n";
    $output = &execute_generic($ba_sim, $args_str);

    return $output;
}

sub make_args_str
{
    my $args = shift;
    my $args_str;
    my $first = true;

    foreach my $key (keys %$args) {
	if($args->{$key} eq "n/a") {
	    next;
	}
	$first == true ? $first = false : $args_str .= " ";

	die if(!defined($arg_letters{$key}));
	
	$args_str .= "-" . $arg_letters{$key};

	if($key ne "any_random_seed") {
	    $args_str .= $args->{$key};
	}
    }

    if($args->{dis_sched} ne "n/a") {
	$first == true ? $first = false : $args_str .= " ";
	$args_str .= "-x";
    }

    return $args_str;
}

######################################################################

sub execute_generic
{
    my $prog = shift;
    my $args_str = shift;
    my $output;

    $output = `$prog $args_str`;
    if($? != 0) {
	if(defined $output) {
	    print $output;
	}
    }
    &parse_execute_error($?, $prog);

    return $output;
}

sub parse_execute_error
{
    my $rc = shift;
    my $prog = shift;

    if($rc == 0xff00) {
	# $! always seems to be empty to me...  XXX
	die "failure trying to execute `$prog': $!";
    } elsif($rc > 0x80) {
        # is this shift correct?  XXX i've seen:
        # `./ba_sim' gave a non-zero exit status: 0 at ./ba_harness.pl
        # line 616.
	$rc >>= 8;
	die "`$prog' gave a non-zero exit status: $rc";
    } elsif($rc != 0) {
	if($rc & 0x80) {
	    $rc &= ~0x80;
	}
	die "`$prog' received signal: $rc";
    }
}

######################################################################
