#include <ba_config.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include "getopt.h"
#include <ba_sim.h>

typedef struct {
    char *exec_name;
    char *command_line_options;

    ba_bool quiet;
    char *data_output_filename;

    unsigned int random_seed;
    ba_bool any_random_seed;

    char *trace_output_filename;
    char *trace_input_filename;

    ba_bool treat_req_dis_diff;
    ba_sched_behavior_t req_sched_behavior;
    ba_sched_behavior_t dis_sched_behavior;

    uint32_t num_users;
    ba_real virtual_clock_terminate;
    uint32_t max_req_dis_tasks_per_user;
    
    char *user_arrivals_str;
    char *task_set_change_range_str;
    char *tasks_per_task_set_str;
    char *service_delta_time_str;
    char *think_delta_time_str;

    ba_rv_dist_t *user_arrivals_dist;
    ba_rv_dist_t *task_set_change_range_dist;
    ba_rv_dist_t *tasks_per_task_set_dist;
    ba_rv_dist_t *service_delta_time_dist;
    ba_rv_dist_t *think_delta_time_dist;
} options_t;

typedef struct {
    ba_bool wants_banner;
    ba_bool wants_usage;
    ba_bool show_options;

    ba_bool got_error;    

    ba_bool random_seed_error;
    ba_bool any_random_seed_error;

    ba_bool trace_error;

    ba_bool treat_req_dis_diff_error;
    ba_bool req_sched_behavior_error;
    ba_bool dis_sched_behavior_error;

    ba_bool num_users_error;
    ba_bool virtual_clock_terminate_error;
    ba_bool max_req_dis_tasks_per_user_error;

    ba_bool user_arrivals_dist_error;
    ba_bool task_set_change_range_dist_error;
    ba_bool tasks_per_task_set_dist_error;
    ba_bool service_delta_time_dist_error;
    ba_bool think_delta_time_dist_error;
} option_actions_t;

static char *make_command_line_options(char **argv);
static void parse_dist_options(void);
static ba_rv_dist_t *parse_dist_option(char *optarg);
static void print_usage(char *progname, ba_bool is_error);
static void print_banner(void);
static unsigned int get_random_seed(void);

static char usage[] =
"Usage: %s [options]\n"
"\n"
"Options:\n"
"  -V, --version             print the version number, then exit\n"
"  -h, --help                print this message, then exit\n"
"  -s, --show-options        show the set options\n"
"\n"
"  -q, --quiet               don't print informative messages (default: off)\n"
"  -o, --data-output <file>  where to send simulation output (default: "
"stdout)\n"
"  -r, --random-seed <num>   set the random number generator seed "
"(default: 1)\n"
"  -R, --any-random-seed     pick the random seed from `random' events\n"
"                            (incompatible with the random-seed option)\n"
"\n"
"  -t, --trace-output <file> where to send a trace (default: no trace "
"output)\n"
"  -u, --trace-input <file>  where to get a trace (default: no trace input)\n"
"\n"
"  -x, --treat-req-dis-diff  treat requested and disclosed tasks differently\n"
"                            (default: off)\n"
"  -y, --req-sched-behavior <sched>  what requested task scheduler to use\n"
"                            (default: FCFS)\n"
"  -z, --dis-sched-behavior <sched>  what disclosed task scheduler to use\n"
"                            (default: FCFS when `-x' is in effect)\n"
"\n"
"  -n, --num-users <n>  number of users to spawn (default: 10)\n"
"  -v, --virtual-clock-terminate <t>  time to stop simulating (default: \n"
"                            345600; 0 means `off')\n"
"  -m, --max-req-dis-tasks-per-user <n>  maximum number of requested\n"
"                            and disclosed only tasks per user\n"
"                            (default: 0; 0 means no limit)\n"
"\n"
"  -a, --user-arrivals <count>          the counting process for the arrival "
"of\n"
"                                       users (default: poisson:100)\n"
"  -b, --task-set-change-range <dist>   the range for when a task set "
"changes\n"
"                                       (default: uniform:0.0,0.3)\n"
"  -c, --tasks-per-task-set <dist>      the distribution for the number of "
"tasks\n"
"                                       per ts (default: "
"exponential:10)\n"
"  -d, --service-time <dist>            the distribution for the task "
"service\n"
"                                       time (default: exponential:20)\n"
"  -e, --think-time <dist>              the distribution for the think\n"
"                                       time (default: exponential:20)\n"
"\n"
"Options controlling user and task behavior are ignored when using a\n"
"trace for input.  Simultaneously sending and getting a trace is not "
"allowed.\n"
"\n"
"Supported schedulers are: FCFS, UFB, URFB, SRPT, RFCFS, HRP, HRR, RSRPT,\n"
"                          RAND, RANDU, and NOOP.\n"
"\n"
"Depending on whether simulating speculative or non-speculative scheduling,\n"
"certain requested and disclosed queue policies are not allowed.\n"
"\n"
"When requested and disclosed tasks are treated the same, only the\n"
"requested scheduler is employed.\n"
"\n"
"When the virtual-clock-terminate option is in use, the simulation\n"
"terminates when its time is met, all users are spawned at the start,\n"
"and the users never depart.  When this option is not in use\n"
"(0.0), the simulation terminates when the given number of users\n"
"depart, users are spawned according to the user arrival\n"
"distribution, and the max-req-dis-tasks-per-users determine when\n"
"they depart.  Thusly, when virtual-clock-terminate is in use,\n"
"max-req-dis-tasks-per-user and user-arrivals are ignored.\n"
"Further, virtual-clock-terminate is incompatible with trace-input and\n"
"trace-output.\n"
"\n"
"Distributions are of the form:\n"
"    <distribution type>:<param1>[,<param2>[...]]\n"
"\n"
"  in which the distribution type can be one of the following:\n"
"    exponential: mean, s.t. mean >= 0;\n"
"    pareto     : k, p, alpha, s.t. 0 < k <= p, alpha >= 1 (bounded pareto),\n"
"                 or, if k and p are -1, these limits are ignored and samples\n"
"                 are at least 1 (regular pareto);\n"
"    uniform    : alpha and beta, s.t. alpha >= 0, and beta >= alpha;\n"
"    bernoulli  : prob, s.t. 0 <= prob <= 1.\n"
"\n"
"Counting processes are of the form:\n"
"    <counting process type>:<param1>[,<param2>[...]]\n"
"\n"
"  in which the counting process type must be the following:\n"
"    poisson    : mean, s.t. mean >= 0.\n"
"\n"
"The author may be reached by e-mail at <dpetrou@cs.cmu.edu>.\n"
"\n"
"The accompanying README supercedes this information.\n";

static char banner[] =
PACKAGE_NAME " " PACKAGE_VERSION " by David Petrou\n"
"Copyright (c) 2004 Carnegie Mellon University,\n"
"All rights reserved.\n"
"\n"
"You may redistribute copies under the terms in the file\n"
"COPYING contained in this program's package.\n";

static options_t *options;
static option_actions_t *option_actions;

extern char *
ba_options_exec_name_get(void)
{
    return options->exec_name;
}

extern char *
ba_options_command_line_options_get(void)
{
    return options->command_line_options;
}

extern ba_bool
ba_options_quiet_get(void)
{
    return options->quiet;
}

extern char *
ba_options_data_output_filename_get(void)
{
    return options->data_output_filename;
}

extern unsigned int
ba_options_random_seed_get(void)
{
    return options->random_seed;
}

extern char *
ba_options_trace_output_filename_get(void)
{
    return options->trace_output_filename;
}

extern char *
ba_options_trace_input_filename_get(void)
{
    return options->trace_input_filename;
}

extern ba_bool
ba_options_treat_req_dis_diff_get(void)
{
    return options->treat_req_dis_diff;
}

extern ba_sched_behavior_t
ba_options_req_sched_behavior_get(void)
{
    return options->req_sched_behavior;
}

extern ba_sched_behavior_t
ba_options_dis_sched_behavior_get(void)
{
    return options->dis_sched_behavior;
}

extern uint32_t
ba_options_num_users_get(void)
{
    return options->num_users;
}

extern ba_real
ba_options_virtual_clock_terminate_get(void)
{
    return options->virtual_clock_terminate;
}

extern uint32_t
ba_options_max_req_dis_tasks_per_user_get(void)
{
    return options->max_req_dis_tasks_per_user;
}

extern ba_rv_dist_t *
ba_options_user_arrivals_dist_get(void)
{
    return options->user_arrivals_dist;
}

extern ba_rv_dist_t *
ba_options_task_set_change_range_dist_get(void)
{
    return options->task_set_change_range_dist;
}

extern ba_rv_dist_t *
ba_options_tasks_per_task_set_dist_get(void)
{
    return options->tasks_per_task_set_dist;
}

extern ba_rv_dist_t *
ba_options_service_delta_time_dist_get(void)
{
    return options->service_delta_time_dist;
}

extern ba_rv_dist_t *
ba_options_think_delta_time_dist_get(void)
{
    return options->think_delta_time_dist;
}

extern ba_status
ba_options_get_args(int argc, char **argv)
{
    BA_ASSERT(argv);
    int val;
    ba_bool dis_sched_set, random_seed_set, any_random_seed_set;
    struct option long_options[] = {
	{ "version", no_argument, 0, 'V' },
	{ "help", no_argument, 0, 'h' },
	{ "show-options", no_argument, 0, 's' },
	{ "quiet", no_argument, 0, 'q' },
	{ "data-output", required_argument, 0, 'o' },
	{ "random-seed", required_argument, 0, 'r' },
	{ "any-random-seed", required_argument, 0, 'R' },
	{ "trace-output", required_argument, 0, 't' },
	{ "trace-input", required_argument, 0, 'u' },
	{ "treat-req-dis-diff", no_argument, 0, 'x' },
	{ "req-sched-behavior", required_argument, 0, 'y' },
	{ "dis-sched-behavior", required_argument, 0, 'z' },
	{ "num-users", required_argument, 0, 'n' },
	{ "virtual-clock-terminate", required_argument, 0, 'v' },
	{ "max-req-dis-tasks-per-user", required_argument, 0, 'm' },
	{ "user-arrivals", required_argument, 0, 'a' },
	{ "task-set-change-range", required_argument, 0, 'b' },
	{ "tasks-per-task-set", required_argument, 0, 'c' },
	{ "service-time", required_argument, 0, 'd' },
	{ "think-time", required_argument, 0, 'e' },
	{ 0, 0, 0, 0 }
    };
    char *endptr;

    options->exec_name = ba_string_new(argv[0]);
    /* what's my problem?

    if(strlen(PACKAGE_NAME) == strlen(options->exec_name)) {
	BA_ASSERT(strcmp(PACKAGE_NAME, options->exec_name) == 0);
    } else {
	BA_ASSERT(strlen(options->exec_name) > strlen("/" PACKAGE_NAME));
	BA_ASSERT(strcmp("/" PACKAGE_NAME,
			 options->exec_name +
			 strlen(options->exec_name) -
				strlen("/" PACKAGE_NAME)) == 0);
				}*/
    options->command_line_options = make_command_line_options(argv);

    dis_sched_set = BA_FALSE;
    random_seed_set = BA_FALSE;
    any_random_seed_set = BA_FALSE;
    while(1) {
	val = getopt_long(argc, argv, "Vhsqo:r:Rt:u:xy:z:n:v:N:m:M:a:b:c:d:e:",
			  long_options, 0);
	if(val == -1) {
	    break;
	}

	if(val == ':' || val == '?') {
	    option_actions->got_error = BA_TRUE;
	    return BA_STATUS_SUCCESS;
	}

	switch(val) {
	case 'V':
	    option_actions->wants_banner = BA_TRUE;
	    return BA_STATUS_SUCCESS;
	    break;
	case 'h':
	    option_actions->wants_usage = BA_TRUE;
	    return BA_STATUS_SUCCESS;
	    break;
	case 's':
	    option_actions->show_options = BA_TRUE;
	    break;

	case 'q':
	    options->quiet = BA_TRUE;
	    break;
	case 'o':
	    options->data_output_filename = ba_string_new(optarg);
	    break;
	case 'r':
	    random_seed_set = BA_TRUE;
	    // XXX doesn't check for negative
	    options->random_seed = (unsigned int)strtoul(optarg, &endptr, 10);
	    if(*endptr != '\000' || options->random_seed == LONG_MAX) {
		option_actions->random_seed_error = BA_TRUE;
	    }
	    break;
	case 'R':
	    any_random_seed_set = BA_TRUE;
	    options->any_random_seed = BA_TRUE;
	    options->random_seed = get_random_seed();
	    break;

	case 't':
	    options->trace_output_filename = ba_string_new(optarg);
	    break;
	case 'u':
	    options->trace_input_filename = ba_string_new(optarg);
	    break;

	case 'x':
	    options->treat_req_dis_diff = TRUE;
	    break;
	case 'y':
	    options->req_sched_behavior =
		ba_sched_str_to_sched_behavior(optarg);
	    break;
	case 'z':
	    options->dis_sched_behavior =
		ba_sched_str_to_sched_behavior(optarg);
	    dis_sched_set = BA_TRUE;
	    break;

	case 'n':
	    // XXX doesn't check for negative
	    options->num_users = (uint32_t)strtol(optarg, &endptr, 10);
	    if(*endptr != '\000' || options->num_users == LONG_MAX) {
		option_actions->num_users_error = BA_TRUE;
	    }
	    BA_ASSERT(options->num_users >= 1);
	    break;
	case 'v':
	    if(ba_string_scan_n_reals(optarg, 1,
				      &options->virtual_clock_terminate)
	       == BA_FALSE) {
		option_actions->virtual_clock_terminate_error = BA_TRUE;
	    }
	    // should get 0.0 from virtual_clock_initial_abs_time XXX,
	    // however, i'm initing options before events, so that var
	    // in events isn't ready yet.  gerr.
	    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
	        options->virtual_clock_terminate, BA_EVENTS_NOW);
	    break;
	case 'm':
	    // XXX doesn't check for negative
	    options->max_req_dis_tasks_per_user =
		(uint32_t)strtol(optarg, &endptr, 10);
	    if(*endptr != '\000' ||
	       options->max_req_dis_tasks_per_user == LONG_MAX) {
		option_actions->max_req_dis_tasks_per_user_error =
		    BA_TRUE;
	    }
	    BA_ASSERT(options->max_req_dis_tasks_per_user >= 0);
	    break;

	case 'a':
	    options->user_arrivals_str = ba_string_new(optarg);
	    break;
	case 'b':
	    options->task_set_change_range_str = ba_string_new(optarg);
	    break;
	case 'c':
	    options->tasks_per_task_set_str = ba_string_new(optarg);
	    break;
	case 'd':
	    options->service_delta_time_str = ba_string_new(optarg);
	    break;
	case 'e':
	    options->think_delta_time_str = ba_string_new(optarg);
	    break;
	    
	default:
	    BA_PANIC();
	    break;
	}
    }

    if(optind != argc) {
	option_actions->got_error = BA_TRUE;
	return BA_STATUS_SUCCESS;
    }

    if(dis_sched_set && options->treat_req_dis_diff == BA_FALSE) {
	option_actions->dis_sched_behavior_error = BA_TRUE;
    }
    if(options->req_sched_behavior == BA_SCHED_BEHAVIOR_RFCFS ||
       options->req_sched_behavior == BA_SCHED_BEHAVIOR_GSVSD ||
       options->req_sched_behavior == BA_SCHED_BEHAVIOR_RSRPT ||
       options->req_sched_behavior == BA_SCHED_BEHAVIOR_HRP ||
       options->req_sched_behavior == BA_SCHED_BEHAVIOR_HRR ||
       options->req_sched_behavior == BA_SCHED_BEHAVIOR_NOOP ||
       options->req_sched_behavior == BA_SCHED_BEHAVIOR_UNKNOWN ||
       options->dis_sched_behavior == BA_SCHED_BEHAVIOR_UNKNOWN) {
	BA_PANIC(); // XXX should be doing something like:
	// option_actions->req_sched_behavior_error = BA_TRUE;
    }
    if(options->treat_req_dis_diff) {
	if(options->dis_sched_behavior == BA_SCHED_BEHAVIOR_NOOP) {
	    // interactive
	    if(options->req_sched_behavior == BA_SCHED_BEHAVIOR_URFB) {
		BA_PANIC();
	    }
	} else {
	    // ba
	    if(options->req_sched_behavior == BA_SCHED_BEHAVIOR_UFB) {
		BA_PANIC();
	    }
	    if(options->dis_sched_behavior == BA_SCHED_BEHAVIOR_UFB ||
	       options->dis_sched_behavior == BA_SCHED_BEHAVIOR_URFB ||
	       options->dis_sched_behavior == BA_SCHED_BEHAVIOR_NOOP) {
		BA_PANIC();
		// XXX should be doing something like:
		// option_actions->dis_sched_behavior_error = BA_TRUE;
	    }
	}
    } else {
	// batch
	if(options->req_sched_behavior == BA_SCHED_BEHAVIOR_URFB) {
	    BA_PANIC();
	}
    }

    if(random_seed_set && any_random_seed_set) {
	option_actions->any_random_seed_error = BA_TRUE;
    }

    if(options->trace_input_filename && options->trace_output_filename) {
	option_actions->trace_error = BA_TRUE;
    }

    /* can't panic as i want, because ba_harness.pl specifies these
       args, even when not used.

    if(options->user_arrivals_str) {
	ba_print_info("haven't used the user arrivals dist in a "
		      "while, so it's unsupported for now.\n");
	BA_PANIC();	
    }

    if(options->max_req_dis_tasks_per_user) {
	ba_print_info("haven't used the max_req_dis_tasks_per_user in a "
		      "while, so it's unsupported for now.\n");
	BA_PANIC();	
	}*/

    if(options->trace_output_filename || options->trace_input_filename) {
	// XXX
	ba_print_info("haven't tried tracing in a "
		      "while, so it's unsupported for now.\n");
	BA_PANIC();	
    }

    if(ba_types_real_equal(options->virtual_clock_terminate, 0.0)) {
	// XXX
	ba_print_info("haven't tried `no virtual clock terminate' in a "
		      "while, so it's unsupported for now.\n");
	BA_PANIC();
    } else {
	if(options->max_req_dis_tasks_per_user) {
	    option_actions->max_req_dis_tasks_per_user_error = BA_TRUE;
	}

	// XXX doh, calling ba_report func before ba_report_init.  doesn't
	// hurt in this case, but not good practice.
	if(ba_types_real_greater_than_or_equal(ba_report_warmup_abs_time_get(),
				       options->virtual_clock_terminate)) {
	    ba_print_info("virtual clock terminate time less than warmup "
			  "time!\n");
	    BA_PANIC();
	}
    }

    if((options->trace_input_filename || options->trace_output_filename) &&
       options->virtual_clock_terminate) {
	option_actions->virtual_clock_terminate_error = BA_TRUE;
    }

    ba_rv_setup_random_generator();
    parse_dist_options();

    return BA_STATUS_SUCCESS;
}

static char *
make_command_line_options(char **argv)
{
    char *command_line_options;
    uint32_t i;

    command_line_options = ba_string_new("");
    for(i = 1; argv[i]; i++) { // skip over command name
	if(i != 1) {
	    command_line_options = ba_string_append(command_line_options, " ");
	}
	command_line_options = ba_string_append(command_line_options,
						argv[i]);
    }

    return command_line_options;
}

static void
parse_dist_options(void)
{
    if(options->user_arrivals_str) {
	options->user_arrivals_dist =
	    parse_dist_option(options->user_arrivals_str);
	if(options->user_arrivals_dist) {
	    option_actions->user_arrivals_dist_error = BA_FALSE;
	    BA_ASSERT(strcmp("poisson",
			     ba_rv_dist_name_get(options->
						 user_arrivals_dist))
		      == 0);
	} else {
	    option_actions->user_arrivals_dist_error = BA_TRUE;
	}
    } else {
	options->user_arrivals_dist =
	    ba_rv_dist_new("poisson", 0.0, 0.0, 100.0, 0.0);
	BA_ASSERT(options->user_arrivals_dist);
    }

    if(options->task_set_change_range_str) {
	options->task_set_change_range_dist =
	    parse_dist_option(options->task_set_change_range_str);
	if(options->task_set_change_range_dist) {
	    option_actions->task_set_change_range_dist_error = BA_FALSE;
	    BA_ASSERT(strcmp("uniform",
			     ba_rv_dist_name_get(options->
						 task_set_change_range_dist))
		      == 0);
	} else {
	    option_actions->task_set_change_range_dist_error = BA_TRUE;
	}

	// haha, would like to verify the alpha beta are between 0
	// and 1, but i can't!  i'll assert on this in
	// ba_user.c.  need to revamp ba_stats.c.  XXX
    } else {
	options->task_set_change_range_dist =
	    ba_rv_dist_new("uniform", 0.0, 0.3, 0.0, 0.0);
	BA_ASSERT(options->task_set_change_range_dist);
    }

    if(options->tasks_per_task_set_str) {
	options->tasks_per_task_set_dist =
	    parse_dist_option(options->tasks_per_task_set_str);
	if(options->tasks_per_task_set_dist) {
	    option_actions->tasks_per_task_set_dist_error = BA_FALSE;
	    BA_ASSERT(strcmp("exponential", ba_rv_dist_name_get(options->
					tasks_per_task_set_dist)) == 0 ||
		      strcmp("uniform", ba_rv_dist_name_get(options->
 				        tasks_per_task_set_dist)) == 0 ||
		      strcmp("pareto", ba_rv_dist_name_get(options->
 				        tasks_per_task_set_dist)) == 0);
	} else {
	    option_actions->tasks_per_task_set_dist_error = BA_TRUE;
	}
    } else {
	options->tasks_per_task_set_dist =
	    ba_rv_dist_new("exponential", 0.0, 0.0, 10.0, 0.0);
	BA_ASSERT(options->tasks_per_task_set_dist);
    }

    if(options->service_delta_time_str) {
	options->service_delta_time_dist =
	    parse_dist_option(options->service_delta_time_str);
	if(options->service_delta_time_dist) {
	    option_actions->service_delta_time_dist_error = BA_FALSE;
	    BA_ASSERT(strcmp("exponential", ba_rv_dist_name_get(options->
				service_delta_time_dist)) == 0 ||
		      strcmp("uniform", ba_rv_dist_name_get(options->
				service_delta_time_dist)) == 0 ||
		      strcmp("pareto", ba_rv_dist_name_get(options->
				service_delta_time_dist)) == 0);
	} else {
	    option_actions->service_delta_time_dist_error = BA_TRUE;
	}
    } else {
	options->service_delta_time_dist =
	    ba_rv_dist_new("exponential", 0.0, 0.0, 20.0, 0.0);
	BA_ASSERT(options->service_delta_time_dist);
    }

    if(options->think_delta_time_str) {
	options->think_delta_time_dist =
	    parse_dist_option(options->think_delta_time_str);
	if(options->think_delta_time_dist) {
	    option_actions->think_delta_time_dist_error = BA_FALSE;
	    BA_ASSERT(strcmp("exponential", ba_rv_dist_name_get(options->
				think_delta_time_dist)) == 0 ||
		      strcmp("uniform", ba_rv_dist_name_get(options->
				think_delta_time_dist)) == 0 ||
		      strcmp("pareto", ba_rv_dist_name_get(options->
				think_delta_time_dist)) == 0);
	} else {
	    option_actions->think_delta_time_dist_error = BA_TRUE;
	}
    } else {
	options->think_delta_time_dist =
	    ba_rv_dist_new("exponential", 0.0, 0.0, 20.0, 0.0);
	BA_ASSERT(options->think_delta_time_dist);
    }
}

static ba_rv_dist_t *
parse_dist_option(char *optarg)
{
    ba_bool ok;
    char *name;
    char *endptr;
    ba_bool alpha_used, beta_used, mean_used, prob_used;
    ba_real alpha, beta, mean, prob;
    ba_real vals[4];
    ba_rv_dist_t *dist;

    alpha = beta = mean = prob = 0.0; // warning elimination.

    name = optarg;
    ok = ba_string_scan_up_to(optarg, &endptr, ':');
    if(ok == BA_FALSE) {
	return BA_FALSE;
    }
    optarg = endptr + 1;

    ok = ba_rv_dist_values_used_get(name, &alpha_used, &beta_used, &mean_used,
				    &prob_used);
    if(ok == BA_FALSE) {
	return BA_FALSE;
    }

    if(alpha_used && beta_used && !mean_used) {
	BA_ASSERT(mean_used == BA_FALSE && prob_used == BA_FALSE);

	ok = ba_string_scan_n_reals(optarg, 2, vals);
	if(ok == BA_FALSE) {
	    return BA_FALSE;
	}

	alpha = vals[0];
	beta = vals[1];
	mean = 0.0;
	prob = 0.0;

    } else if(mean_used && !alpha_used && !beta_used) {
	BA_ASSERT(alpha_used == BA_FALSE && beta_used == BA_FALSE &&
		  prob_used == BA_FALSE);

	ok = ba_string_scan_n_reals(optarg, 1, vals);
	if(ok == BA_FALSE) {
	    return BA_FALSE;
	}
	
	mean = vals[0];
	alpha = 0.0;
	beta = 0.0;
	prob = 0.0;

    } else if(prob_used) {
	BA_ASSERT(alpha_used == BA_FALSE && beta_used == BA_FALSE &&
		  mean_used == BA_FALSE);

	ok = ba_string_scan_n_reals(optarg, 1, vals);
	if(ok == BA_FALSE) {
	    return BA_FALSE;
	}
	
	mean = 0.0;
	alpha = 0.0;
	beta = 0.0;
	prob = vals[0];

    } else if(alpha_used && beta_used && mean_used){
	BA_ASSERT(prob_used == BA_FALSE);

	ok = ba_string_scan_n_reals(optarg, 3, vals);
	if(ok == BA_FALSE) {
	    return BA_FALSE;
	}

	alpha = vals[0];
	beta = vals[1];
	mean = vals[2];
	prob = 0.0;

    } else {
	BA_PANIC();
    }

    dist = ba_rv_dist_new(name, alpha, beta, mean, prob);
    return dist;
}

extern ba_status
ba_options_parse_args(void)
{
    if(option_actions->got_error) {
	ba_print_error("Couldn't parse command-line option.\n");
	print_usage(options->exec_name, BA_TRUE);
	return BA_STATUS_OPTIONS_BAD;
    }
    if(option_actions->wants_banner) {
	print_banner();
	return BA_STATUS_OPTIONS_EXIT;
    }
    if(option_actions->wants_usage) {
	print_usage(options->exec_name, BA_FALSE);
	return BA_STATUS_OPTIONS_EXIT;
    }

    if(option_actions->random_seed_error) {
	ba_print_error("Couldn't parse the random seed option.\n");
	print_usage(options->exec_name, BA_TRUE);
	return BA_STATUS_OPTIONS_BAD;	
    }
    if(option_actions->any_random_seed_error) {
	ba_print_error("Couldn't parse the any random seed option.\n");
	print_usage(options->exec_name, BA_TRUE);
	return BA_STATUS_OPTIONS_BAD;	
    }

    if(option_actions->trace_error) {
	ba_print_error("Couldn't parse the trace option(s).\n");
	print_usage(options->exec_name, BA_TRUE);
	return BA_STATUS_OPTIONS_BAD;	
    }

    if(option_actions->treat_req_dis_diff_error) {
	ba_print_error("Couldn't parse the treat requested and disclosed "
		       "differently option.\n");
	print_usage(options->exec_name, BA_TRUE);
	return BA_STATUS_OPTIONS_BAD;	
    }
    if(option_actions->req_sched_behavior_error) {
	ba_print_error("Couldn't parse the requested scheduler behavior "
		       "option.\n");
	print_usage(options->exec_name, BA_TRUE);
	return BA_STATUS_OPTIONS_BAD;	
    }
    if(option_actions->dis_sched_behavior_error) {
	ba_print_error("Couldn't parse the disclosed scheduler behavior "
		       "option.\n");
	print_usage(options->exec_name, BA_TRUE);
	return BA_STATUS_OPTIONS_BAD;	
    }

    if(option_actions->num_users_error) {
	ba_print_error("Couldn't parse the number of users "
		       "option.\n");
	print_usage(options->exec_name, BA_TRUE);
	return BA_STATUS_OPTIONS_BAD;	
    }
    if(option_actions->virtual_clock_terminate_error) {
	ba_print_error("Couldn't parse the virtual clock terminate option.\n");
	print_usage(options->exec_name, BA_TRUE);
	return BA_STATUS_OPTIONS_BAD;	
    }
    if(option_actions->max_req_dis_tasks_per_user_error) {
	ba_print_error("Couldn't parse the maximum requested and disclosed "
		       "only tasks\n"
		       "per user option.\n");
	print_usage(options->exec_name, BA_TRUE);
	return BA_STATUS_OPTIONS_BAD;	
    }

    if(option_actions->user_arrivals_dist_error) {
	ba_print_error("Couldn't parse user arrivals counting process "
		       "option.\n");
	print_usage(options->exec_name, BA_TRUE);
	return BA_STATUS_OPTIONS_BAD;
    }
    if(option_actions->task_set_change_range_dist_error) {
	ba_print_error("Couldn't parse task set change range option.\n");
	print_usage(options->exec_name, BA_TRUE);
	return BA_STATUS_OPTIONS_BAD;
    }
    if(option_actions->tasks_per_task_set_dist_error) {
	ba_print_error("Couldn't parse tasks per task set distribution "
		       "option.\n");
	print_usage(options->exec_name, BA_TRUE);
	return BA_STATUS_OPTIONS_BAD;
    }
    if(option_actions->service_delta_time_dist_error) {
	ba_print_error("Couldn't parse service time distribution option.\n");
	ba_options_show_options();
	print_usage(options->exec_name, BA_TRUE);
	return BA_STATUS_OPTIONS_BAD;
    }
    if(option_actions->think_delta_time_dist_error) {
	ba_print_error("Couldn't parse think time distribution option.\n");
	ba_options_show_options();
	print_usage(options->exec_name, BA_TRUE);
	return BA_STATUS_OPTIONS_BAD;
    }

    if(option_actions->show_options) {
	ba_options_show_options();
	ba_print_info("\n");
    }

    return BA_STATUS_SUCCESS;
}

static void
print_usage(char *progname,
	    ba_bool is_error)
{
    if(is_error) {
	ba_print_error(usage, progname);
    } else {
	ba_print_info(usage, progname);
    }
}

static void
print_banner(void)
{
    ba_print_info(banner);
}

extern void
ba_options_show_options(void)
{
    ba_print_info("Options:\n");

    if(options->command_line_options) {
	ba_print_info("  args: %s\n", options->command_line_options);
    } else {
	ba_print_info("  args: 0\n", options->command_line_options);
    }

    ba_print_info("  quiet: ");
    ba_types_bool_print(options->quiet);
    ba_print_info("\n");

    if(options->data_output_filename) {
	ba_print_info("  data output filename: %s\n",
		      options->data_output_filename);
    } else {
	ba_print_info("  data output filename: (stdout)\n");
    }

    ba_print_info("  random seed: %u\n", options->random_seed);
    ba_print_info("  any random seed: ");
    ba_types_bool_print(options->any_random_seed);
    ba_print_info("\n");

    if(options->trace_output_filename) {
	ba_print_info("  trace output filename: %s\n",
		      options->trace_output_filename);
    } else {
	ba_print_info("  trace output filename: (no trace output)\n");
    }
    if(options->trace_input_filename) {
	ba_print_info("  trace input filename: %s\n",
		      options->trace_input_filename);
    } else {
	ba_print_info("  trace input filename: (no trace input)\n");
    }

    ba_print_info("  treat requested and disclosed tasks differently: ");
    ba_types_bool_print(options->treat_req_dis_diff);
    ba_print_info("\n");
    ba_print_info("  requested sched behavior: ");
    ba_sched_behavior_print_short(options->req_sched_behavior);
    ba_print_info("\n");
    ba_print_info("  disclosed sched behavior: ");
    if(options->treat_req_dis_diff) {
	ba_sched_behavior_print_short(options->dis_sched_behavior);
    } else {
	ba_print_info("n/a (disclosed tasks treated as requested)");
    }
    ba_print_info("\n");

    ba_print_info("  num users: %d\n", options->num_users);
    ba_print_info("  virtual clock terminate: " BA_REAL_FMT "\n",
		  options->virtual_clock_terminate);

    ba_print_info("  max requested and disclosed only tasks per user: "
		  "%d\n", options->max_req_dis_tasks_per_user);

    if(options->user_arrivals_dist) {
	ba_print_info("User arrivals counting process:\n");
	ba_rv_dist_print_nice(options->user_arrivals_dist);
    } else {
	ba_print_info("User arrivals counting process: (not init'd)\n");
    }

    if(options->task_set_change_range_dist) {
	ba_print_info("Task set change range distribution:\n");
	ba_rv_dist_print_nice(options->task_set_change_range_dist);
    } else {
	ba_print_info("Task set change range distribution: (not init'd)\n");
    }

    if(options->tasks_per_task_set_dist) {
	ba_print_info("Tasks per task set distribution:\n");
	ba_rv_dist_print_nice(options->tasks_per_task_set_dist);
    } else {
	ba_print_info("Tasks per task set distribution: (not init'd)\n");
    }

    if(options->service_delta_time_dist) {
	ba_print_info("Service time distribution:\n");
	ba_rv_dist_print_nice(options->service_delta_time_dist);
    } else {
	ba_print_info("Service time distribution: (not init'd)\n");
    }

    if(options->think_delta_time_dist) {
	ba_print_info("Think time distribution:\n");
	ba_rv_dist_print_nice(options->think_delta_time_dist);
    } else {
	ba_print_info("Think time distribution: (not init'd)\n");
    }
}

static unsigned int
get_random_seed(void)
{
    unsigned int seed;
    int fd, ret;

    // wanted to use srandomdev(), but can't find it on OS X.

    fd = open("/dev/urandom", O_RDONLY);
    BA_ASSERT(fd != -1);
    ret = read(fd, &seed, sizeof(seed));
    BA_ASSERT(ret == sizeof(seed));
    ret = close(fd);
    BA_ASSERT(ret == 0);

    // or wuss out: return time(0) ^ (getpid() << 8);

    return seed;
}

extern ba_status
ba_options_init(void)
{
    options = ba_mem_malloc(sizeof(options_t));

    options->exec_name = 0;
    options->command_line_options = 0;
    options->quiet = BA_FALSE;
    options->data_output_filename = 0;
    options->random_seed = 1;
    options->any_random_seed = BA_FALSE;

    options->trace_output_filename = 0;
    options->trace_input_filename = 0;

    options->treat_req_dis_diff = BA_FALSE;
    options->req_sched_behavior = BA_SCHED_BEHAVIOR_FCFS;
    options->dis_sched_behavior = BA_SCHED_BEHAVIOR_FCFS;

    options->num_users = 10;
    options->virtual_clock_terminate = 60 * 60 * 24 * 4;
    options->max_req_dis_tasks_per_user = 0;

    options->user_arrivals_str = 0;
    options->task_set_change_range_str = 0;
    options->tasks_per_task_set_str = 0;
    options->service_delta_time_str = 0;
    options->think_delta_time_str = 0;

    option_actions = ba_mem_malloc(sizeof(option_actions_t));

    option_actions->wants_banner = BA_FALSE;
    option_actions->wants_usage = BA_FALSE;
    option_actions->show_options = BA_FALSE;

    option_actions->got_error = BA_FALSE;

    option_actions->random_seed_error = BA_FALSE;
    option_actions->any_random_seed_error = BA_FALSE;

    option_actions->trace_error = BA_FALSE;

    option_actions->treat_req_dis_diff_error = BA_FALSE;
    option_actions->req_sched_behavior_error = BA_FALSE;
    option_actions->dis_sched_behavior_error = BA_FALSE;

    option_actions->num_users_error = BA_FALSE;
    option_actions->virtual_clock_terminate_error = BA_FALSE;
    option_actions->max_req_dis_tasks_per_user_error = BA_FALSE;

    option_actions->user_arrivals_dist_error = BA_FALSE;
    option_actions->task_set_change_range_dist_error = BA_FALSE;
    option_actions->tasks_per_task_set_dist_error = BA_FALSE;
    option_actions->service_delta_time_dist_error = BA_FALSE;
    option_actions->think_delta_time_dist_error = BA_FALSE;

    return BA_STATUS_SUCCESS;
}

extern
ba_status ba_options_destroy(void)
{
    ba_mem_free(option_actions);

    if(options->exec_name) {
	ba_string_delete(options->exec_name);
    }
    if(options->command_line_options) {
	ba_string_delete(options->command_line_options);
    }
    if(options->data_output_filename) {
	ba_string_delete(options->data_output_filename);
    }
    if(options->trace_output_filename) {
	ba_string_delete(options->trace_output_filename);
    }
    if(options->trace_input_filename) {
	ba_string_delete(options->trace_input_filename);
    }

    if(options->user_arrivals_str) {
	ba_string_delete(options->user_arrivals_str);
    }
    if(options->task_set_change_range_str) {
	ba_string_delete(options->task_set_change_range_str);
    }
    if(options->tasks_per_task_set_str) {
	ba_string_delete(options->tasks_per_task_set_str);
    }
    if(options->service_delta_time_str) {
	ba_string_delete(options->service_delta_time_str);
    }
    if(options->think_delta_time_str) {
	ba_string_delete(options->think_delta_time_str);
    }

    if(options->user_arrivals_dist) {
	ba_rv_dist_delete(options->user_arrivals_dist);
    }
    if(options->task_set_change_range_dist) {
	ba_rv_dist_delete(options->task_set_change_range_dist);
    }
    if(options->tasks_per_task_set_dist) {
	ba_rv_dist_delete(options->tasks_per_task_set_dist);
    }
    if(options->service_delta_time_dist) {
	ba_rv_dist_delete(options->service_delta_time_dist);
    }
    if(options->think_delta_time_dist) {
	ba_rv_dist_delete(options->think_delta_time_dist);
    }

    ba_mem_free(options);

    return BA_STATUS_SUCCESS;
}

extern void
ba_options_destroy_from_hook(void *unused)
{
    ba_status status;

    status = ba_options_destroy();
    BA_ASSERT(status == BA_STATUS_SUCCESS);
}
