#include <ba_config.h>
#include <ba_sim.h>

#define GLIB_RAND

#if !defined(SYS_RAND) && !defined(GLIB_RAND)
#error need to specify rand type
#endif

typedef enum {
    DIST_KIND_UNIFORM = 0,
    DIST_KIND_EXPONENTIAL,
    DIST_KIND_PARETO,
    DIST_KIND_BERNOULLI,

    DIST_KIND_POISSON,

    DIST_KIND_UNKNOWN,
} dist_kind_t;

typedef struct {
#if defined(SYS_RAND)
    char rstate_val[256];
#elif defined(GLIB_RAND)
    GRand *rstate_val;
#endif
} rstate_t;

struct ba_rv_dist_s {
    char *name;
    dist_kind_t kind;

    rstate_t *rstate;

    ba_real alpha;
    ba_real beta;
    ba_real mean;
    ba_real prob;
};

typedef ba_bool (dist_verify_t)(ba_real alpha,
				ba_real beta,
				ba_real mean,
				ba_real prob);

typedef struct {
    char *name;
    dist_kind_t kind;
    dist_verify_t *verify;
    ba_bool alpha_used;
    ba_bool beta_used;
    ba_bool mean_used;
    ba_bool prob_used;
} dist_desc_t;

static uint32_t get_random(rstate_t *rstate);
static void dist_desc_print(dist_desc_t *dist_desc);
static void dist_desc_print_from_hash(gpointer key_unused,
				      gpointer dist_desc_in,
				      gpointer generic_data_unused);
static dist_desc_t *dist_desc_new(char *name,
				  dist_kind_t kind,
				  dist_verify_t *verify,
				  ba_bool alpha_used,
				  ba_bool beta_used,
				  ba_bool mean_used,
				  ba_bool prob_used);
static void dist_desc_delete(dist_desc_t *dist_desc);
static void dist_desc_delete_from_hash(gpointer dist_desc_in);
static dist_verify_t uniform_verify;
static ba_real uniform(ba_real alpha,
		       ba_real beta,
		       rstate_t *rstate);
static dist_verify_t bernoulli_verify;
static ba_real bernoulli(ba_real prob,
			 rstate_t *rstate);
static dist_verify_t exponential_verify;
static ba_real exponential(ba_real mean,
			   rstate_t *rstate);
static dist_verify_t pareto_verify;
static ba_real pareto(ba_real k,
		      ba_real p,
		      ba_real alpha,
		      rstate_t *rstate);
static ba_real regular_pareto(ba_real alpha,
			      ba_real u);
static ba_real bounded_pareto(ba_real k,
			      ba_real p,
			      ba_real alpha,
			      ba_real u);
static inline ba_real random_u(rstate_t *rstate);
static void dist_kind_print(dist_kind_t kind);
static void distributions_print(void);
static void distributions_init(void);
static void distributions_destroy(void);

static uint32_t get_next_rstate_seed(void);

static GHashTable *distributions;

static ba_bool need_to_set_meta_seed;
static rstate_t *meta_rstate;

#if defined(SYS_RAND)
static char sys_dummy_rstate[256];
static char *sys_current_rstate;
#endif

extern ba_bool
ba_rv_dist_values_used_get(char *name,
			   ba_bool *alpha_used,
			   ba_bool *beta_used,
			   ba_bool *mean_used,
			   ba_bool *prob_used)
{
    BA_ASSERT(name);
    dist_desc_t *dist_desc;

    dist_desc = g_hash_table_lookup(distributions, name);
    if(dist_desc == 0) {
	return BA_FALSE;
    }

    if(alpha_used) {
	*alpha_used = dist_desc->alpha_used;
    }
    if(beta_used) {
	*beta_used = dist_desc->beta_used;
    }
    if(mean_used) {
	*mean_used = dist_desc->mean_used;
    }
    if(prob_used) {
	*prob_used = dist_desc->prob_used;
    }

    return BA_TRUE;
}

extern void
ba_rv_distributions_print(void)
{
    distributions_print();
}

static void
dist_desc_print(dist_desc_t *dist_desc)
{
    BA_ASSERT(dist_desc);

    ba_print_info("rv dist desc %p:\n", dist_desc);
    ba_print_info("  name: %s\n", dist_desc->name);
    dist_kind_print(dist_desc->kind);
    ba_print_info("  verify: %p\n", dist_desc->verify);

    ba_print_info("  alpha_used: ");
    ba_types_bool_print(dist_desc->alpha_used);
    ba_print_info("\n");

    ba_print_info("  beta_used: ");
    ba_types_bool_print(dist_desc->beta_used);
    ba_print_info("\n");

    ba_print_info("  mean_used: ");
    ba_types_bool_print(dist_desc->mean_used);
    ba_print_info("\n");

    ba_print_info("  prob_used: ");
    ba_types_bool_print(dist_desc->prob_used);
    ba_print_info("\n");
}

static void
dist_desc_print_from_hash(gpointer key_unused,
			  gpointer dist_desc_in,
			  gpointer generic_data_unused)
{
    BA_ASSERT(key_unused && dist_desc_in);
    dist_desc_t *dist_desc;

    dist_desc = (dist_desc_t *)dist_desc_in;
    dist_desc_print(dist_desc);
}

static dist_desc_t *
dist_desc_new(char *name,
	      dist_kind_t kind,
	      dist_verify_t *verify,
	      ba_bool alpha_used,
	      ba_bool beta_used,
	      ba_bool mean_used,
	      ba_bool prob_used)
{
    BA_ASSERT(name && verify);
    dist_desc_t *dist_desc;
    
    dist_desc = ba_mem_malloc(sizeof(dist_desc_t));
    dist_desc->name = ba_string_new(name);
    dist_desc->kind = kind;
    dist_desc->verify = verify;
    dist_desc->alpha_used = alpha_used;
    dist_desc->beta_used = beta_used;
    dist_desc->mean_used = mean_used;
    dist_desc->prob_used = prob_used;

    return dist_desc;
}

static void
dist_desc_delete(dist_desc_t *dist_desc)
{
    BA_ASSERT(dist_desc);

    ba_string_delete(dist_desc->name);

    ba_mem_free(dist_desc);
}

static void
dist_desc_delete_from_hash(gpointer dist_desc_in)
{
    BA_ASSERT(dist_desc_in);
    dist_desc_t *dist_desc;

    dist_desc = (dist_desc_t *)dist_desc_in;
    dist_desc_delete(dist_desc);
}

extern void
ba_rv_setup_random_generator(void)
{
    unsigned int seed;

    BA_ASSERT(need_to_set_meta_seed);
    need_to_set_meta_seed = BA_FALSE;

    seed = ba_options_random_seed_get();

#if defined(SYS_RAND)
    initstate(seed, meta_rstate->rstate_val, 256);
    sys_current_rstate = meta_rstate->rstate_val;
#elif defined(GLIB_RAND)
    meta_rstate->rstate_val = g_rand_new_with_seed(seed);
#endif
}

static uint32_t
get_next_rstate_seed(void)
{
    uint32_t generator_seed;

    BA_ASSERT(need_to_set_meta_seed == BA_FALSE);
    generator_seed = get_random(meta_rstate);

    return generator_seed;
}

static uint32_t
get_random(rstate_t *rstate)
{
#if defined(SYS_RAND)
    ba_bool install_new;
#endif
    uint32_t random_value;

    BA_ASSERT(need_to_set_meta_seed == BA_FALSE);

#if defined(SYS_RAND)
    if(sys_current_rstate != meta_rstate->rstate_val) {
	install_new = BA_TRUE;
    } else {
	install_new = BA_FALSE;
    }

    if(install_new) {
	sys_current_rstate = rstate->rstate_val;
	setstate(rstate->rstate_val);
    }

    random_value = random();

    // gerr, need to change to dummy_rstate because it seems that
    // SOMETHING out there not in my code (maybe glib?) is calling
    // random() and screwing up my random generators.  thus, the code
    // above to check whether to install a new generator will always
    // be true, since we are always switching to this dummy rstate for
    // now.  if i switch to g_rand in glib, this will all change.  XXX
    sys_current_rstate = sys_dummy_rstate;
    setstate(sys_dummy_rstate);
#elif defined(GLIB_RAND)
    random_value = g_rand_int(rstate->rstate_val);
#endif

    return random_value;
}

extern char *
ba_rv_dist_name_get(ba_rv_dist_t *dist)
{
    BA_ASSERT(dist);

    return dist->name;
}

extern ba_real
ba_rv_dist(ba_rv_dist_t *dist)
{
    BA_ASSERT(dist);

    switch(dist->kind) {
    case DIST_KIND_UNIFORM:
	return uniform(dist->alpha, dist->beta, dist->rstate);
	break;
    case DIST_KIND_EXPONENTIAL:
	return exponential(dist->mean, dist->rstate);
	break;
    case DIST_KIND_PARETO:
	return pareto(dist->alpha, dist->beta, dist->mean, dist->rstate);
	break;
    case DIST_KIND_BERNOULLI:
	return bernoulli(dist->prob, dist->rstate);
	break;
    case DIST_KIND_POISSON:
	return exponential(dist->mean, dist->rstate);
	break;
    case DIST_KIND_UNKNOWN:
	BA_PANIC();
	break;
    }

    return 0.0; // not reached
}

static ba_bool
uniform_verify(ba_real alpha,
	       ba_real beta,
	       ba_real mean,
	       ba_real prob)
{
    // not using the tolerance comparators because these values are
    // virgin.  further, there is not a tolerance comparator for `>'
    // (as there is for `>=').

    // ross says that beta should be strictly greater than alpha.  you
    // need that to take the pdf at the point where alpha == beta,
    // least the prob == infinity.  but we are never taking the pdf.
    // also, why do i say that alpha >= 0?  i don't think ross
    // requires this.  XXX
    if(finite(alpha) && finite(beta) && beta >= alpha && alpha >= 0) {
	return BA_TRUE;
    } else {
	return BA_FALSE;
    }
}

static ba_bool
bernoulli_verify(ba_real alpha,
		 ba_real beta,
		 ba_real mean,
		 ba_real prob)
{
    // not using the tolerance comparators because these values are
    // virgin.
    if(prob >= 0.0 && prob <= 1.0) {
	return BA_TRUE;	
    } else {
	return BA_FALSE;
    }
}

static ba_real
uniform(ba_real alpha,
	ba_real beta,
	rstate_t *rstate)
{
    ba_real u, uniform;

    u = random_u(rstate);

    uniform = (beta - alpha) * u + alpha;

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(uniform, 0.0);

    return uniform;
}

static ba_real
bernoulli(ba_real prob,
	  rstate_t *rstate)
{
    ba_real u, bernoulli;

    u = random_u(rstate);

    if(u < prob) {
	bernoulli = 1.0;
    } else {
	bernoulli = 0.0;
    }

    return bernoulli;
}

static ba_bool
exponential_verify(ba_real alpha,
		   ba_real beta,
		   ba_real mean,
		   ba_real prob)
{
    // not using the tolerance comparators because these values are
    // virgin.  further, there is not a tolerance comparator for `>'
    // (as there is for `>=').
    if(finite(mean) && mean > 0) {
	return BA_TRUE;
    } else {
	return BA_FALSE;
    }
}

static ba_real
exponential(ba_real mean,
	    rstate_t *rstate)
{
    ba_real u, exponential;

    u = random_u(rstate);

    exponential = -mean * log(u);

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(exponential, 0.0);

    return exponential;
}

static ba_bool
pareto_verify(ba_real alpha_in,
	      ba_real beta,
	      ba_real mean,
	      ba_real prob)
{
    ba_real k, p, alpha;

    k = alpha_in;
    p = beta;
    alpha = mean;

    // not using the tolerance comparators because these values are
    // virgin.

    if(k == -1.0 && p != -1.0) {
	return BA_FALSE;
    }
    if(p == -1.0 && k != -1.0) {
	return BA_FALSE;
    }
    if(p != -1.0) {
	if(k > p) {
	    return BA_FALSE;
	}
	if(k <= 0.0) {
	    return BA_FALSE;
	}
    }

    if(alpha <= 0.0) {
	return BA_FALSE;
    }

    if(!finite(k) || !finite(p) || !finite(alpha)) {
	return BA_FALSE;
    }

    return BA_TRUE;
}

static ba_real
pareto(ba_real k,
       ba_real p,
       ba_real alpha,
       rstate_t *rstate)
{
    ba_real u;

    u = random_u(rstate);

    if(k == -1.0) {
	return regular_pareto(alpha, u);
    } else {
	return bounded_pareto(k, p, alpha, u);
    }
}

static ba_real
regular_pareto(ba_real alpha,
	       ba_real u)
{
    ba_real pareto;
    ba_real inv_alpha, one_minus_u, denom;

    one_minus_u = 1.0 - u;
    inv_alpha = 1.0 / alpha;

    BA_ASSERT(one_minus_u >= 0.0);
    BA_ASSERT(inv_alpha > 0.0);
    denom = pow(one_minus_u, inv_alpha);

    pareto = 1.0 / denom;
    
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(pareto, 0.0);

    return pareto;
}

static ba_real
bounded_pareto(ba_real k,
	       ba_real p,
	       ba_real alpha,
	       ba_real u)
{
    ba_real pareto;
    ba_real inv_alpha, k_over_p_to_alpha, denom_base, denom;

    BA_ASSERT(k / p >= 0.0);
    BA_ASSERT(alpha > 0.0);
    k_over_p_to_alpha = pow(k / p, alpha);

    denom_base = 1.0 + u * (k_over_p_to_alpha - 1.0);
    inv_alpha = 1.0 / alpha;

    BA_ASSERT(denom_base >= 0.0);
    BA_ASSERT(inv_alpha > 0.0);
    denom = pow(denom_base, inv_alpha);

    pareto = k / denom;

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(pareto, 0.0);

    return pareto;
}

static inline ba_real
random_u(rstate_t *rstate)
{
    uint32_t num;
    ba_real u;

    num = get_random(rstate);

#if defined(SYS_RAND)
    // INT32_MAX == (2 ** 31) - 1, which is the max that random(3) returns.
    u = (ba_real)num / (ba_real)INT32_MAX;
#elif defined(GLIB_RAND)
    // UINT32_MAX == (2 ** 32) - 1, which is the max that glib rand returns.
    u = (ba_real)num / (ba_real)UINT32_MAX;
#endif

    return u;
}

static void
dist_kind_print(dist_kind_t kind)
{
    switch(kind) {
    case DIST_KIND_UNIFORM:
	ba_print_info("  kind: uniform\n");
	break;
    case DIST_KIND_EXPONENTIAL:
	ba_print_info("  kind: exponential\n");
	break;
    case DIST_KIND_PARETO:
	ba_print_info("  kind: pareto\n");
	break;
    case DIST_KIND_BERNOULLI:
	ba_print_info("  kind: bernoulli\n");
	break;
    case DIST_KIND_POISSON:
	ba_print_info("  kind: poisson\n");
	break;
    case DIST_KIND_UNKNOWN:
	BA_PANIC();
	break;
    }
}

extern void
ba_rv_dist_print_nice(ba_rv_dist_t *dist)
{
    BA_ASSERT(dist);
    ba_bool ok;
    ba_bool alpha_used, beta_used, mean_used, prob_used;

    ba_print_info("Distribution name: %s\n", dist->name);
    ok = ba_rv_dist_values_used_get(dist->name, &alpha_used, &beta_used,
				    &mean_used, &prob_used);
    BA_ASSERT(ok);

    if(alpha_used) {
	ba_print_info("  alpha: " BA_REAL_FMT "\n", dist->alpha);
    }
    if(beta_used) {
	ba_print_info("  beta: " BA_REAL_FMT "\n", dist->beta);
    }
    if(mean_used) {
	ba_print_info("  mean: " BA_REAL_FMT "\n", dist->mean);
    }
    if(prob_used) {
	ba_print_info("  prob: " BA_REAL_FMT "\n", dist->prob);
    }
}

extern void
ba_rv_dist_print_short_to(ba_rv_dist_t *dist,
			  ba_print_func_t *print_func)
{
    BA_ASSERT(dist);
    BA_ASSERT(print_func);
    ba_bool ok, showed_one;
    ba_bool alpha_used, beta_used, mean_used, prob_used;

    print_func("%s:", dist->name);
    ok = ba_rv_dist_values_used_get(dist->name, &alpha_used, &beta_used,
				    &mean_used, &prob_used);
    BA_ASSERT(ok);

    showed_one = BA_FALSE;
    if(alpha_used) {
	if(showed_one) {
	    print_func(",");
	}
	print_func(BA_REAL_FMT, dist->alpha);
	showed_one = BA_TRUE;
    }
    if(beta_used) {
	if(showed_one) {
	    print_func(",");
	}
	print_func(BA_REAL_FMT, dist->beta);
	showed_one = BA_TRUE;
    }
    if(mean_used) {
	if(showed_one) {
	    print_func(",");
	}
	print_func(BA_REAL_FMT, dist->mean);
	showed_one = BA_TRUE;
    }
    if(prob_used) {
	if(showed_one) {
	    print_func(",");
	}
	print_func(BA_REAL_FMT, dist->prob);
	showed_one = BA_TRUE;
    }
}

extern void
ba_rv_dist_print_short(ba_rv_dist_t *dist)
{
    BA_ASSERT(dist);

    ba_rv_dist_print_short_to(dist, ba_print_info);
}

extern void
ba_rv_dist_print(ba_rv_dist_t *dist)
{
    BA_ASSERT(dist);

    ba_print_info("Distribution %p:\n", dist);
    ba_print_info("  name: %s\n", dist->name);
    dist_kind_print(dist->kind);
    ba_print_info("  (not all of the following values are used for all "
		  "dists.)\n");
    ba_print_info("  alpha: " BA_REAL_FMT "\n", dist->alpha);
    ba_print_info("  beta: " BA_REAL_FMT "\n", dist->beta);
    ba_print_info("  mean: " BA_REAL_FMT "\n", dist->mean);
    ba_print_info("  prob: " BA_REAL_FMT "\n", dist->prob);
}

extern ba_rv_dist_t *
ba_rv_dist_copy(ba_rv_dist_t *dist_src)
{
    BA_ASSERT(dist_src);
    ba_rv_dist_t *dist_dst;

    dist_dst = ba_rv_dist_new(dist_src->name, dist_src->alpha, dist_src->beta,
			      dist_src->mean, dist_src->prob);

    return dist_dst;
}

extern ba_rv_dist_t *
ba_rv_dist_new(char *name,
	       ba_real alpha,
	       ba_real beta,
	       ba_real mean,
	       ba_real prob)
{
    BA_ASSERT(name);
    ba_rv_dist_t *dist;
    dist_desc_t *dist_desc;
    ba_bool ok;
    uint32_t generator_seed;

    dist_desc = g_hash_table_lookup(distributions, name);
    if(dist_desc == 0) {
	return 0;
    }

    ok = dist_desc->verify(alpha, beta, mean, prob);
    if(ok == BA_FALSE) {
	return 0;
    }

    dist = ba_mem_malloc(sizeof(ba_rv_dist_t));

    dist->name = dist_desc->name; // name is deleted by dist_desc.
    dist->kind = dist_desc->kind;
    if(dist_desc->alpha_used) {
	dist->alpha = alpha;
    }
    if(dist_desc->beta_used) {
	dist->beta = beta;
    }
    if(dist_desc->mean_used) {
	dist->mean = mean;
    }
    if(dist_desc->prob_used) {
	dist->prob = prob;
    }

    generator_seed = get_next_rstate_seed();

    dist->rstate = ba_mem_malloc(sizeof(rstate_t));

#if defined(SYS_RAND)
    initstate(generator_seed, dist->rstate->rstate_val, 256);
    sys_current_rstate = dist->rstate->rstate_val;
#elif defined(GLIB_RAND)
    dist->rstate->rstate_val = g_rand_new_with_seed(generator_seed);
#endif

    return dist;
}

extern void
ba_rv_dist_delete(ba_rv_dist_t *dist)
{
    BA_ASSERT(dist);
    
#if defined(GLIB_RAND)
    g_rand_free(dist->rstate->rstate_val);
#endif

    ba_mem_free(dist->rstate);

    ba_mem_free(dist);
}

static void
distributions_print(void)
{
    g_hash_table_foreach(distributions, dist_desc_print_from_hash, 0);
}

static void
distributions_init(void)
{
    dist_desc_t *dist_desc;
    
    distributions = g_hash_table_new_full(g_str_hash, g_str_equal, 0,
					  dist_desc_delete_from_hash);
    BA_ASSERT(distributions);

    dist_desc = dist_desc_new("uniform", DIST_KIND_UNIFORM, uniform_verify,
			      BA_TRUE, BA_TRUE, BA_FALSE, BA_FALSE);
    g_hash_table_insert(distributions, "uniform", dist_desc);

    dist_desc = dist_desc_new("exponential", DIST_KIND_EXPONENTIAL,
			      exponential_verify, BA_FALSE, BA_FALSE, BA_TRUE,
			      BA_FALSE);
    g_hash_table_insert(distributions, "exponential", dist_desc);

    dist_desc = dist_desc_new("pareto", DIST_KIND_PARETO, pareto_verify,
			      BA_TRUE, BA_TRUE, BA_TRUE, BA_FALSE);
    g_hash_table_insert(distributions, "pareto", dist_desc);

    dist_desc = dist_desc_new("bernoulli", DIST_KIND_BERNOULLI,
			      bernoulli_verify, BA_FALSE, BA_FALSE, BA_FALSE,
			      BA_TRUE);
    g_hash_table_insert(distributions, "bernoulli", dist_desc);

    dist_desc = dist_desc_new("poisson", DIST_KIND_POISSON,
			      exponential_verify, BA_FALSE, BA_FALSE, BA_TRUE,
			      BA_FALSE);
    g_hash_table_insert(distributions, "poisson", dist_desc);
}

static void
distributions_destroy(void)
{
    g_hash_table_destroy(distributions);
}

extern ba_status
ba_rv_init(void)
{
    need_to_set_meta_seed = BA_TRUE;
    meta_rstate = ba_mem_malloc(sizeof(rstate_t));

#if defined(SYS_RAND)
    initstate(1, sys_dummy_rstate, 256);
    sys_current_rstate = 0;
#endif

#if defined(GLIB_RAND)
    meta_rstate->rstate_val = 0;
#endif

    distributions_init();

    return BA_STATUS_SUCCESS;
}

extern ba_status
ba_rv_destroy(void)
{
    distributions_destroy();

#if defined(GLIB_RAND)
    if(meta_rstate->rstate_val) {
	g_rand_free(meta_rstate->rstate_val);
    }
#endif

    ba_mem_free(meta_rstate);

    return BA_STATUS_SUCCESS;
}

extern void
ba_rv_destroy_from_hook(void *unused)
{
    ba_status status;

    status = ba_rv_destroy();
    BA_ASSERT(status == BA_STATUS_SUCCESS);
}
