#include <ba_config.h>
#include <string.h>
#include <ba_sim.h>

#define COUNT_SCHEDULER_DECISIONS 0

// only one should be used at once.  output can be sent to qlen_to_graph.pl
#define PRINT_NUM_ALL_CANDIDATES 0
#define PRINT_NUM_REQ_CANDIDATES 0
#define PRINT_NUM_DIS_CANDIDATES 0

typedef ba_task_t *(sched_func_t)(GSList *candidates);

static void reset_candidates(void);
static void all_candidates_print(void);
static void requested_candidates_print(void);
static void disclosed_candidates_print(void);

static void make_candidates(void);
static void make_candidates_helper(GSList **add_list,
				   ba_glib_fast_slist_append_t *appender,
				   uint32_t *add_len,
				   ba_real *add_weight,
				   GSList *iter_list);

static void set_sched_func(ba_sched_behavior_t sched_behavior,
			   sched_func_t **sched_func);

static sched_func_t fcfs;
static sched_func_t rfcfs;
static sched_func_t ufb;
static GSList *ufb_best_user(GSList *candidates);
static sched_func_t urfb;
static GSList *urfb_best_user(GSList *candidates);
static sched_func_t hrp;
static GSList *hrp_best_user(GSList *candidates);
static sched_func_t hrr;
static GSList *hrr_best_user(GSList *candidates);
static sched_func_t gsvsd;
static GSList *gsvsd_best_user(GSList *candidates);
static sched_func_t srpt;
static sched_func_t rsrpt;
static sched_func_t lprt;
static sched_func_t lprtp;
static sched_func_t rand_sched; // `_sched' to not conflict with internal rand
static sched_func_t randu;
static GSList *randu_best_user(GSList *candidates);
static sched_func_t noop;

static ba_bool treat_req_dis_diff;
static ba_sched_behavior_t requested_sched_behavior;
static ba_sched_behavior_t disclosed_sched_behavior;

static GSList *all_candidates, *requested_candidates,
    *disclosed_candidates;
static ba_glib_fast_slist_append_t *all_candidates_app,
    *requested_candidates_app, *disclosed_candidates_app;
uint32_t num_all_candidates, num_req_candidates, num_dis_candidates;
ba_real weighed_all_candidates, weighed_req_candidates, weighed_dis_candidates;

static sched_func_t *all_sched_func, *requested_sched_func,
    *disclosed_sched_func;

ba_rv_dist_t *rand_sched_dist;
ba_rv_dist_t *randu_sched_dist;

// XXX should be reporting this, but no time
ba_stats_t *sched_func_stats;
GTimer *sched_func_timer;

static void
reset_candidates(void)
{
    g_slist_free(all_candidates);
    g_slist_free(requested_candidates);
    g_slist_free(disclosed_candidates);

    all_candidates = 0;
    requested_candidates = 0;
    disclosed_candidates = 0;
    
    num_all_candidates = 0;
    num_req_candidates = 0;
    num_dis_candidates = 0;

    weighed_all_candidates = 0.0;
    weighed_req_candidates = 0.0;
    weighed_dis_candidates = 0.0;

    ba_glib_fast_slist_appender_reset(all_candidates_app);
    ba_glib_fast_slist_appender_reset(requested_candidates_app);
    ba_glib_fast_slist_appender_reset(disclosed_candidates_app);
}

extern void
ba_sched_candidates_print(void)
{
    make_candidates();
   
    all_candidates_print();
    requested_candidates_print();
    disclosed_candidates_print();

    reset_candidates();
}

static void
all_candidates_print(void)
{
    ba_print_info("All candidates:\n");
    g_slist_foreach(all_candidates, ba_task_print_from_list, 0);
}

static void
requested_candidates_print(void)
{
    ba_print_info("Requested candidates:\n");
    g_slist_foreach(requested_candidates, ba_task_print_from_list, 0);
}

static void
disclosed_candidates_print(void)
{
    ba_print_info("Disclosed candidates:\n");
    g_slist_foreach(disclosed_candidates, ba_task_print_from_list, 0);
}

static void
make_candidates(void)
{
    GSList *node;

    // assumption held by the schedulers: all tasks from a particular
    // user must be back-to-back in the lists i'm creating here.
    // this is ensured by this outer loop that goes across the
    // users.
    for(node = ba_users_get(); node; node = g_slist_next(node)) {
	ba_user_t *user;
	ba_bool stop_running_tasks;
	GSList *requested_tasks;
	GSList *disclosed_tasks;

	user = (ba_user_t *)node->data;
	BA_ASSERT(user);
	
	stop_running_tasks = ba_user_stop_running_tasks_get(user);
	if(stop_running_tasks) {
	    continue;
	}

	requested_tasks = ba_user_requested_tasks_get(user);
	disclosed_tasks = ba_user_disclosed_tasks_get(user);

	make_candidates_helper(&requested_candidates, requested_candidates_app,
			       &num_req_candidates, &weighed_req_candidates,
			       requested_tasks);
	make_candidates_helper(&disclosed_candidates, disclosed_candidates_app,
			       &num_dis_candidates, &weighed_dis_candidates,
			       disclosed_tasks);
    }
}

static void
make_candidates_helper(GSList **add_list,
		       ba_glib_fast_slist_append_t *appender,
		       uint32_t *add_len,
		       ba_real *add_weight,
		       GSList *iter_list)
{
    GSList *node;
    
    // disclosed tasks from a single user are
    // ordered in the way that the user will wind up requesting
    // them.  the dissertation discusses when this is realistic and
    // when it is now.  (in think_about_output() in ba_user.c, requests
    // are always taken from the front of the disclosed list.)

    for(node = iter_list; node; node = g_slist_next(node)) {
	ba_task_t *task;
	
	task = (ba_task_t *)node->data;
	BA_ASSERT(task);

	if(ba_task_running_get(task)) {
	    continue;
	}

	*add_list = ba_glib_fast_slist_append(appender, *add_list, task);
	(*add_len)++;
	(*add_weight) += ba_task_remaining_delta_time_get(task);
	all_candidates = ba_glib_fast_slist_append(all_candidates_app,
						   all_candidates, task);
	num_all_candidates++;
	weighed_all_candidates += ba_task_remaining_delta_time_get(task);
    }
}

extern void
ba_sched_choose(void)
{
    BA_ASSERT(ba_resource_is_idle());
    ba_task_t *task;
#if COUNT_SCHEDULER_DECISIONS
    gdouble elapsed_time;
#endif
#if BA_SIM_INSPECT
    ba_bool req_empty = BA_FALSE;
#endif

    make_candidates();

#if PRINT_NUM_ALL_CANDIDATES
    ba_print_info("%f %d\n", ba_events_virtual_clock_abs_time_get(),
		  num_all_candidates);
#endif
#if PRINT_NUM_REQ_CANDIDATES
    ba_print_info("%f %d\n", ba_events_virtual_clock_abs_time_get(),
		  num_req_candidates);
#endif
#if PRINT_NUM_DIS_CANDIDATES
    ba_print_info("%f %d\n", ba_events_virtual_clock_abs_time_get(),
		  num_dis_candidates);
#endif

    if(ba_report_warmup_met()) {
	ba_report_sched_candidates(num_all_candidates,
				   num_req_candidates,
				   num_dis_candidates,
				   ba_events_virtual_clock_abs_time_get());

	ba_report_sched_weighed_candidates(weighed_all_candidates,
					   weighed_req_candidates,
					   weighed_dis_candidates,
				   ba_events_virtual_clock_abs_time_get());
    }

#if COUNT_SCHEDULER_DECISIONS
    g_timer_start(sched_func_timer);
#endif

    if(treat_req_dis_diff == BA_FALSE) {
	task = all_sched_func(all_candidates);
    } else {
	task = requested_sched_func(requested_candidates);
	if(task == 0) {

#if BA_SIM_INSPECT
	    req_empty = BA_TRUE;
#endif

	    task = disclosed_sched_func(disclosed_candidates);
	}
    }

#if BA_SIM_INSPECT
    if(task) {
	if(req_empty) {
	    ba_print_info("scheduler selected user %d's task %d (remain: "
			  BA_REAL_LOW_PREC_FMT ") from dis queue at time "
			  BA_REAL_LOW_PREC_FMT "\n",
			  ba_user_id_get(ba_task_user_get(task)),
			  ba_task_id_get(task),
			  ba_task_remaining_delta_time_get(task),
			  ba_events_virtual_clock_abs_time_get());
	} else {
	    ba_print_info("scheduler selected user %d's task %d (remain: "
			  BA_REAL_LOW_PREC_FMT ") from req queue at time "
			  BA_REAL_LOW_PREC_FMT "\n",
			  ba_user_id_get(ba_task_user_get(task)),
			  ba_task_id_get(task),
			  ba_task_remaining_delta_time_get(task),
			  ba_events_virtual_clock_abs_time_get());
	}
    } else {
	ba_print_info("scheduler found no candidate task at time "
		      BA_REAL_LOW_PREC_FMT "\n",
		      ba_events_virtual_clock_abs_time_get());
    }
#endif

#if COUNT_SCHEDULER_DECISIONS
    elapsed_time = g_timer_elapsed(sched_func_timer, 0);
    if(ba_report_warmup_met()) {
	ba_stats_insert(sched_func_stats, elapsed_time);
    }
#endif

    if(task) {
	ba_events_add_event(BA_EVENT_KIND_RESOURCE_BUSY, BA_EVENTS_NOW,
			    ba_resource_busy_callback, task);
    }

    reset_candidates();
}

//////////////////////////////////////////////////////////////////////

static ba_task_t *
fcfs(GSList *candidates)
{
    ba_real earliest_abs_time;
    ba_bool first;
    GSList *node;
    ba_task_t *best_task;

    earliest_abs_time = 0.0; // appease the compiler.
    first = BA_TRUE;
    best_task = 0;

    for(node = candidates; node; node = g_slist_next(node)) {
	ba_task_t *task;
	ba_real task_abs_time;

	task_abs_time = 0.0; // warning elimination.

	task = (ba_task_t *)node->data;
	BA_ASSERT(task);

	if(ba_options_treat_req_dis_diff_get()) {
	    if(candidates == requested_candidates) {
		// inter and ba
		task_abs_time = ba_task_requested_abs_time_get(task);
	    } else if(candidates == disclosed_candidates) {
		// only ba (inter disclosed tasks get shunted to NOOP)
		task_abs_time = ba_task_disclosed_abs_time_get(task);
	    } else {
		BA_PANIC();
	    }
	} else {
	    // batch
	    BA_ASSERT(candidates == all_candidates);

	    // whether or not the task is now `requested' (needed),
	    // this is when it entered the system
	    task_abs_time = ba_task_disclosed_abs_time_get(task);
	}

	/*	ba_print_info("b4: %p, %f, %f\n", task, task_abs_time,
	  earliest_abs_time);*/

	if(first) {
	    first = BA_FALSE;
	    earliest_abs_time = task_abs_time;
	    best_task = task;
	} else {
	    // ok to use regular comparator since either way a task
	    // will get chosen.
	    if(task_abs_time < earliest_abs_time) {
		earliest_abs_time = task_abs_time;
		best_task = task;
	    }
	}

	/*	ba_print_info("after: %p, %f, %f\n", task, task_abs_time,
	  earliest_abs_time);*/
    }

    //    ba_print_info("fcfs choose %p\n", best_task);

    return best_task;
}

static ba_task_t *
rfcfs(GSList *candidates)
{
    ba_real earliest_abs_time;
    ba_bool first;
    GSList *node;
    ba_task_t *best_task;

    earliest_abs_time = 0.0; // appease the compiler.
    first = BA_TRUE;
    best_task = 0;

    for(node = candidates; node; node = g_slist_next(node)) {
	ba_task_t *task;
	ba_real task_abs_time;

	task_abs_time = 0.0; // warning elimination.

	task = (ba_task_t *)node->data;
	BA_ASSERT(task);

	if(ba_task_will_be_requested_get(task) == BA_FALSE) {
	    // this is how this policy is different than fcfs
	    continue;
	}

	/* we can panic on non-ba non-dis queue cases... */

	if(ba_options_treat_req_dis_diff_get()) {
	    if(candidates == requested_candidates) {
		BA_PANIC();
	    } else if(candidates == disclosed_candidates) {
		task_abs_time = ba_task_disclosed_abs_time_get(task);
	    } else {
		BA_PANIC();
	    }
	} else {
	    BA_PANIC();
	}

      /*ba_print_info("%p, %f, %f\n", task, task_abs_time, earliest_abs_time);*/

	if(first) {
	    first = BA_FALSE;
	    earliest_abs_time = task_abs_time;
	    best_task = task;
	} else {
	    // ok to use regular comparator since either way a task
	    // will get chosen.
	    if(task_abs_time < earliest_abs_time) {
		earliest_abs_time = task_abs_time;
		best_task = task;
	    }
	}
    }

    /*    ba_print_info("choose %p\n", best_task);*/
    /*    if(best_task) {
	ba_print_info("confirm: %f, %f\n", earliest_abs_time,
		      ba_task_disclosed_abs_time_get(best_task));
		      }*/

    return best_task;
}

//////////////////////////////////////////////////////////////////////

static ba_task_t *
ufb(GSList *candidates)
{
    GSList *node_starting_best_user;
    ba_task_t *best_task;

    node_starting_best_user = ufb_best_user(candidates);
    if(node_starting_best_user == 0) {
	return 0;
    }

    best_task = (ba_task_t *)node_starting_best_user->data;
    BA_ASSERT(best_task);

    return best_task;
}

static GSList *
ufb_best_user(GSList *candidates)
{
    ba_bool first;
    GSList *node, *node_starting_best_user;
    ba_real smallest_resource_total_usage;
    ba_user_t *last_user_seen;

    smallest_resource_total_usage = 0.0; // appease the compiler.
    last_user_seen = 0; // appease the compiler.
    node_starting_best_user = 0;
    first = BA_TRUE;

    for(node = candidates; node; node = g_slist_next(node)) {
	ba_task_t *task;
	ba_user_t *user;
	ba_real resource_total_usage;

	task = (ba_task_t *)node->data;
	BA_ASSERT(task);

	user = ba_task_user_get(task);
	resource_total_usage = ba_user_resource_total_usage_get(user);

	/*	ba_print_info("%p, %f, %f\n", user, resource_total_usage,
	  smallest_resource_total_usage);*/

	if(first) {
	    first = BA_FALSE;
	    smallest_resource_total_usage = resource_total_usage;
	    node_starting_best_user = node;
	    last_user_seen = user;
	} else {
	    if(last_user_seen != user) {
		// ok to use regular comparator since either way a task
		// will get chosen.
		if(resource_total_usage < smallest_resource_total_usage) {
		    smallest_resource_total_usage = resource_total_usage;
		    node_starting_best_user = node;
		}
		last_user_seen = user;
	    }
	}
    }

    /*    if(node_starting_best_user) {
	ba_task_t *task;
	ba_user_t *user;

	task = (ba_task_t *)node_starting_best_user->data;
	BA_ASSERT(task);
	  
	user = ba_task_user_get(task);
	ba_print_info("confirm: %f %f\n",
		      ba_user_resource_total_usage_get(user),
		      smallest_resource_total_usage);
		      }*/

    return node_starting_best_user;
}

static ba_task_t *
urfb(GSList *candidates)
{
    GSList *node_starting_best_user;
    ba_task_t *best_task;

    node_starting_best_user = urfb_best_user(candidates);
    if(node_starting_best_user == 0) {
	return 0;
    }

    best_task = (ba_task_t *)node_starting_best_user->data;
    BA_ASSERT(best_task);

    return best_task;
}

static GSList *
urfb_best_user(GSList *candidates)
{
    ba_bool first;
    GSList *node, *node_starting_best_user;
    ba_real smallest_req_resource_total_usage;
    ba_user_t *last_user_seen;

    smallest_req_resource_total_usage = 0.0; // appease the compiler.
    last_user_seen = 0; // appease the compiler.
    node_starting_best_user = 0;
    first = BA_TRUE;

    for(node = candidates; node; node = g_slist_next(node)) {
	ba_task_t *task;
	ba_user_t *user;
	ba_real req_resource_total_usage;

	task = (ba_task_t *)node->data;
	BA_ASSERT(task);

	user = ba_task_user_get(task);
	req_resource_total_usage =
	    ba_user_running_req_resource_total_usage_get(user);

	/*	ba_print_info("%p, %f, %f\n", user, req_resource_total_usage,
	  smallest_req_resource_total_usage);*/

	if(first) {
	    first = BA_FALSE;
	    smallest_req_resource_total_usage = req_resource_total_usage;
	    node_starting_best_user = node;
	    last_user_seen = user;
	} else {
	    if(last_user_seen != user) {
		// ok to use regular comparator since either way a task
		// will get chosen.
		if(req_resource_total_usage <
		   smallest_req_resource_total_usage) {
		    smallest_req_resource_total_usage =
			req_resource_total_usage;
		    node_starting_best_user = node;
		}
		last_user_seen = user;
	    }
	}
    }

    /*    if(node_starting_best_user) {
	ba_task_t *task;
	ba_user_t *user;

	task = (ba_task_t *)node_starting_best_user->data;
	BA_ASSERT(task);
	  
	user = ba_task_user_get(task);
	ba_print_info("confirm: %f %f\n",
		      ba_user_running_req_resource_total_usage_get(user),
		      smallest_req_resource_total_usage);
		      }*/

    return node_starting_best_user;
}

//////////////////////////////////////////////////////////////////////

static ba_task_t *
gsvsd(GSList *candidates)
{
    GSList *node_starting_best_user;
    ba_task_t *best_task;
    
    node_starting_best_user = gsvsd_best_user(candidates);
    if(node_starting_best_user == 0) {
	return 0;
    }
    
    best_task = (ba_task_t *)node_starting_best_user->data;
    BA_ASSERT(best_task);
    
    return best_task;
}

static GSList *
gsvsd_best_user(GSList *candidates)
{
    ba_bool first;
    GSList *node, *node_starting_best_user;
    ba_real greatest_vsd;
    ba_user_t *last_user_seen;

    greatest_vsd = 0.0; // appease the compiler.
    last_user_seen = 0; // appease the compiler.
    node_starting_best_user = 0;
    first = BA_TRUE;

    for(node = candidates; node; node = g_slist_next(node)) {
	ba_task_t *task;
	ba_user_t *user;
	ba_real vsd;

	task = (ba_task_t *)node->data;
	BA_ASSERT(task);

	user = ba_task_user_get(task);
	vsd = ba_user_total_vsd_get(user);

	if(first) {
	    first = BA_FALSE;
	    greatest_vsd = vsd;
	    node_starting_best_user = node;
	    last_user_seen = user;
	} else {
	    if(last_user_seen != user) {
		// ok to use regular comparator since either way a task
		// will get chosen.
		if(vsd > greatest_vsd) {
		    greatest_vsd = vsd;
		    node_starting_best_user = node;
		}
		last_user_seen = user;
	    }
	}
    }

    return node_starting_best_user;
}

//////////////////////////////////////////////////////////////////////

static ba_task_t *
srpt(GSList *candidates)
{
    ba_real shortest_remaining_delta_time;
    ba_bool first;
    GSList *node;
    ba_task_t *best_task;

    shortest_remaining_delta_time = 0.0; // appease the compiler.
    first = BA_TRUE;
    best_task = 0;

    for(node = candidates; node; node = g_slist_next(node)) {
	ba_task_t *task;
	ba_real remaining_delta_time;

	task = (ba_task_t *)node->data;
	BA_ASSERT(task);

	remaining_delta_time = ba_task_remaining_delta_time_get(task);

	/*	ba_print_info("%p, %f, %f\n", task, remaining_delta_time,
	  shortest_remaining_delta_time);*/

	if(first) {
	    first = BA_FALSE;
	    shortest_remaining_delta_time = remaining_delta_time;
	    best_task = task;
	} else {
	    // ok to use regular comparator since either way a task
	    // will get chosen.
	    if(remaining_delta_time < shortest_remaining_delta_time) {
		shortest_remaining_delta_time = remaining_delta_time;
		best_task = task;
	    }
	}
    }

    /*    if(best_task) {
	  ba_print_info("confirm: %f, %f\n", shortest_remaining_delta_time,
			ba_task_remaining_delta_time_get(best_task));
			}*/

    return best_task;
}

static ba_task_t *
rsrpt(GSList *candidates)
{
    ba_real shortest_remaining_delta_time;
    ba_bool first;
    GSList *node;
    ba_task_t *best_task;

    shortest_remaining_delta_time = 0.0; // appease the compiler.
    first = BA_TRUE;
    best_task = 0;

    for(node = candidates; node; node = g_slist_next(node)) {
	ba_task_t *task;
	ba_real remaining_delta_time;

	task = (ba_task_t *)node->data;
	BA_ASSERT(task);

	if(ba_task_will_be_requested_get(task) == BA_FALSE) {
	    continue;
	}

	remaining_delta_time = ba_task_remaining_delta_time_get(task);

	if(first) {
	    first = BA_FALSE;
	    shortest_remaining_delta_time = remaining_delta_time;
	    best_task = task;
	} else {
	    // ok to use regular comparator since either way a task
	    // will get chosen.
	    if(remaining_delta_time < shortest_remaining_delta_time) {
		shortest_remaining_delta_time = remaining_delta_time;
		best_task = task;
	    }
	}
    }

    return best_task;
}

//////////////////////////////////////////////////////////////////////

static ba_task_t *
hrp(GSList *candidates)
{
    GSList *node_starting_best_user;
    ba_task_t *best_task;

    node_starting_best_user = hrp_best_user(candidates);
    if(node_starting_best_user == 0) {
	return 0;
    }

    best_task = (ba_task_t *)node_starting_best_user->data;
    BA_ASSERT(best_task);

    /*    ba_print_info("hrp choose %p\n", best_task);*/

    return best_task;
}

static GSList *
hrp_best_user(GSList *candidates)
{
    ba_bool first;
    GSList *node, *node_starting_best_user;
    ba_real highest_request_probability;
    ba_user_t *last_user_seen;

    highest_request_probability = 0.0; // appease the compiler.
    last_user_seen = 0; // appease the compiler.
    node_starting_best_user = 0;
    first = BA_TRUE;

    for(node = candidates; node; node = g_slist_next(node)) {
	ba_task_t *task;
	ba_user_t *user;
	ba_real request_probability;

	task = (ba_task_t *)node->data;
	BA_ASSERT(task);

	user = ba_task_user_get(task);
	request_probability = ba_user_request_probability_get(user);

	/*	ba_print_info("%p, %f, %f\n", user, request_probability,
	  highest_request_probability);*/

	if(first) {
	    first = BA_FALSE;
	    highest_request_probability = request_probability;
	    node_starting_best_user = node;
	    last_user_seen = user;
	} else {
	    if(last_user_seen != user) {
		// ok to use regular comparator since either way a task
		// will get chosen.
		if(request_probability > highest_request_probability) {
		    highest_request_probability = request_probability;
		    node_starting_best_user = node;
		}
		last_user_seen = user;
	    }
	}
    }

    /*    if(node_starting_best_user) {
	  ba_task_t *task;
	  ba_user_t *user;
	  ba_print_info("%f\n", highest_request_probability);
	  task = (ba_task_t *)node_starting_best_user->data;
	  BA_ASSERT(task);
	  
	  user = ba_task_user_get(task);
	  ba_print_info("confirm: %f\n", ba_user_request_probability_get(user));
	  }*/

    return node_starting_best_user;
}

//////////////////////////////////////////////////////////////////////

static ba_task_t *
hrr(GSList *candidates)
{
    GSList *node_starting_best_user;
    ba_task_t *best_task;

    node_starting_best_user = hrr_best_user(candidates);
    if(node_starting_best_user == 0) {
	return 0;
    }

    best_task = (ba_task_t *)node_starting_best_user->data;
    BA_ASSERT(best_task);

    return best_task;
}

static GSList *
hrr_best_user(GSList *candidates)
{
    ba_bool first;
    GSList *node, *node_starting_best_user;
    ba_real highest_req_resource_total_usage;
    ba_user_t *last_user_seen;

    highest_req_resource_total_usage = 0.0; // appease the compiler.
    last_user_seen = 0; // appease the compiler.
    node_starting_best_user = 0;
    first = BA_TRUE;

    for(node = candidates; node; node = g_slist_next(node)) {
	ba_task_t *task;
	ba_user_t *user;
	ba_real req_resource_total_usage;

	task = (ba_task_t *)node->data;
	BA_ASSERT(task);

	user = ba_task_user_get(task);
	req_resource_total_usage =
	    ba_user_running_req_resource_total_usage_get(user);

	/*	ba_print_info("%p, %f, %f\n", user, req_resource_total_usage,
	  highest_req_resource_total_usage);*/

	if(first) {
	    first = BA_FALSE;
	    highest_req_resource_total_usage = req_resource_total_usage;
	    node_starting_best_user = node;
	    last_user_seen = user;
	} else {
	    if(last_user_seen != user) {
		// ok to use regular comparator since either way a task
		// will get chosen.
		if(req_resource_total_usage >
		   highest_req_resource_total_usage) {
		    highest_req_resource_total_usage = req_resource_total_usage;
		    node_starting_best_user = node;
		}
		last_user_seen = user;
	    }
	}
    }

    /*    if(node_starting_best_user) {
	ba_task_t *task;
	ba_user_t *user;

	task = (ba_task_t *)node_starting_best_user->data;
	BA_ASSERT(task);
	
	user = ba_task_user_get(task);
	ba_print_info("confirm: %f %f\n",
		      ba_user_running_req_resource_total_usage_get(user),
		      highest_req_resource_total_usage);
		      }*/

    return node_starting_best_user;
}

//////////////////////////////////////////////////////////////////////

// Mor suggests OrigSize * (1 - p) ^ i.  for now, i == 1, statically.  XXX
static ba_task_t *
lprt(GSList *candidates)
{
    ba_real shortest_prob_scaled_remaining_delta_time;
    ba_bool first;
    GSList *node;
    ba_task_t *best_task;

    shortest_prob_scaled_remaining_delta_time = 0.0; // appease the compiler.
    first = BA_TRUE;
    best_task = 0;

    for(node = candidates; node; node = g_slist_next(node)) {
	ba_task_t *task;
	ba_real remaining_delta_time;
	ba_real prob_scaled_remaining_delta_time;
	ba_real task_set_change_prob;
	ba_user_t *user;

	task = (ba_task_t *)node->data;
	BA_ASSERT(task);

	user = ba_task_user_get(task);
	task_set_change_prob =
	    ba_user_task_set_change_prob_get(user);

	remaining_delta_time = ba_task_remaining_delta_time_get(task);

	prob_scaled_remaining_delta_time = (1.0 - task_set_change_prob) *
	    remaining_delta_time;

	if(first) {
	    first = BA_FALSE;
	    shortest_prob_scaled_remaining_delta_time =
		prob_scaled_remaining_delta_time;
	    best_task = task;
	} else {
	    // ok to use regular comparator since either way a task
	    // will get chosen.
	    if(prob_scaled_remaining_delta_time <
	       shortest_prob_scaled_remaining_delta_time) {
		shortest_prob_scaled_remaining_delta_time =
		    prob_scaled_remaining_delta_time;
		best_task = task;
	    }
	}
    }

    return best_task;
}

// Mor suggests OrigSize * (1 - p) ^ i.  for now, i == 1, statically.  XXX
// definition of a deadline before far away is foo.
static ba_task_t *
lprtp(GSList *candidates)
{
    // XXX i think this is just LPRT (no postponement) now.
    ba_real shortest_prob_scaled_remaining_delta_time;
    ba_bool first;
    GSList *node;
    ba_task_t *best_task;

    // don't want to do this until i get a better handle on the
    // performance of other scheds.
    BA_PANIC();

    shortest_prob_scaled_remaining_delta_time = 0.0; // appease the compiler.
    first = BA_TRUE;
    best_task = 0;

    for(node = candidates; node; node = g_slist_next(node)) {
	ba_task_t *task;
	ba_real remaining_delta_time;
	ba_real prob_scaled_remaining_delta_time;
	ba_real task_set_change_prob;
	ba_user_t *user;

	task = (ba_task_t *)node->data;
	BA_ASSERT(task);

	user = ba_task_user_get(task);
	task_set_change_prob =
	    ba_user_task_set_change_prob_get(user);

	remaining_delta_time = ba_task_remaining_delta_time_get(task);

	prob_scaled_remaining_delta_time = (1.0 - task_set_change_prob) *
	    remaining_delta_time;

	if(first) {
	    first = BA_FALSE;
	    shortest_prob_scaled_remaining_delta_time =
		prob_scaled_remaining_delta_time;
	    best_task = task;
	} else {
	    // ok to use regular comparator since either way a task
	    // will get chosen.
	    if(prob_scaled_remaining_delta_time <
	       shortest_prob_scaled_remaining_delta_time) {
		shortest_prob_scaled_remaining_delta_time =
		    prob_scaled_remaining_delta_time;
		best_task = task;
	    }
	}
    }

    return best_task;
}

//////////////////////////////////////////////////////////////////////

static ba_task_t *
rand_sched(GSList *candidates)
{
    uint32_t total_tasks, selected_task;
    ba_real rv;
    ba_task_t *task;
    int i;
    GSList *node;

    total_tasks = g_slist_length(candidates);
    if(total_tasks == 0) {
	return 0;
    }

    rv = ba_rv_dist(rand_sched_dist);
    if(rv == 1.0) {
	// exceedingly unlikely to occur
	selected_task = total_tasks - 1;
    } else {
	selected_task = floor(rv * (ba_real)total_tasks);
    }

    i = 0;
    for(node = candidates; node; node = g_slist_next(node)) {
	if(i == selected_task) {
	    break;
	}
	i++;
    }
    BA_ASSERT(node);

    task = (ba_task_t *)node->data;
    BA_ASSERT(task);

    return task;
}

static ba_task_t *
randu(GSList *candidates)
{
    GSList *node_starting_best_user;
    ba_task_t *best_task;
    
    node_starting_best_user = randu_best_user(candidates);
    if(node_starting_best_user == 0) {
	return 0;
    }
    
    best_task = (ba_task_t *)node_starting_best_user->data;
    BA_ASSERT(best_task);
    
    return best_task;
}

static GSList *
randu_best_user(GSList *candidates)
{
    uint32_t total_users, selected_user, i;
    ba_real rv;
    ba_user_t *last_user_seen;
    ba_bool first;
    GSList *node;

    // find the number of users
    total_users = 0;
    last_user_seen = 0;
    first = BA_TRUE;

    for(node = candidates; node; node = g_slist_next(node)) {
	ba_task_t *task;
	ba_user_t *user;

	task = (ba_task_t *)node->data;
	BA_ASSERT(task);

	user = ba_task_user_get(task);
	if(first) {
	    first = BA_FALSE;
	    last_user_seen = user;
	    total_users++;
	} else {
	    if(user != last_user_seen) {
		last_user_seen = user;
		total_users++;
	    }
	}
    }
    if(total_users == 0) {
	return 0;
    }

    rv = ba_rv_dist(randu_sched_dist);
    if(rv == 1.0) {
	// exceedingly unlikely to occur
	selected_user = total_users - 1;
    } else {
	selected_user = floor(rv * (ba_real)total_users);
    }

    // find the selected user
    i = 0;
    last_user_seen = 0;
    first = BA_TRUE;

    for(node = candidates; node; node = g_slist_next(node)) {
	ba_task_t *task;
	ba_user_t *user;

	if(i == selected_user) {
	    break;
	}

	task = (ba_task_t *)node->data;
	BA_ASSERT(task);

	user = ba_task_user_get(task);
	if(first) {
	    first = BA_FALSE;
	    last_user_seen = user;
	    i++;
	} else {
	    if(user != last_user_seen) {
		last_user_seen = user;
		i++;
	    }
	}
    }
    BA_ASSERT(node);

    return node;
}

//////////////////////////////////////////////////////////////////////

static ba_task_t *
noop(GSList *candidates)
{
    return 0;
}

//////////////////////////////////////////////////////////////////////

extern ba_sched_behavior_t
ba_sched_str_to_sched_behavior(char *str)
{
    if(strcasecmp(str, "FCFS") == 0) {
	return BA_SCHED_BEHAVIOR_FCFS;
    } else if(strcasecmp(str, "RFCFS") == 0) {
	return BA_SCHED_BEHAVIOR_RFCFS;
    } else if(strcasecmp(str, "PS") == 0) {
	return BA_SCHED_BEHAVIOR_PS;
    } else if(strcasecmp(str, "UFB") == 0) {
	return BA_SCHED_BEHAVIOR_UFB;
    } else if(strcasecmp(str, "URFB") == 0) {
	return BA_SCHED_BEHAVIOR_URFB;
    } else if(strcasecmp(str, "DUFB") == 0) {
	return BA_SCHED_BEHAVIOR_DUFB;
    } else if(strcasecmp(str, "GSVSD") == 0) {
	return BA_SCHED_BEHAVIOR_GSVSD;
    } else if(strcasecmp(str, "SRPT") == 0) {
	return BA_SCHED_BEHAVIOR_SRPT;
    } else if(strcasecmp(str, "RSRPT") == 0) {
	return BA_SCHED_BEHAVIOR_RSRPT;
    } else if(strcasecmp(str, "HRP") == 0) {
	return BA_SCHED_BEHAVIOR_HRP;
    } else if(strcasecmp(str, "HRR") == 0) {
	return BA_SCHED_BEHAVIOR_HRR;
    } else if(strcasecmp(str, "EDF") == 0) {
	return BA_SCHED_BEHAVIOR_EDF;
    } else if(strcasecmp(str, "LPRT") == 0) {
	return BA_SCHED_BEHAVIOR_LPRT;
    } else if(strcasecmp(str, "LPRTP") == 0) {
	return BA_SCHED_BEHAVIOR_LPRTP;
    } else if(strcasecmp(str, "RAND") == 0) {
	return BA_SCHED_BEHAVIOR_RAND;
    } else if(strcasecmp(str, "RANDU") == 0) {
	return BA_SCHED_BEHAVIOR_RANDU;
    } else if(strcasecmp(str, "NOOP") == 0) {
	return BA_SCHED_BEHAVIOR_NOOP;
    } else {
	return BA_SCHED_BEHAVIOR_UNKNOWN;
    }
}

extern void
ba_sched_behavior_print_short_to(ba_sched_behavior_t behavior,
				 ba_print_func_t *print_func)
{
    BA_ASSERT(print_func);

    switch(behavior) {
    case BA_SCHED_BEHAVIOR_FCFS:
	print_func("FCFS (first-come-first-serve)");
	break;
    case BA_SCHED_BEHAVIOR_RFCFS:
	print_func("RFCFS (request-first-come-first-serve)");
	break;
    case BA_SCHED_BEHAVIOR_PS:
	print_func("PS (processor sharing)");
	break;
    case BA_SCHED_BEHAVIOR_UFB:
	print_func("UFB (user-FB)");
	break;
    case BA_SCHED_BEHAVIOR_URFB:
	print_func("URFB (user-requested-FB)");
	break;
    case BA_SCHED_BEHAVIOR_DUFB:
	print_func("DUFB (decayed user-FB)");
	break;
    case BA_SCHED_BEHAVIOR_GSVSD:
	print_func("GSVSD (greatest visible slowdown)");
	break;
    case BA_SCHED_BEHAVIOR_SRPT:
	print_func("SRPT (shortest-remaining-processing-time)");
	break;
    case BA_SCHED_BEHAVIOR_RSRPT:
	print_func("RSRPT (req-shortest-remaining-processing-time)");
	break;
    case BA_SCHED_BEHAVIOR_HRP:
	print_func("HRP (highest request probability)");
	break;
    case BA_SCHED_BEHAVIOR_HRR:
	print_func("HRR (highest requested resources)");
	break;
    case BA_SCHED_BEHAVIOR_EDF:
	print_func("EDF (earliest-deadline-first)");
	break;
    case BA_SCHED_BEHAVIOR_LPRT:
	print_func("LPRT (least probability-scaled remaining time)");
	break;
    case BA_SCHED_BEHAVIOR_LPRTP:
	print_func("LPRTP (least probability-scaled remaining time w/ "
		   "postponement)");
	break;
    case BA_SCHED_BEHAVIOR_RAND:
	print_func("RAND (random)");
	break;
    case BA_SCHED_BEHAVIOR_RANDU:
	print_func("RANDU (random user)");
	break;
    case BA_SCHED_BEHAVIOR_NOOP:
	print_func("NOOP (do nothing)");
	break;
    case BA_SCHED_BEHAVIOR_UNKNOWN:
	BA_PANIC();
	break;
    }
}

extern void
ba_sched_behavior_print_short(ba_sched_behavior_t behavior)
{
    ba_sched_behavior_print_short_to(behavior, ba_print_info);
}

static void
set_sched_func(ba_sched_behavior_t sched_behavior,
	       sched_func_t **sched_func)
{
    switch(sched_behavior) {
    case BA_SCHED_BEHAVIOR_FCFS:
	*sched_func = fcfs;
	break;
    case BA_SCHED_BEHAVIOR_RFCFS:
	*sched_func = rfcfs;
	break;
    case BA_SCHED_BEHAVIOR_PS:
	BA_PANIC();
	break;
    case BA_SCHED_BEHAVIOR_UFB:
	*sched_func = ufb;
	break;
    case BA_SCHED_BEHAVIOR_URFB:
	*sched_func = urfb;
	break;
    case BA_SCHED_BEHAVIOR_DUFB:
	BA_PANIC();
	break;
    case BA_SCHED_BEHAVIOR_GSVSD:
	*sched_func = gsvsd;
	break;
    case BA_SCHED_BEHAVIOR_SRPT:
	*sched_func = srpt;
	break;
    case BA_SCHED_BEHAVIOR_RSRPT:
	*sched_func = rsrpt;
	break;
    case BA_SCHED_BEHAVIOR_HRP:
	*sched_func = hrp;
	break;
    case BA_SCHED_BEHAVIOR_HRR:
	*sched_func = hrr;
	break;
    case BA_SCHED_BEHAVIOR_EDF:
	BA_PANIC();
	break;
    case BA_SCHED_BEHAVIOR_LPRT:
	BA_PANIC();
	*sched_func = lprt;
	break;
    case BA_SCHED_BEHAVIOR_LPRTP:
	BA_PANIC();
	*sched_func = lprtp;
	break;
    case BA_SCHED_BEHAVIOR_RAND:
	*sched_func = rand_sched;
	break;
    case BA_SCHED_BEHAVIOR_RANDU:
	*sched_func = randu;
	break;
    case BA_SCHED_BEHAVIOR_NOOP:
	*sched_func = noop;
	break;
    case BA_SCHED_BEHAVIOR_UNKNOWN:
	BA_PANIC();
	break;
    }
}

extern ba_status
ba_sched_init(void)
{
    sched_func_stats = ba_stats_new();
    sched_func_timer = g_timer_new();

    treat_req_dis_diff = ba_options_treat_req_dis_diff_get();
    requested_sched_behavior = ba_options_req_sched_behavior_get();
    disclosed_sched_behavior = ba_options_dis_sched_behavior_get();

    all_candidates = 0;
    requested_candidates = 0;
    disclosed_candidates = 0;

    all_candidates_app = ba_glib_fast_slist_append_new();
    requested_candidates_app = ba_glib_fast_slist_append_new();
    disclosed_candidates_app = ba_glib_fast_slist_append_new();

    num_all_candidates = 0;
    num_req_candidates = 0;
    num_dis_candidates = 0;

    weighed_all_candidates = 0;
    weighed_req_candidates = 0;
    weighed_dis_candidates = 0;

    all_sched_func = 0;
    requested_sched_func = 0;
    disclosed_sched_func = 0;

    if(treat_req_dis_diff == BA_FALSE) {
	set_sched_func(requested_sched_behavior, &all_sched_func);
    } else {
	set_sched_func(requested_sched_behavior, &requested_sched_func);
	set_sched_func(disclosed_sched_behavior, &disclosed_sched_func);
    }

    rand_sched_dist = ba_rv_dist_new("uniform", 0.0, 1.0, 0.0, 0.0);
    randu_sched_dist = ba_rv_dist_new("uniform", 0.0, 1.0, 0.0, 0.0);

    return BA_STATUS_SUCCESS;
}

extern ba_status
ba_sched_destroy(void)
{
#if COUNT_SCHEDULER_DECISIONS
    ba_print_info("sched_func_stats:\n");
    ba_stats_print(sched_func_stats);
#endif
    ba_stats_delete(sched_func_stats);
    g_timer_destroy(sched_func_timer);

    g_slist_free(all_candidates);
    g_slist_free(requested_candidates);
    g_slist_free(disclosed_candidates);

    ba_glib_fast_slist_append_delete(all_candidates_app);
    ba_glib_fast_slist_append_delete(requested_candidates_app);
    ba_glib_fast_slist_append_delete(disclosed_candidates_app);

    ba_rv_dist_delete(rand_sched_dist);
    ba_rv_dist_delete(randu_sched_dist);

    return BA_STATUS_SUCCESS;
}

extern void
ba_sched_destroy_from_hook(void *unused)
{
    ba_status status;

    status = ba_sched_destroy();
    BA_ASSERT(status == BA_STATUS_SUCCESS);
}
