#include <ba_config.h>
#include <ba_sim.h>

#define USE_QUANT 0

struct ba_stats_s {
    GArray *values;

    ba_bool at_least_one_val;

    // valid after processing...
    uint64_t count;

    ba_real total;

    ba_real min;
    ba_real max;

    ba_real sample_mean;
    ba_real sample_variance;

    ba_real standard_deviation;
    ba_real standard_error;

    ba_real conf_delta_90;
    ba_real conf_delta_95;
    ba_real conf_delta_99;

    ba_real quant_25;
    ba_real quant_75;
    
    //
    ba_bool reprocess;
};

// this running mean is used when we wish to estimate the mean of a
// function that changes over time and when samples of the function
// are taken at irregular points in time.  this structure does not
// remember old samples, so it take little memory.
struct ba_stats_running_mean_s {
    uint64_t count;

    ba_real sample_mean;
    ba_real abs_time;
};

static void process(ba_stats_t *stats);

#if USE_QUANT
static void compute_quantiles(ba_stats_t *stats);
static gint compare_func(gconstpointer a_in,
			 gconstpointer b_in);
static ba_real get_quantile(ba_stats_t *stats,
			    ba_real percent);
#endif

extern ba_bool
ba_stats_equal(ba_stats_t *stats1,
	       ba_stats_t *stats2)
{
    // this func ensures two stats are the same (essentially) even if
    // the insertations of their elements was in different orders.
    if(stats1->reprocess) {
	stats1->reprocess = BA_FALSE;
	process(stats1);
    }
    if(stats2->reprocess) {
	stats2->reprocess = BA_FALSE;
	process(stats2);
    }

    if(stats1->count != stats2->count) {
	return BA_FALSE;
    }

    if(!ba_types_real_equal(stats1->total, stats2->total)) {
	return BA_FALSE;
    }
    if(!ba_types_real_equal(stats1->min, stats2->min)) {
	return BA_FALSE;
    }
    if(!ba_types_real_equal(stats1->max, stats2->max)) {
	return BA_FALSE;
    }
    if(!ba_types_real_equal(stats1->sample_mean, stats2->sample_mean)) {
	return BA_FALSE;
    }
    if(!ba_types_real_equal(stats1->sample_variance,
			    stats2->sample_variance)) {
	return BA_FALSE;
    }
    
    return BA_TRUE;
}

extern ba_bool
ba_stats_non_empty(ba_stats_t *stats)
{
    BA_ASSERT(stats);

    return stats->at_least_one_val;
}

extern ba_real
ba_stats_count_get(ba_stats_t *stats)
{
    BA_ASSERT(stats);

    if(stats->reprocess) {
	stats->reprocess = BA_FALSE;
	process(stats);
    }

    return (ba_real)stats->count;
}

extern ba_real
ba_stats_total_get(ba_stats_t *stats)
{
    BA_ASSERT(stats);

    if(stats->reprocess) {
	stats->reprocess = BA_FALSE;
	process(stats);
    }

    return stats->total;
}

extern ba_real
ba_stats_min_get(ba_stats_t *stats)
{
    BA_ASSERT(stats);

    if(stats->reprocess) {
	stats->reprocess = BA_FALSE;
	process(stats);
    }

    return stats->min;
}

extern ba_real
ba_stats_max_get(ba_stats_t *stats)
{
    BA_ASSERT(stats);

    if(stats->reprocess) {
	stats->reprocess = BA_FALSE;
	process(stats);
    }

    return stats->max;
}

extern ba_real
ba_stats_sample_mean_get(ba_stats_t *stats)
{
    BA_ASSERT(stats);

    if(stats->reprocess) {
	stats->reprocess = BA_FALSE;
	process(stats);
    }

    return stats->sample_mean;
}

extern ba_real
ba_stats_sample_variance_get(ba_stats_t *stats)
{
    BA_ASSERT(stats);

    if(stats->reprocess) {
	stats->reprocess = BA_FALSE;
	process(stats);
    }

    return stats->sample_variance;
}

extern ba_real
ba_stats_standard_deviation_get(ba_stats_t *stats)
{
    BA_ASSERT(stats);

    if(stats->reprocess) {
	stats->reprocess = BA_FALSE;
	process(stats);
    }

    return stats->standard_deviation;
}

extern void
ba_stats_insert(ba_stats_t *stats,
		ba_real value)
{
    BA_ASSERT(stats);

    BA_ASSERT(finite(value));
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(value, 0.0);

    stats->values = g_array_append_val(stats->values, value);
    stats->reprocess = BA_TRUE;

    stats->at_least_one_val = BA_TRUE;
}

static void
process(ba_stats_t *stats)
{
    BA_ASSERT(stats);
    uint64_t i;
    ba_real value;
    ba_bool first;
    ba_real deviance, total_squared_deviance;

    stats->count = stats->values->len;

    BA_ASSERT(stats->count >= 1);

    first = BA_TRUE;
    for(i = 0; i < stats->count; i++) {
	value = g_array_index(stats->values, ba_real, i);

	stats->total += value;

	if(first) {
	    first = BA_FALSE;
	    stats->min = value;
	    stats->max = value;
	} else {
	    stats->min = BA_REAL_MIN(value, stats->min);
	    stats->max = BA_REAL_MAX(value, stats->max);
	}
    }

    stats->sample_mean = stats->total / (ba_real)stats->count;

    total_squared_deviance = 0.0;
    for(i = 0; i < stats->count; i++) {
	value = g_array_index(stats->values, ba_real, i);

	deviance = value - stats->sample_mean;
	total_squared_deviance += deviance * deviance;
    }

    if(stats->count > 1) {
	stats->sample_variance = total_squared_deviance /
	    (ba_real)(stats->count - 1);
    } else {
	stats->sample_variance = 0;
    }

    stats->standard_deviation = sqrt(stats->sample_variance);
    stats->standard_error = stats->standard_deviation /
	sqrt((ba_real)stats->count);

    stats->conf_delta_90 = 1.65 * stats->standard_error;
    stats->conf_delta_95 = 1.96 * stats->standard_error;
    stats->conf_delta_99 = 2.48 * stats->standard_error;

#if USE_QUANT
    compute_quantiles(stats);
#endif

    stats->reprocess = BA_FALSE;
}

#if USE_QUANT
static void
compute_quantiles(ba_stats_t *stats)
{
    g_array_sort(stats->values, compare_func);

    stats->quant_25 = get_quantile(stats, 0.25);
    stats->quant_75 = get_quantile(stats, 0.75);
}

static ba_real
get_quantile(ba_stats_t *stats,
	     ba_real percent)
{
    ba_real real_index;
    uint32_t int_index;
    ba_real quant;

    real_index = ((ba_real)stats->count - 1.0) * percent;
    int_index = rint(real_index); // XXX is this rint() correct?
    BA_ASSERT(int_index >= 0 && int_index <= (stats->count - 1));
    quant = g_array_index(stats->values, ba_real, int_index);

    return quant;
}
#endif

#if USE_QUANT
static gint
compare_func(gconstpointer a_in,
	     gconstpointer b_in)
{
    ba_real *a, *b;

    a = (ba_real *)a_in;
    b = (ba_real *)b_in;

    if(*a < *b) {
	return -1;
    } else if(*a == *b) {
	return 0;
    } else {
	return 1;
    }
}
#endif

extern void
ba_stats_print_nice_to(ba_stats_t *stats,
		       ba_print_func_t *print_func)
{
    BA_ASSERT(stats);
    BA_ASSERT(print_func);

    if(stats->reprocess) {
	stats->reprocess = BA_FALSE;
	process(stats);
    }

    print_func(BA_REAL_LOW_PREC_FMT ", ", stats->sample_mean);
    print_func(BA_REAL_LOW_PREC_FMT ", ", stats->total);
    print_func(BA_REAL_LOW_PREC_FMT "-" BA_REAL_LOW_PREC_FMT ", ", stats->min,
	       stats->max);
#if USE_QUANT
    print_func(BA_REAL_LOW_PREC_FMT "-" BA_REAL_LOW_PREC_FMT ", ",
	       stats->quant_25, stats->quant_75);
#endif
    print_func(BA_REAL_LOW_PREC_FMT ", ", stats->standard_deviation);

    //    print_func(BA_REAL_LOW_PREC_FMT, stats->count); why does
    //    this not work?? XXX

    print_func("%d", stats->values->len);
}

extern void
ba_stats_print_nice_maybe_none_to(ba_stats_t *stats,
				  ba_print_func_t *print_func)
{
    BA_ASSERT(stats);
    BA_ASSERT(print_func);

    if(ba_options_virtual_clock_terminate_get() == 0.0) {
	ba_stats_print_nice_to(stats, print_func);
	return;
    }

    if(ba_stats_non_empty(stats)) {
	ba_stats_print_nice_to(stats, print_func);
    } else {
	ba_real val = 0.0;

	print_func(BA_REAL_LOW_PREC_FMT ", ", val);
	print_func(BA_REAL_LOW_PREC_FMT ", ", val);
	print_func(BA_REAL_LOW_PREC_FMT "-" BA_REAL_LOW_PREC_FMT ", ", val,
		   val);
#if USE_QUANT
	print_func(BA_REAL_LOW_PREC_FMT "-" BA_REAL_LOW_PREC_FMT ", ",
		   val, val);
#endif
	print_func(BA_REAL_LOW_PREC_FMT ", ", val);
	print_func(BA_REAL_LOW_PREC_FMT, val);
    }
}

extern void
ba_stats_print_nice(ba_stats_t *stats)
{
    BA_ASSERT(stats);

    ba_stats_print_nice_to(stats, ba_print_info);
}

extern void
ba_stats_print(ba_stats_t *stats)
{
    BA_ASSERT(stats);

    if(stats->reprocess) {
	stats->reprocess = BA_FALSE;
	process(stats);
    }

    ba_print_info("stats %p:\n", stats);
    ba_print_info("  count: %qd\n", stats->count);    

    ba_print_info("  total: " BA_REAL_FMT "\n", stats->total);

    ba_print_info("  min: " BA_REAL_FMT "\n", stats->min);
    ba_print_info("  max: " BA_REAL_FMT "\n", stats->max);

    ba_print_info("  sample_mean: " BA_REAL_FMT "\n", stats->sample_mean);
    ba_print_info("  sample_variance: " BA_REAL_FMT "\n",
		  stats->sample_variance);

    ba_print_info("  standard_deviation: " BA_REAL_FMT "\n",
		  stats->standard_deviation);
    ba_print_info("  standard_error: " BA_REAL_FMT "\n",
		  stats->standard_error);

    ba_print_info("  conf_delta_90: " BA_REAL_FMT "\n", stats->conf_delta_90);
    ba_print_info("  conf_delta_95: " BA_REAL_FMT "\n", stats->conf_delta_95);
    ba_print_info("  conf_delta_99: " BA_REAL_FMT "\n", stats->conf_delta_99);

#if USE_QUANT
    ba_print_info("  quant_25: " BA_REAL_FMT "\n", stats->quant_25);
    ba_print_info("  quant_75: " BA_REAL_FMT "\n", stats->quant_75);
#endif

    ba_print_info("  reprocess: ");
    ba_types_bool_print(stats->reprocess);
    ba_print_info("\n");
}

extern ba_stats_t *
ba_stats_new(void)
{
    ba_stats_t *stats;

    stats = ba_mem_malloc(sizeof(ba_stats_t));

    stats->values = g_array_new(FALSE, FALSE, sizeof(ba_real));

    stats->at_least_one_val = BA_FALSE;

    stats->count = 0;

    stats->total = 0.0;

    stats->min = 0.0;
    stats->max = 0.0;

    stats->sample_mean = 0.0;
    stats->sample_variance = 0.0;

    stats->standard_deviation = 0.0;
    stats->standard_error = 0.0;
    
    stats->conf_delta_90 = 0.0;
    stats->conf_delta_95 = 0.0;
    stats->conf_delta_99 = 0.0;

    stats->quant_25 = 0.0;
    stats->quant_75 = 0.0;
    
    stats->reprocess = BA_TRUE;

    return stats;
}

extern void
ba_stats_delete(ba_stats_t *stats)
{
    BA_ASSERT(stats);

    g_array_free(stats->values, TRUE);

    ba_mem_free(stats);
}

extern void
ba_stats_delete_from_list(void *stats_in,
			  void *unused)
{
    BA_ASSERT(stats_in);
    ba_stats_t *stats;

    stats = (ba_stats_t *)stats_in;

    ba_stats_delete(stats);
}

extern void
ba_stats_running_mean_print(ba_stats_running_mean_t *rm_stats)
{
    BA_ASSERT(rm_stats);

    ba_print_info("running stats %p:\n", rm_stats);
    ba_print_info("  count: %qd\n", rm_stats->count);
    ba_print_info("  sample_mean: " BA_REAL_FMT "\n", rm_stats->sample_mean);
    ba_print_info("  abs_time: " BA_REAL_FMT "\n", rm_stats->abs_time);
}

extern ba_real
ba_stats_running_mean_get(ba_stats_running_mean_t *rm_stats)
{
    BA_ASSERT(rm_stats);

    BA_ASSERT(rm_stats->count >= 1);

    return rm_stats->sample_mean;
}

extern void
ba_stats_running_mean_insert(ba_stats_running_mean_t *rm_stats,
			    ba_real value,
			    ba_real abs_time)
{
    BA_ASSERT(rm_stats);
    ba_real last_sample_mean;
    ba_real last_abs_time;

    if(rm_stats->count == 0) {
	rm_stats->count++;
	rm_stats->sample_mean = value;
	rm_stats->abs_time = abs_time;
	return;
    }

    last_sample_mean = rm_stats->sample_mean;
    last_abs_time = rm_stats->abs_time;
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(abs_time, last_abs_time);

    rm_stats->count++;
    rm_stats->sample_mean = ((last_sample_mean * last_abs_time) +
		     value * (abs_time - last_abs_time))
	/ abs_time;
    rm_stats->abs_time = abs_time;
}

extern ba_stats_running_mean_t *
ba_stats_running_mean_new(void)
{
    ba_stats_running_mean_t *rm_stats;

    rm_stats = ba_mem_malloc(sizeof(ba_stats_running_mean_t));

    rm_stats->count = 0;
    rm_stats->sample_mean = 0.0;
    rm_stats->abs_time = 0.0;

    return rm_stats;
}

extern void
ba_stats_running_mean_delete(ba_stats_running_mean_t *rm_stats)
{
    BA_ASSERT(rm_stats);

    ba_mem_free(rm_stats);
}

extern void
ba_stats_running_mean_delete_from_list(void *rm_stats_in,
				      void *unused)
{
    BA_ASSERT(rm_stats_in);
    ba_stats_running_mean_t *rm_stats;

    rm_stats = (ba_stats_running_mean_t *)rm_stats_in;

    ba_stats_running_mean_delete(rm_stats);
}
