#include <ba_config.h>
#include <ba_sim.h>
#include <string.h>
#include <stdio.h>

typedef struct {
    ba_real delta_time;
} new_user_entry_t;

typedef struct {
    ba_trace_entry_t type;
    ba_real delta_time;
    ba_id_t task_id;
} user_entry_t;

typedef struct {
    uint32_t num_users;
    uint32_t max_req_dis_tasks_per_user;
} traced_user_behavior_t;

GSList *new_user_entries;
ba_glib_fast_slist_append_t *new_user_entries_appender;

GArray *users_array;
GArray *users_appender_array;

ba_bool read_trace_input = BA_FALSE;
ba_bool write_trace_output = BA_FALSE;

traced_user_behavior_t *traced_user_behavior;

static ba_trace_entry_t str_to_trace_entry(char *str);
static char *entry_type_string(ba_trace_entry_t type);
static void write_entry(ba_trace_entry_t type,
			ba_id_t user_id,
			ba_id_t task_id,
			ba_real rv);
static ba_real traverse_trace(ba_trace_entry_t type,
			      ba_id_t user_id,
			      ba_id_t task_id);
static ba_real traverse_new_user_entry(void);
static ba_real traverse_user_entry(ba_trace_entry_t type,
					ba_id_t user_id,
					ba_id_t task_id);
static uint32_t get_max_trace_entry_string_len(void);
static void write_header(void);
static void write_parsable_user_behavior(void);
static void read_header(void);
static void read_parsable_user_behavior(void);
static void read_entries(void);
static void free_appender_structs(void);
static void parse_entry(char *str,
			ba_id_t user_id,
			ba_id_t task_id,
			ba_real rv);
static void parse_new_user_entry(ba_real rv);
static void parse_user_entry(ba_trace_entry_t type,
				  ba_id_t user_id,
				  ba_id_t task_id,
				  ba_real rv);
static void expand_arrays(ba_id_t user_id);
static void read_trace(void);
static new_user_entry_t *new_user_entry_new(ba_real delta_time);
static void new_user_entry_delete(new_user_entry_t
				       *new_user_entry);
static user_entry_t *user_entry_new(ba_trace_entry_t type,
					      ba_real delta_time,
					      ba_id_t task_id);
static void user_entry_delete(user_entry_t *user_entry);

extern ba_bool
ba_trace_read_trace_input_get(void)
{
    return read_trace_input;
}

extern ba_bool
ba_trace_write_trace_output_get(void)
{
    return write_trace_output;
}

extern uint32_t
ba_trace_num_users_get(void)
{
    // would be stricter to assert that we actually got far enough
    // along to read the trace...
    BA_ASSERT(read_trace_input);

    return traced_user_behavior->num_users;
}

extern uint32_t
ba_trace_max_req_dis_tasks_per_user_get(void)
{
    // would be stricter to assert that we actually got far enough
    // along to read the trace...
    BA_ASSERT(read_trace_input);

    return traced_user_behavior->max_req_dis_tasks_per_user;
}

static ba_trace_entry_t
str_to_trace_entry(char *str)
{
    if(strcmp(str, "new_user") == 0) {
	return BA_TRACE_ENTRY_NEW_USER;
    } else if(strcmp(str, "task_set_change_prob") == 0) {
	return BA_TRACE_ENTRY_TASK_SET_CHANGE_PROB;
    } else if(strcmp(str, "tasks_per_task_set") == 0) {
	return BA_TRACE_ENTRY_TASKS_PER_TASK_SET;
    } else if(strcmp(str, "task_set_change") == 0) {
	return BA_TRACE_ENTRY_TASK_SET_CHANGE;
    } else if(strcmp(str, "service_delta_time") == 0) {
	return BA_TRACE_ENTRY_SERVICE_DELTA_TIME;
    } else if(strcmp(str, "think_delta_time") == 0) {
	return BA_TRACE_ENTRY_THINK_DELTA_TIME;
    } else {
	BA_PANIC();
    }

    return 0; // warning elimination
}

static char *
entry_type_string(ba_trace_entry_t type)
{
#define KASE(_v_,_s_) case _v_: return(_s_)
    
    switch(type) {
	KASE(BA_TRACE_ENTRY_NEW_USER, "new_user");
	KASE(BA_TRACE_ENTRY_TASK_SET_CHANGE_PROB, "task_set_change_prob");
	KASE(BA_TRACE_ENTRY_TASKS_PER_TASK_SET, "tasks_per_task_set");
	KASE(BA_TRACE_ENTRY_TASK_SET_CHANGE, "task_set_change");
	KASE(BA_TRACE_ENTRY_SERVICE_DELTA_TIME, "service_delta_time");
	KASE(BA_TRACE_ENTRY_THINK_DELTA_TIME, "think_delta_time");
	
    default:
	BA_PANIC();
    }

#undef KASE

    return 0; // warning elimination
}

static void
write_entry(ba_trace_entry_t type,
	    ba_id_t user_id,
	    ba_id_t task_id,
	    ba_real rv)
{
    // not bothering to save space by omitting the new_user id.
    // doing so would complicate parsing.
    ba_print_write_trace_output("%s " BA_ID_FMT " " BA_ID_FMT " "
				BA_REAL_HIGH_PREC_FMT "\n",
				entry_type_string(type), user_id, task_id,
				rv);
}

extern ba_real
ba_trace_get_and_log_rv(ba_trace_entry_t type,
			ba_id_t user_id,
			ba_id_t task_id,
			ba_trace_gen_rv_t gen_rv_func,
			void *gen_rv_data)
{
    ba_real rv;

    if(read_trace_input) {
	rv = traverse_trace(type, user_id, task_id);
    } else {
	rv = gen_rv_func(gen_rv_data);
    }

    if(write_trace_output) {
	write_entry(type, user_id, task_id, rv);
    }

    return rv;
}

static ba_real
traverse_trace(ba_trace_entry_t type,
	       ba_id_t user_id,
	       ba_id_t task_id)
{
    ba_real rv;

    rv = 0.0; // warning elimination

    switch(type) {
    case BA_TRACE_ENTRY_NEW_USER:
	rv = traverse_new_user_entry();
	break;

    case BA_TRACE_ENTRY_TASK_SET_CHANGE_PROB:
    case BA_TRACE_ENTRY_TASKS_PER_TASK_SET:
    case BA_TRACE_ENTRY_TASK_SET_CHANGE:
    case BA_TRACE_ENTRY_SERVICE_DELTA_TIME:
    case BA_TRACE_ENTRY_THINK_DELTA_TIME:
	rv = traverse_user_entry(type, user_id, task_id);
	break;

    default:
	BA_PANIC();
	break;
    }

    return rv;
}

static ba_real
traverse_new_user_entry(void)
{
    new_user_entry_t *new_user_entry;
    ba_real rv;

    BA_ASSERT(new_user_entries);

    new_user_entry = (new_user_entry_t *)new_user_entries->data;
    BA_ASSERT(new_user_entry);

    rv = new_user_entry->delta_time;

    new_user_entries =
	g_slist_remove(new_user_entries, new_user_entry);
    new_user_entry_delete(new_user_entry);

    return rv;
}

static ba_real
traverse_user_entry(ba_trace_entry_t type,
			 ba_id_t user_id,
			 ba_id_t task_id)
{
    user_entry_t *user_entry;
    ba_real rv;
    GSList **user_entries;
    ba_bool found;
    GSList *elem;

    BA_ASSERT(user_id < users_array->len);

    user_entry = 0; // warning elimination

    user_entries = &g_array_index(users_array, GSList *,
				       user_id);
    if(*user_entries == 0) {
	ba_print_info("wanted type %d, user %d, task %d.\n",
		      type, user_id, task_id);
    }
    BA_ASSERT(*user_entries);

    found = BA_FALSE;
    // usually, what i want is near the top of the list...
    for(elem = *user_entries; elem; elem = g_slist_next(elem)) {
	user_entry = (user_entry_t *)elem->data;
	BA_ASSERT(user_entry);

	if(user_entry->type != type) {
	    continue;
	}
	if(user_entry->task_id != task_id) {
	    continue;
	}

	found = BA_TRUE;
	break;
    }

    BA_ASSERT(found);

    rv = user_entry->delta_time;

    *user_entries = g_slist_remove(*user_entries, user_entry);
    user_entry_delete(user_entry);

    return rv;
}

static uint32_t
get_max_trace_entry_string_len(void)
{
    uint32_t i;
    uint32_t len;
    ba_bool first;

    first = BA_TRUE;
    len = 0; // appease the compiler.

    for(i = 0; i < BA_TRACE_ENTRY_UNKNOWN; i++) {
	if(first) {
	    first = BA_FALSE;
	    len = strlen(entry_type_string(i));
	} else {
	    if(strlen(entry_type_string(i)) > len) {
		len = strlen(entry_type_string(i));
	    }
	}
    }

    return len;
}

static void
write_header(void)
{
    ba_report_print_proem_to(ba_print_write_trace_output);
    ba_report_print_user_behavior_to(ba_print_write_trace_output);
}

static void
write_parsable_user_behavior(void)
{
    ba_print_write_trace_output("---- parsable user behavior ----\n");
    ba_print_write_trace_output("num_users %d\n",
				ba_options_num_users_get());
    ba_print_write_trace_output("max_req_dis_tasks_per_user %d\n",
	       ba_options_max_req_dis_tasks_per_user_get());
    ba_print_write_trace_output("\n");
    ba_print_write_trace_output("---- trace ----\n");
}

static void
read_header(void)
{
    ba_bool eof;
    char *line_str;
    ba_bool read_header;

    // 80 is hardcoded to avoid awesomely complex macro expansion stuff.
    line_str = ba_string_new_len("", 80);

    read_header = BA_FALSE;
    while(1) {
	// no easy way to get a string up to a certain number of
	// characters including whitespace using a scanf variant.

	eof = ba_print_read_trace_input_line(line_str, 80, BA_TRUE);
	if(eof) {
	    break;
	}

	if(strcmp(line_str, "---- parsable user behavior ----\n") == 0) {
	    read_header = BA_TRUE;
	    break;
	}
    }
    BA_ASSERT(read_header == BA_TRUE);
    ba_string_delete(line_str);
}

static void
read_parsable_user_behavior(void)
{
    ba_bool eof;
    char *line_str;
    uint32_t num_conv;

    // 80 is hardcoded to avoid awesomely complex macro expansion stuff.
    line_str = ba_string_new_len("", 80);

    eof = ba_print_read_trace_input_line(line_str, 80, BA_FALSE);
    BA_ASSERT(eof == BA_FALSE);
    num_conv = sscanf(line_str, "num_users %d\n",
		      &traced_user_behavior->num_users);
    BA_ASSERT(num_conv == 1);

    eof = ba_print_read_trace_input_line(line_str, 80, BA_FALSE);
    BA_ASSERT(eof == BA_FALSE);
    num_conv = sscanf(line_str, "max_req_dis_tasks_per_user %d\n",
		      &traced_user_behavior->
		      max_req_dis_tasks_per_user);
    BA_ASSERT(num_conv == 1);
    
    eof = ba_print_read_trace_input_line(line_str, 80, BA_FALSE);
    BA_ASSERT(eof == BA_FALSE);
    BA_ASSERT(strcmp(line_str, "\n") == 0);

    eof = ba_print_read_trace_input_line(line_str, 80, BA_FALSE);
    BA_ASSERT(eof == BA_FALSE);
    BA_ASSERT(strcmp(line_str, "---- trace ----\n") == 0);

    ba_string_delete(line_str);
}

static void
read_entries(void)
{
    ba_bool eof;
    uint32_t num_conv;
    uint32_t trace_entry_type_len;
    char *trace_entry_type_str;
    ba_id_t user_id, task_id;
    ba_real rv;
    char *line_str;

    users_appender_array =
	g_array_new(FALSE, FALSE, sizeof(ba_glib_fast_slist_append_t *));

    trace_entry_type_len = get_max_trace_entry_string_len();
    trace_entry_type_str = ba_string_new_len("", trace_entry_type_len);

    line_str = ba_string_new_len("", 80);

    while(1) {
	eof = ba_print_read_trace_input_line(line_str, 80, BA_FALSE);
	if(eof) {
	    break;
	}
	
	num_conv = sscanf(line_str, "%s " BA_ID_FMT " " BA_ID_FMT " "
			  BA_REAL_SCAN_FMT "\n", trace_entry_type_str,
			  &user_id, &task_id, &rv);
	BA_ASSERT(num_conv == 4);	

	/*	ba_print_info("%s " BA_ID_FMT " " BA_ID_FMT " "
		BA_REAL_HIGH_PREC_FMT "\n",
		trace_entry_type_str, user_id, task_id, rv);*/

	parse_entry(trace_entry_type_str, user_id, task_id, rv);
    }

    ba_string_delete(line_str);
    ba_string_delete(trace_entry_type_str);

    free_appender_structs();
}

static void
free_appender_structs(void)
{
    uint32_t i;
    ba_glib_fast_slist_append_t *fast_slist_append;

    if(new_user_entries_appender) {
	ba_glib_fast_slist_append_delete(new_user_entries_appender);
    }
    new_user_entries_appender = 0;

    BA_ASSERT(users_array->len == users_appender_array->len);
    for(i = 0; i < users_appender_array->len; i++) {
	fast_slist_append = g_array_index(users_appender_array,
					  ba_glib_fast_slist_append_t *, i);
	ba_glib_fast_slist_append_delete(fast_slist_append);
    }
    g_array_free(users_appender_array, TRUE);
    users_appender_array = 0;
}

static void
parse_entry(char *str,
	    ba_id_t user_id,
	    ba_id_t task_id,
	    ba_real rv)
{
    ba_trace_entry_t type;

    type = str_to_trace_entry(str);
    switch(type) {
    case BA_TRACE_ENTRY_NEW_USER:
	parse_new_user_entry(rv);
	break;

    case BA_TRACE_ENTRY_TASK_SET_CHANGE_PROB:
    case BA_TRACE_ENTRY_TASKS_PER_TASK_SET:
    case BA_TRACE_ENTRY_TASK_SET_CHANGE:
    case BA_TRACE_ENTRY_SERVICE_DELTA_TIME:
    case BA_TRACE_ENTRY_THINK_DELTA_TIME:
	parse_user_entry(type, user_id, task_id, rv);
	break;

    default:
	BA_PANIC();
	break;
    }
}

static void
parse_new_user_entry(ba_real rv)
{
    new_user_entry_t *new_user_entry;

    if(new_user_entries_appender == 0) {
	new_user_entries_appender = ba_glib_fast_slist_append_new();
    }
    
    new_user_entry = new_user_entry_new(rv);
    new_user_entries =
	ba_glib_fast_slist_append(new_user_entries_appender,
				  new_user_entries, new_user_entry);
}

static void
parse_user_entry(ba_trace_entry_t type,
		      ba_id_t user_id,
		      ba_id_t task_id,
		      ba_real rv)
{
    GSList **user_entries;
    ba_glib_fast_slist_append_t *fast_slist_append;
    user_entry_t *user_entry;
    
    if(user_id >= users_array->len) {
	expand_arrays(user_id);
    }

    user_entries = &g_array_index(users_array, GSList *,
				       user_id);
    fast_slist_append = g_array_index(users_appender_array,
				      ba_glib_fast_slist_append_t *,
				      user_id);
    user_entry = user_entry_new(type, rv, task_id);
    *user_entries = ba_glib_fast_slist_append(fast_slist_append,
						   *user_entries,
						   user_entry);
}

static void
expand_arrays(ba_id_t user_id)
{
    GSList **user_entries;
    ba_glib_fast_slist_append_t **fast_slist_append;
    
    users_array = g_array_set_size(users_array, user_id + 1);
    user_entries = &g_array_index(users_array, GSList *,
				       user_id);
    *user_entries = 0;
    
    users_appender_array = g_array_set_size(users_appender_array,
						 user_id + 1);
    fast_slist_append = &g_array_index(users_appender_array,
				       ba_glib_fast_slist_append_t *,
				       user_id);
    *fast_slist_append = ba_glib_fast_slist_append_new();
}

static void
read_trace(void)
{
    read_header();
    read_parsable_user_behavior();
    read_entries();
}

static new_user_entry_t *
new_user_entry_new(ba_real delta_time)
{
    new_user_entry_t *new_user_entry;

    new_user_entry = ba_mem_malloc(sizeof(new_user_entry_t));
    new_user_entry->delta_time = delta_time;

    return new_user_entry;
}

static void
new_user_entry_delete(new_user_entry_t *new_user_entry)
{
    BA_ASSERT(new_user_entry);

    // nothing to do.

    ba_mem_free(new_user_entry);
}

static user_entry_t *
user_entry_new(ba_trace_entry_t type,
		    ba_real delta_time,
		    ba_id_t task_id)
{
    user_entry_t *user_entry;

    user_entry = ba_mem_malloc(sizeof(user_entry_t));
    user_entry->type = type;
    user_entry->delta_time = delta_time;
    user_entry->task_id = task_id;

    return user_entry;
}

static void
user_entry_delete(user_entry_t *user_entry)
{
    BA_ASSERT(user_entry);

    // nothing to do.

    ba_mem_free(user_entry);
}

extern ba_status
ba_trace_init(void)
{
    new_user_entries = 0;
    new_user_entries_appender = 0;

    users_array = g_array_new(FALSE, FALSE, sizeof(GSList *));
    users_appender_array = 0;

    if(ba_options_trace_output_filename_get()) {
	write_trace_output = BA_TRUE;
    }
    if(ba_options_trace_input_filename_get()) {
	read_trace_input = BA_TRUE;
    }

    traced_user_behavior =
	ba_mem_malloc(sizeof(traced_user_behavior_t));
    traced_user_behavior->num_users = 0;
    traced_user_behavior->max_req_dis_tasks_per_user = 0;

    if(write_trace_output) {
	write_header();
	write_parsable_user_behavior();
    }

    if(read_trace_input) {
	read_trace();
    }

    return BA_STATUS_SUCCESS;
}

extern ba_status
ba_trace_destroy(void)
{
    int i;
    GSList *user_entries;
    
    BA_ASSERT(new_user_entries == 0);
    g_slist_free(new_user_entries); // moot, since 0

    if(read_trace_input) {
	BA_ASSERT(users_array->len >= 1);
    }
    for(i = 0; i < users_array->len; i++) {
	user_entries = g_array_index(users_array, GSList *, i);
	BA_ASSERT(user_entries == 0);
	g_slist_free(user_entries); // moot, since 0
    }
    g_array_free(users_array, TRUE);
    users_array = 0;

    BA_ASSERT(new_user_entries_appender == 0);
    BA_ASSERT(users_appender_array == 0);

    ba_mem_free(traced_user_behavior);

    return BA_STATUS_SUCCESS;
}

extern void
ba_trace_destroy_from_hook(void *unused)
{
    ba_status status;

    status = ba_trace_destroy();
    BA_ASSERT(status == BA_STATUS_SUCCESS);
}
