#include <ba_config.h>
#include <ba_sim.h>

#define THOROUGH_CHECKS 0 // XXX ./configure option?

struct ba_user_s {
    ba_id_t id;

    ba_bool queued_for_destruction;
    ba_bool stop_running_tasks;
    ba_real stop_running_tasks_abs_time;

    GSList *disclosed_tasks;
    GSList *executed_tasks;
    GSList *requested_tasks;
    ba_glib_fast_slist_append_t *finished_tasks_app;
    GSList *finished_tasks;
    // disclosed, requested, or executed tasks that were canceled (the
    // task set was deactivated or the user was removed).
    GSList *canceled_tasks;

    ba_id_gen_t *task_set_id_gen;
    ba_glib_fast_slist_append_t *task_sets_app;
    GSList *task_sets;
    ba_task_set_t *active_task_set;

    uint32_t max_req_dis_tasks;

    ba_id_gen_t *task_id_gen;

    ba_rv_dist_t *task_set_change_range_dist;
    ba_real task_set_change_prob;
    ba_rv_dist_t *task_set_change_dist;
    
    ba_rv_dist_t *tasks_per_task_set_dist;
    uint32_t tasks_per_task_set;

    ba_rv_dist_t *service_delta_time_dist;
    ba_rv_dist_t *think_delta_time_dist;

    // stuff that's reported (and asserted)
    ba_real resource_total_usage;
    ba_real warmup_resource_total_usage;
    ba_real req_resource_total_usage;
    ba_real running_req_resource_total_usage; // not reported, just for policy
    // the following var is WRONG XXX.  instead use
    // correct_warmup_req_resource_total_usage.
    ba_real warmup_req_resource_total_usage;
    ba_real correct_warmup_req_resource_total_usage;
    ba_real warmup_dis_resource_total_usage;
    ba_real warmup_scaled_resource_total_usage;

    uint32_t num_task_sets;
    uint32_t num_tasks_disclosed;
    uint32_t num_tasks_disclosed_only;
    uint32_t num_tasks_executed;
    uint32_t num_tasks_requested;
    uint32_t num_tasks_finished;
    uint32_t num_tasks_canceled;
    ba_stats_t *service_delta_time_stats;
    ba_real mean_service_delta_time;
    ba_stats_t *think_delta_time_stats;
    ba_real mean_think_delta_time;

    ba_real total_vrt_delta_time;
    ba_real total_vsd;
    uint32_t deadlines_met;
};

static void add_resource_usage_warmup_helper(ba_user_t *user,
					     ba_real resource_usage,
					     ba_real start_abs_time,
					     ba_task_t *task);
static void add_resource_usage_warmup_requested_helper(ba_user_t *user,
						       ba_real resource_usage,
						       ba_real start_abs_time,
						       ba_real warmup_abs_time,
						       ba_task_t *task);
static void stop_running_tasks_abs_time_set(ba_user_t *user,
				    ba_real stop_running_tasks_abs_time);
static void activate_task_set_helper(ba_user_t *user,
				     ba_task_set_t *task_set);
static void deactivate_task_set_helper(ba_user_t *user);
static void deactivate_ts_task_counted_in_warmup_helper(ba_user_t *user,
							ba_task_t *task,
						ba_real completed_before_warmup,
						ba_real st_after_warmup);
static void remove_tasks_from_list_helper(ba_user_t *user,
					  GSList **tasks);
static void add_canceled_task(ba_user_t *user,
			      ba_task_t *task);
static void add_executed_or_requested_task_helper(ba_user_t *user,
						  GSList **add_tasks,
						  GSList **search_tasks,
						  ba_task_t *task);
static void add_finished_task(ba_user_t *user,
			      ba_task_t *task);
static void think_about_output(ba_user_t *user,
			       ba_task_t *task_to_think_about);
static ba_event_callback_t request_task_callback;
static void add_requested_task(ba_user_t *user,
			       ba_task_t *task);

static void disclosed_tasks_print(ba_user_t *user);
static void executed_tasks_print(ba_user_t *user);
static void requested_tasks_print(ba_user_t *user);
static void finished_tasks_print(ba_user_t *user);
static void canceled_tasks_print(ba_user_t *user);

static void report_info(ba_user_t *user);
static void assert_info(ba_user_t *user);

static gint find_user_request_task(gconstpointer event_in,
					gconstpointer given_user_in);
static gint find_user_activate_new_task_set(gconstpointer event_in,
						 gconstpointer
						 given_user_in);

static void make_task(ba_user_t *user,
		      ba_task_set_t *new_task_set,
		      uint32_t i,
		      ba_bool *no_more_requests);

static ba_trace_gen_rv_t gen_service_delta_time;
static ba_trace_gen_rv_t gen_think_delta_time;
static ba_trace_gen_rv_t gen_task_set_change;
static ba_trace_gen_rv_t gen_task_set_change_prob;
static ba_trace_gen_rv_t gen_tasks_per_task_set;

extern ba_real
ba_user_request_probability_get(ba_user_t *user)
{
    BA_ASSERT(user);
    ba_real request_probability;

    if(user->num_tasks_disclosed == 0.0) {
	BA_ASSERT(user->num_tasks_requested == 0);
	return 0.0;
    }

    BA_ASSERT(user->num_tasks_disclosed >= user->num_tasks_requested);
    BA_ASSERT(user->num_tasks_requested >= 0);
    
    request_probability = (ba_real) user->num_tasks_requested /
	(ba_real) user->num_tasks_disclosed;
    BA_ASSERT(request_probability >= 0.0);
    BA_ASSERT(request_probability <= 1.0);

    return request_probability;
}

extern ba_real
ba_user_task_set_change_prob_get(ba_user_t *user)
{
    BA_ASSERT(user);

    return user->task_set_change_prob;
}

extern GSList *
ba_user_requested_tasks_get(ba_user_t *user)
{
    BA_ASSERT(user);

    return user->requested_tasks;
}

extern GSList *
ba_user_disclosed_tasks_get(ba_user_t *user)
{
    BA_ASSERT(user);

    return user->disclosed_tasks;
}

extern ba_real
ba_user_total_vsd_get(ba_user_t *user)
{
    BA_ASSERT(user);
    ba_real current_total_vsd;

    // user->req_resource_total_usage reflects total requested
    // resources, even if they haven't run yet.  so this makes this
    // computation totally cool.  in any case, this function is only
    // used by the unneeded gsvsd.

    current_total_vsd = ba_user_total_vrt_delta_time_get(user) /
	user->req_resource_total_usage;

    return current_total_vsd;
}

extern ba_real
ba_user_total_vrt_delta_time_get(ba_user_t *user)
{
    BA_ASSERT(user);
    ba_real current_total_vrt_delta_time, last_recorded;
    ba_real additional;
    GSList *node;

    last_recorded = user->total_vrt_delta_time;

    // last recorded doesn't count things the user is currently
    // waiting for, so that's the purpose of the following loop...
    additional = 0.0;
    for(node = user->requested_tasks; node; node = g_slist_next(node)) {
	ba_real requested_abs_time, virtual_clock_abs_time;
	ba_task_t *task;

	task = (ba_task_t *)node->data;
	BA_ASSERT(task);

	BA_ASSERT(ba_task_executed_get(task) == BA_FALSE);
	BA_ASSERT(ba_task_requested_get(task) == BA_TRUE);
	
	virtual_clock_abs_time = ba_events_virtual_clock_abs_time_get();
	requested_abs_time = ba_task_requested_abs_time_get(task);
	BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(virtual_clock_abs_time -
					     requested_abs_time, 0.0);
	additional += virtual_clock_abs_time - requested_abs_time;
    }

    current_total_vrt_delta_time = last_recorded + additional;

    return current_total_vrt_delta_time;
}

extern ba_id_t
ba_user_id_get(ba_user_t *user)
{
    BA_ASSERT(user);

    return user->id;
}

extern ba_bool
ba_user_stop_running_tasks_get(ba_user_t *user)
{
    BA_ASSERT(user);

    return user->stop_running_tasks;
}

static void
stop_running_tasks_abs_time_set(ba_user_t *user,
				ba_real stop_running_tasks_abs_time)
{
    BA_ASSERT(user);
    ba_task_set_t *task_set;

    user->stop_running_tasks = BA_TRUE;

    task_set = user->active_task_set;
    if(ba_task_set_return_unexecuted_task(task_set)) {
	user->stop_running_tasks_abs_time = stop_running_tasks_abs_time;
    } else {
	user->stop_running_tasks_abs_time =
	    ba_task_set_last_task_executed_abs_time_get(task_set);
    }

    // gotta preempt the resource since it still may be running
    // something from us...  there's no worry that the scheduler will
    // pick us again during this preemption because we just queued
    // ourself for destruction.
    ba_resource_preempt();
}

extern ba_real
ba_user_stop_running_tasks_abs_time_get(ba_user_t *user)
{
    BA_ASSERT(user);

    BA_ASSERT(user->stop_running_tasks);
    return user->stop_running_tasks_abs_time;
}

extern ba_id_t
ba_user_task_id_next(ba_user_t *user)
{
    BA_ASSERT(user);

    return ba_id_gen_next(user->task_id_gen);
}

static void
disclosed_tasks_print(ba_user_t *user)
{
    BA_ASSERT(user);

    ba_print_info("User disclosed tasks:\n");
    g_slist_foreach(user->disclosed_tasks, ba_task_print_from_list, 0);
}

static void
executed_tasks_print(ba_user_t *user)
{
    BA_ASSERT(user);

    ba_print_info("User executed tasks:\n");
    g_slist_foreach(user->executed_tasks, ba_task_print_from_list, 0);
}

static void
requested_tasks_print(ba_user_t *user)
{
    BA_ASSERT(user);

    ba_print_info("User requested tasks:\n");
    g_slist_foreach(user->requested_tasks, ba_task_print_from_list, 0);
}

static void
finished_tasks_print(ba_user_t *user)
{
    BA_ASSERT(user);

    ba_print_info("User finished tasks %p:\n", user);
    g_slist_foreach(user->finished_tasks, ba_task_print_from_list, 0);
}

static void
canceled_tasks_print(ba_user_t *user)
{
    BA_ASSERT(user);

    ba_print_info("User canceled tasks %p:\n", user);
    g_slist_foreach(user->canceled_tasks, ba_task_print_from_list, 0);
}

extern ba_task_set_t *
ba_user_task_set_get(ba_user_t *user)
{
    BA_ASSERT(user);

    return user->active_task_set;
}

extern void
ba_user_print(ba_user_t *user)
{
    BA_ASSERT(user);

    ba_print_info("user: %p\n", user);
    ba_print_info("  id: " BA_ID_FMT "\n", user->id);
    ba_print_info("  queued_for_destruction: ");
    ba_types_bool_print(user->queued_for_destruction);
    ba_print_info("\n");
    ba_print_info("  stop_running_tasks: ");
    ba_types_bool_print(user->stop_running_tasks);
    ba_print_info("\n");
    ba_print_info("  stop_running_tasks_abs_time: " BA_REAL_FMT "\n",
		  user->stop_running_tasks_abs_time);
    disclosed_tasks_print(user);
    executed_tasks_print(user);
    requested_tasks_print(user);
    finished_tasks_print(user);
    canceled_tasks_print(user);
    ba_print_info("  task_set_id_gen: %p\n", user->task_set_id_gen);
    ba_print_info("  task_sets: %p\n", user->task_sets);
    g_slist_foreach(user->task_sets, ba_task_set_print_from_list, 0);
    ba_print_info("  active_task_set: %p\n", user->active_task_set);
    ba_print_info("  max_req_dis_tasks: %d\n", user->max_req_dis_tasks);
    ba_print_info("  task_id_gen: %p\n", user->task_id_gen);

    ba_print_info("  task_set_change_range_dist %p:\n",
		  user->task_set_change_range_dist);
    ba_rv_dist_print(user->task_set_change_range_dist);
    ba_print_info("  task_set_change_dist %p:\n",
		  user->task_set_change_dist);
    ba_rv_dist_print(user->task_set_change_dist);

    ba_print_info("  tasks_per_task_set_dist %p:\n",
		  user->tasks_per_task_set_dist);
    ba_rv_dist_print(user->tasks_per_task_set_dist);
    ba_print_info("  tasks_per_task_set: %d\n", user->tasks_per_task_set);

    ba_print_info("  service_delta_time_dist %p:\n",
		  user->service_delta_time_dist);
    ba_rv_dist_print(user->service_delta_time_dist);

    ba_print_info("  think_delta_time_dist %p:\n",
		  user->think_delta_time_dist);
    ba_rv_dist_print(user->think_delta_time_dist);

    ba_print_info("  resource_total_usage: " BA_REAL_FMT "\n",
		  user->resource_total_usage);
    ba_print_info("  warmup_resource_total_usage: " BA_REAL_FMT "\n",
		  user->warmup_resource_total_usage);
    ba_print_info("  req_resource_total_usage: " BA_REAL_FMT "\n",
		  user->req_resource_total_usage);
    ba_print_info("  running_req_resource_total_usage: " BA_REAL_FMT "\n",
		  user->running_req_resource_total_usage);
    ba_print_info("  warmup_req_resource_total_usage: " BA_REAL_FMT "\n",
		  user->warmup_req_resource_total_usage);
    ba_print_info("  correct_warmup_req_resource_total_usage: "
		  BA_REAL_FMT "\n",
		  user->correct_warmup_req_resource_total_usage);
    // warmup_dis_resource_total_usage only set in report_info().
    // warmup_scaled_resource_total_usage only set in report_info().

    ba_print_info("  num_task_sets: %d\n", user->num_task_sets);
    ba_print_info("  num_tasks_disclosed: %d\n",
		  user->num_tasks_disclosed);
    ba_print_info("  num_tasks_disclosed_only: %d\n",
		  user->num_tasks_disclosed_only);
    ba_print_info("  num_tasks_executed: %d\n", user->num_tasks_executed);
    ba_print_info("  num_tasks_requested: %d\n",
		  user->num_tasks_requested);
    ba_print_info("  num_tasks_finished: %d\n", user->num_tasks_finished);
    ba_print_info("  num_tasks_canceled: %d\n", user->num_tasks_canceled);
    ba_print_info("  service_delta_time_stats: ");
    ba_stats_print(user->service_delta_time_stats);
    ba_print_info("  think_delta_time_stats: ");
    ba_stats_print(user->think_delta_time_stats);

    ba_print_info("  total_vrt_delta_time: " BA_REAL_FMT "\n",
		  user->total_vrt_delta_time);
    // total_vsd only set in report_info().
    ba_print_info("  deadlines_met: %d\n", user->deadlines_met);
}

extern void
ba_user_print_from_list(void *user_in,
			     void *unused)
{
    BA_ASSERT(user_in);
    ba_user_t *user;
    
    user = (ba_user_t *)user_in;

    ba_user_print(user);
}

extern void
ba_user_activate_new_task_set(ba_event_kind_t kind,
			      ba_real delay_delta_time,
			      void *closure)
{
    BA_ASSERT(kind == BA_EVENT_KIND_USER_ACTIVATE_NEW_TASK_SET);
    BA_ASSERT(closure);
    ba_user_t *user;
    ba_id_t task_set_id;
    ba_task_set_t *new_task_set;
    uint32_t i;
    ba_bool no_more_requests;

    user = (ba_user_t *)closure;

    task_set_id = ba_id_gen_next(user->task_set_id_gen);
    new_task_set = ba_task_set_new(task_set_id, user);
    user->num_task_sets++;
    user->task_sets = ba_glib_fast_slist_append(user->task_sets_app,
						user->task_sets, new_task_set);

    no_more_requests = BA_FALSE;
    for(i = 0; i < user->tasks_per_task_set; i++) {
	make_task(user, new_task_set, i, &no_more_requests);
    }

#if BA_SIM_INSPECT
    ba_print_info("user %d planned a task set of %d task(s) at time "
		  BA_REAL_LOW_PREC_FMT "\n", ba_user_id_get(user),
		  user->tasks_per_task_set,
		  ba_events_virtual_clock_abs_time_get());
#endif

    deactivate_task_set_helper(user);
    activate_task_set_helper(user, new_task_set);

    // start first request
    think_about_output(user, 0);
}

static void
make_task(ba_user_t *user,
	  ba_task_set_t *new_task_set,
	  uint32_t i,
	  ba_bool *no_more_requests)
{
    ba_bool will_be_requested;
    ba_real task_set_change;
    ba_task_t *task;
    ba_id_t task_id;
    ba_real service_delta_time;
    ba_real think_delta_time;
    
    will_be_requested = BA_FALSE; // absurd warning elimination
    
    // more complex user modeling: possibly call
    // ba_task_clone_new_from_old to put a copy of a
    // canceled task back into operation.  but will be tricky with
    // task ids...
    
    task_id = ba_user_task_id_next(user);
    task = ba_task_new(task_id, new_task_set);
    
    service_delta_time =
	ba_trace_get_and_log_rv(BA_TRACE_ENTRY_SERVICE_DELTA_TIME,
				user->id, task_id, gen_service_delta_time,
				user);
    ba_task_service_delta_time_set(task, service_delta_time);
    
    think_delta_time =
	ba_trace_get_and_log_rv(BA_TRACE_ENTRY_THINK_DELTA_TIME,
				user->id, task_id, gen_think_delta_time, user);
    
    ba_task_think_delta_time_set(task, think_delta_time);
    
    if(i == 0) {
	// will always request the first task of a task set
	will_be_requested = BA_TRUE;
    } else if(*no_more_requests) {
	will_be_requested = BA_FALSE;
    } else {
	task_set_change =
	    ba_trace_get_and_log_rv(BA_TRACE_ENTRY_TASK_SET_CHANGE,
				    user->id, task_id, gen_task_set_change,
				    user);
	if(task_set_change == 1.0) {
	    *no_more_requests = BA_TRUE;
	    will_be_requested = BA_FALSE;
	} else if(task_set_change == 0.0) {
	    will_be_requested = BA_TRUE;
	} else {
	    BA_PANIC();
	}
    }
    ba_task_will_be_requested_set(task, will_be_requested);
    
    ba_task_set_append_task(new_task_set, task);
    
    ba_stats_insert(user->service_delta_time_stats, service_delta_time);
}

static ba_real
gen_service_delta_time(void *generic_data)
{
    BA_ASSERT(generic_data);
    ba_user_t *user;
    ba_real rv;

    user = (ba_user_t *)generic_data;

 retry:
    rv = ba_rv_dist(user->service_delta_time_dist);
    if(ba_types_real_equal(rv, 0.0)) {
	goto retry;
    }

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(rv, 0.0);

    return rv;
}

static ba_real
gen_think_delta_time(void *generic_data)
{
    BA_ASSERT(generic_data);
    ba_user_t *user;
    ba_real rv;

    // more complex modeling: make think time a function of what
    // the task to think about is?
    // would need to change the generic_data closure to be the task (and
    // then access the user through the task).

    user = (ba_user_t *)generic_data;

 retry:
    rv = ba_rv_dist(user->think_delta_time_dist);
    if(ba_types_real_equal(rv, 0.0)) {
	goto retry;
    }

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(rv, 0.0);

    return rv;
}

// only called in one place, but here to complement deactivate_task_set_helper
static void
activate_task_set_helper(ba_user_t *user,
			 ba_task_set_t *task_set)
{
    BA_ASSERT(user && task_set);
    ba_real virtual_clock_abs_time;

    virtual_clock_abs_time = ba_events_virtual_clock_abs_time_get();
    ba_task_set_activated_abs_time_set(task_set, virtual_clock_abs_time);

    user->active_task_set = task_set;

    ba_task_set_disclose_all(task_set);
}

extern void
ba_user_deactivate_task_set(ba_user_t *user)
{
    BA_ASSERT(user);
    BA_ASSERT(user->active_task_set);

    deactivate_task_set_helper(user);
}

static void
deactivate_task_set_helper(ba_user_t *user)
{
    BA_ASSERT(user);
    ba_real virtual_clock_abs_time;
    GSList *node;

    if(user->active_task_set == 0) {
	return;
    }

    virtual_clock_abs_time = ba_events_virtual_clock_abs_time_get();
    ba_task_set_deactivated_abs_time_set(user->active_task_set,
					 virtual_clock_abs_time);
    
    ba_task_set_report(user->active_task_set);
    
    ba_events_remove_events(user, find_user_request_task);
    ba_events_remove_events(user, find_user_activate_new_task_set);
    
    remove_tasks_from_list_helper(user, &user->disclosed_tasks);
    remove_tasks_from_list_helper(user, &user->executed_tasks);
    
    for(node = user->requested_tasks; node; node = g_slist_next(node)) {
	ba_task_t *task;
	ba_real remaining_delta_time;
	ba_real completed_before_warmup, st_after_warmup;
	
	task = (ba_task_t *)node->data;
	BA_ASSERT(task);
	
	remaining_delta_time = ba_task_remaining_delta_time_get(task);
	
	user->req_resource_total_usage -= remaining_delta_time;
	BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(user->req_resource_total_usage,
					     0.0);
	if(ba_task_counted_in_warmup_get(task, &completed_before_warmup,
					 &st_after_warmup)) {

	    deactivate_ts_task_counted_in_warmup_helper(user, task,
							completed_before_warmup,
							st_after_warmup);
	}

	// new code begin
	user->correct_warmup_req_resource_total_usage += 
	    ba_task_completed_warmup_delta_time_get(task);
	// new code end
    }

    remove_tasks_from_list_helper(user, &user->requested_tasks);

    user->active_task_set = 0;
}

static void
deactivate_ts_task_counted_in_warmup_helper(ba_user_t *user,
					    ba_task_t *task,
					    ba_real completed_before_warmup,
					    ba_real st_after_warmup)
{
    ba_real should_be_counted;
    ba_real adjustment;

    BA_ASSERT(ba_report_warmup_met());
	    
    should_be_counted = ba_task_completed_delta_time_get(task) -
	completed_before_warmup;
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(should_be_counted, 0.0);
    
    adjustment = st_after_warmup - should_be_counted;
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(adjustment, 0.0);
    
    // turns out, with some algebra, that `adjustment' is just
    // the current remaining time.
    BA_ASSERT_REAL_EQUAL(adjustment, ba_task_remaining_delta_time_get(task));
    
    user->warmup_req_resource_total_usage -= adjustment;
    
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(user->warmup_req_resource_total_usage,
					 0.0);
}

static gint
find_user_request_task(gconstpointer event_in,
		       gconstpointer given_user_in)
{
    BA_ASSERT(given_user_in && event_in);
    ba_user_t *given_user, *user;
    ba_event_t *event;

    given_user = (ba_user_t *)given_user_in;
    event = (ba_event_t *)event_in;

    if(ba_event_kind_get(event) != BA_EVENT_KIND_USER_REQUEST_TASK) {
	return 1;
    }

    user = (ba_user_t *)ba_event_closure_get(event);
    BA_ASSERT(user);

    if(given_user == user) {
	return 0;
    } else {
	return 1;
    }
}

static gint
find_user_activate_new_task_set(gconstpointer event_in,
				gconstpointer given_user_in)
{
    BA_ASSERT(given_user_in && event_in);
    ba_event_t *event;
    ba_user_t *given_user, *user;

    given_user = (ba_user_t *)given_user_in;
    event = (ba_event_t *)event_in;

    if(ba_event_kind_get(event) !=
       BA_EVENT_KIND_USER_ACTIVATE_NEW_TASK_SET) {
	return 1;
    }

    user = (ba_user_t *)ba_event_closure_get(event);

    if(given_user == user) {
	return 0;
    } else {
	return 1;
    }
}

static void
remove_tasks_from_list_helper(ba_user_t *user,
			      GSList **tasks)
{
    BA_ASSERT(user && tasks);
    ba_task_t *task;

    while(*tasks) {
	task = (ba_task_t *)(*tasks)->data;

	*tasks = g_slist_remove(*tasks, task);

	add_canceled_task(user, task);
    }
}

static void
add_canceled_task(ba_user_t *user,
		 ba_task_t *task)
{
    BA_ASSERT(user && task);
    ba_real virtual_clock_abs_time;

    virtual_clock_abs_time = ba_events_virtual_clock_abs_time_get();
    ba_task_canceled_abs_time_set(task, virtual_clock_abs_time);

    user->num_tasks_canceled++;

#if BA_SIM_INSPECT
    ba_print_info("user %d canceled task %d (remaining: " BA_REAL_LOW_PREC_FMT
		  ") at time "
		  BA_REAL_LOW_PREC_FMT "\n", ba_user_id_get(user),
		  ba_task_id_get(task), ba_task_remaining_delta_time_get(task),
		  ba_events_virtual_clock_abs_time_get());
#endif

    user->canceled_tasks = g_slist_append(user->canceled_tasks, task);
}

extern void
ba_user_add_resource_usage(ba_user_t *user,
			   ba_real resource_usage,
			   ba_real start_abs_time,
			   ba_real end_abs_time,
			   ba_task_t *task)
{
    BA_ASSERT(user);
    BA_ASSERT(task);
    // we don't use end_abs_time
    
    user->resource_total_usage += resource_usage;

    if(ba_task_requested_get(task)) {
	user->running_req_resource_total_usage += resource_usage;
    }

    if(ba_report_warmup_met()) {
	add_resource_usage_warmup_helper(user, resource_usage, start_abs_time,
					 task);
    }
}

static void
add_resource_usage_warmup_helper(ba_user_t *user,
				 ba_real resource_usage,
				 ba_real start_abs_time,
				 ba_task_t *task)
{
    ba_real warmup_abs_time;

    warmup_abs_time = ba_report_warmup_abs_time_get();

    // we've met the warmup time, so inc resource_usage
    user->warmup_resource_total_usage += resource_usage;
    
    if(ba_types_real_less_than_or_equal(start_abs_time, warmup_abs_time)) {
	// ahh, the resource run started before warm up.  so we
	// over counted the resource usage.  let's fix that.
	user->warmup_resource_total_usage -= (warmup_abs_time - start_abs_time);
	BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(user->warmup_resource_total_usage,
					     0.0);
    }

    if(ba_task_requested_get(task)) {
	add_resource_usage_warmup_requested_helper(user, resource_usage,
						   start_abs_time,
						   warmup_abs_time, task);
    }
}

static void
add_resource_usage_warmup_requested_helper(ba_user_t *user,
					   ba_real resource_usage,
					   ba_real start_abs_time,
					   ba_real warmup_abs_time,
					   ba_task_t *task)
{
    ba_real virtual_clock_abs_time;
    ba_real completed_before_warmup;
    ba_real st_after_warmup;

    virtual_clock_abs_time = ba_events_virtual_clock_abs_time_get();

    if(ba_types_real_less_than_or_equal(start_abs_time, warmup_abs_time)) {
	// case that the task has been requested and was scheduled
	// from before the warmup to after the warmup
	
	BA_ASSERT(ba_task_counted_in_warmup_get(task, 0, 0) == BA_FALSE);
	
	completed_before_warmup = ba_task_completed_delta_time_get(task) -
	    (virtual_clock_abs_time - warmup_abs_time);
	BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(completed_before_warmup, 0.0);
	
	st_after_warmup = ba_task_service_delta_time_get(task) -
	    completed_before_warmup;
	BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(st_after_warmup, 0.0);
	
	BA_ASSERT_REAL_EQUAL(st_after_warmup,
			     ba_task_remaining_delta_time_get(task) +
			     (virtual_clock_abs_time - warmup_abs_time));

	goto update;
    }

    // because of the goto, from here on start_abs_time > warmup_abs_time:

    if(ba_task_counted_in_warmup_get(task, 0, 0) == BA_FALSE) {
	// case of the first time that the task was requested and
	// scheduled from after the warmup to after the warmup

	completed_before_warmup = ba_task_completed_delta_time_get(task) -
	    resource_usage;
	BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(completed_before_warmup, 0.0);
	
	st_after_warmup = ba_task_remaining_delta_time_get(task) +
	    resource_usage;
	
	goto update;
    }

    return; // the task's requested resource usage after warmup has
            // already been counted

 update:
    user->warmup_req_resource_total_usage += st_after_warmup;
    ba_task_counted_in_warmup_set(task, completed_before_warmup,
				  st_after_warmup);
}

extern ba_real
ba_user_resource_total_usage_get(ba_user_t *user)
{
    BA_ASSERT(user);

    return user->resource_total_usage;
}

extern ba_real
ba_user_req_resource_total_usage_get(ba_user_t *user)
{
    BA_ASSERT(user);

    // not used 

    return user->req_resource_total_usage;
}

extern ba_real
ba_user_running_req_resource_total_usage_get(ba_user_t *user)
{
    BA_ASSERT(user);

    return user->running_req_resource_total_usage;
}

extern void
ba_user_add_executed_task(ba_user_t *user,
			  ba_task_t *task)
{
    BA_ASSERT(user && task);
    ba_user_t *user_comp;
    ba_real virtual_clock_abs_time;

    user_comp = ba_task_user_get(task);
    BA_ASSERT(user == user_comp);

#if THOROUGH_CHECKS
    // can't be on the following lists
    BA_ASSERT(g_slist_find(user->executed_tasks, task) == 0);
    BA_ASSERT(g_slist_find(user->finished_tasks, task) == 0);
    BA_ASSERT(g_slist_find(user->canceled_tasks, task) == 0);
#endif

    // if it was disclosed, then it will be no longer...
    if(g_slist_find(user->disclosed_tasks, task)) {
	user->disclosed_tasks = g_slist_remove(user->disclosed_tasks, task);
    }

    virtual_clock_abs_time = ba_events_virtual_clock_abs_time_get();
    ba_task_executed_abs_time_set(task, virtual_clock_abs_time);

    user->num_tasks_executed++;

#if BA_SIM_INSPECT
    ba_print_info("user %d's task %d executed at time "
		  BA_REAL_LOW_PREC_FMT "\n", ba_user_id_get(user),
		  ba_task_id_get(task), ba_events_virtual_clock_abs_time_get());
#endif

    // if it was requested, the following will do the right thing to
    // make the task finished, if not the following just makes it
    // executed.
    add_executed_or_requested_task_helper(user, &user->executed_tasks,
					  &user->requested_tasks, task);
}

static void
add_executed_or_requested_task_helper(ba_user_t *user,
				      GSList **add_tasks,
				      GSList **search_tasks,
				      ba_task_t *task)
{
    BA_ASSERT(user && add_tasks && search_tasks && task);
    GSList *elem;

    elem = g_slist_find(*search_tasks, task);
    if(elem) {
	*search_tasks = g_slist_remove(*search_tasks, task);

	add_finished_task(user, task);
	think_about_output(user, task);
    } else {
	*add_tasks = g_slist_append(*add_tasks, task);
    }
}

static void
add_finished_task(ba_user_t *user,
		  ba_task_t *task)
{
    BA_ASSERT(user && task);
    ba_real virtual_clock_abs_time;
    ba_real executed_abs_time, requested_abs_time, vrt_delta_time, vsd,
	service_delta_time;

#if THOROUGH_CHECKS
    // shouldn't be on any list now
    BA_ASSERT(g_slist_find(user->disclosed_tasks, task) == 0);
    BA_ASSERT(g_slist_find(user->executed_tasks, task) == 0);
    BA_ASSERT(g_slist_find(user->requested_tasks, task) == 0);
    BA_ASSERT(g_slist_find(user->finished_tasks, task) == 0);
    BA_ASSERT(g_slist_find(user->canceled_tasks, task) == 0);
#endif

    virtual_clock_abs_time = ba_events_virtual_clock_abs_time_get();
    ba_task_finished_abs_time_set(task, virtual_clock_abs_time);

    if(ba_report_warmup_met()) {
	user->num_tasks_finished++;
    }

    user->finished_tasks = ba_glib_fast_slist_append(user->finished_tasks_app,
						     user->finished_tasks,
						     task);

    executed_abs_time = ba_task_executed_abs_time_get(task);
    requested_abs_time = ba_task_requested_abs_time_get(task);
    vrt_delta_time = executed_abs_time - requested_abs_time;
    if(ba_types_real_less_than_or_equal(vrt_delta_time, 0.0)) {
	vrt_delta_time = 0.0;

	if(ba_report_warmup_met()) {
	    user->deadlines_met++;
	}
    }
    user->total_vrt_delta_time += vrt_delta_time;

    service_delta_time = ba_task_service_delta_time_get(task);
    vsd = vrt_delta_time / service_delta_time;

    if(ba_report_warmup_met()) {
	ba_report_user_vrt_delta_time(vrt_delta_time);
	ba_report_user_vsd(vsd);
    }

    ba_resource_add_req_resource_usage(service_delta_time);
}

static void
think_about_output(ba_user_t *user,
		   ba_task_t *task_to_think_about)
{
    BA_ASSERT(user);
    ba_real think_delta_time;
    
    if(task_to_think_about) {
	think_delta_time = ba_task_think_delta_time_get(task_to_think_about);
    } else {
	think_delta_time = BA_EVENTS_NOW;
    }

#if BA_SIM_INSPECT
    if(task_to_think_about) {
	ba_print_info("user %d started thinking about task %d for "
		      BA_REAL_LOW_PREC_FMT " at time "
		      BA_REAL_LOW_PREC_FMT "\n", ba_user_id_get(user),
		      ba_task_id_get(task_to_think_about),
		      think_delta_time,
		      ba_events_virtual_clock_abs_time_get());
    }
#endif

    ba_events_add_event(BA_EVENT_KIND_USER_REQUEST_TASK, think_delta_time,
			request_task_callback, user);
}

extern void
ba_user_queue_for_destruction(ba_user_t *user)
{
    BA_ASSERT(user);

    BA_ASSERT(user->queued_for_destruction == BA_FALSE);
    user->queued_for_destruction = BA_TRUE;

    if(ba_user_stop_running_tasks_get(user) == BA_FALSE) {
	stop_running_tasks_abs_time_set(user,
					ba_events_virtual_clock_abs_time_get());
    }

    ba_events_add_event(BA_EVENT_KIND_USERS_REMOVE_USER, BA_EVENTS_NOW,
			ba_users_remove_user, user);
}

static void
request_task_callback(ba_event_kind_t kind,
		      ba_real delay_delta_time,
		      void *closure)
{
    BA_ASSERT(kind == BA_EVENT_KIND_USER_REQUEST_TASK);
    BA_ASSERT(closure);
    ba_task_t *task;
    ba_user_t *user;
    ba_bool will_be_requested;

    user = (ba_user_t *)closure;
    BA_ASSERT(user);

    // if delay_delta_time is BA_EVENTS_NOW, that means that we
    // weren't thinking about anything.
    if(delay_delta_time != BA_EVENTS_NOW) {
	ba_stats_insert(user->think_delta_time_stats, delay_delta_time);
    }

    task = ba_task_set_return_next_unrequested_task(user->active_task_set);

    if(ba_options_virtual_clock_terminate_get()) {
	if(task == 0) {
	    will_be_requested = BA_FALSE;
	    // greater modeling complexity: could set a var for when
	    // the activate_new_task_set
	    // event should occur.  instead of NOW, why not
	    // delay_delta_time, so in effect the think time is
	    // doubled when designing a new experiment.
	} else {
	    will_be_requested = ba_task_will_be_requested_get(task);
	}
    } else {
	if(task == 0 || ba_user_can_req_or_dis(user, BA_TRUE) == BA_FALSE) {
	    ba_user_queue_for_destruction(user);
	    return;
	}

	will_be_requested = ba_task_will_be_requested_get(task);
    }

    if(!will_be_requested) {
	ba_events_add_event(BA_EVENT_KIND_USER_ACTIVATE_NEW_TASK_SET,
			    BA_EVENTS_NOW, ba_user_activate_new_task_set, user);
	return;
    }

    add_requested_task(user, task);
    if(g_slist_find(user->requested_tasks, task)) {
	// this `if' fails when the task was executed; when
	// add_requested_task just puts it on the finished list.
	ba_resource_preempt();
    }
}

static ba_real
gen_task_set_change(void *generic_data)
{
    BA_ASSERT(generic_data);
    ba_user_t *user;
    ba_real rv;

    user = (ba_user_t *)generic_data;

    rv = ba_rv_dist(user->task_set_change_dist);

    return rv;
}

extern ba_bool
ba_user_can_req_or_dis(ba_user_t *user,
			    ba_bool will_request)
{
    if(ba_options_virtual_clock_terminate_get()) {
	return BA_TRUE;
    }

    if(user->max_req_dis_tasks == 0) {
	return BA_TRUE;
    }

    BA_ASSERT(user->num_tasks_requested + user->num_tasks_disclosed_only <=
	      user->max_req_dis_tasks);

    if(will_request && user->num_tasks_requested == 0) {
	// XXX always allow the user to request at least one task,
	// even if we temporarily violate the constraint.  it'll get
	// quickly fixed since requested decrements
	// num_tasks_disclosed_only.
	return BA_TRUE;
    }

    if(user->num_tasks_requested + user->num_tasks_disclosed_only ==
       user->max_req_dis_tasks) {
	if(will_request) {
	    ba_report_user_inc_num_hit_req_dis_limit();
	}
	return BA_FALSE;
    }

    return BA_TRUE;
}

extern void
ba_user_add_disclosed_task(ba_user_t *user,
			   ba_task_t *task)
{
    BA_ASSERT(user && task);
    ba_real virtual_clock_abs_time;

#if THOROUGH_CHECKS
    // shouldn't be on any list now
    BA_ASSERT(g_slist_find(user->disclosed_tasks, task) == 0);
    BA_ASSERT(g_slist_find(user->executed_tasks, task) == 0);
    BA_ASSERT(g_slist_find(user->requested_tasks, task) == 0);
    BA_ASSERT(g_slist_find(user->finished_tasks, task) == 0);
    BA_ASSERT(g_slist_find(user->canceled_tasks, task) == 0);
#endif

    virtual_clock_abs_time = ba_events_virtual_clock_abs_time_get();
    ba_task_disclosed_abs_time_set(task, virtual_clock_abs_time);

    user->num_tasks_disclosed++;
    user->num_tasks_disclosed_only++;
    user->disclosed_tasks = g_slist_append(user->disclosed_tasks, task);

#if BA_SIM_INSPECT
    ba_print_info("user %d disclosed task %d (st: " BA_REAL_LOW_PREC_FMT
		  ") at time "
		  BA_REAL_LOW_PREC_FMT "\n", ba_user_id_get(user),
		  ba_task_id_get(task), ba_task_service_delta_time_get(task),
		  ba_events_virtual_clock_abs_time_get());
#endif
}

static void
add_requested_task(ba_user_t *user,
		   ba_task_t *task)
{
    BA_ASSERT(user && task);
    ba_real virtual_clock_abs_time;
    ba_real service_delta_time;

#if THOROUGH_CHECKS
    // can't be on the following lists
    BA_ASSERT(g_slist_find(user->requested_tasks, task) == 0);
    BA_ASSERT(g_slist_find(user->finished_tasks, task) == 0);
    BA_ASSERT(g_slist_find(user->canceled_tasks, task) == 0);
#endif

    // if it was disclosed, then it will be no longer...
    if(g_slist_find(user->disclosed_tasks, task)) {
	user->disclosed_tasks = g_slist_remove(user->disclosed_tasks, task);
    }

    virtual_clock_abs_time = ba_events_virtual_clock_abs_time_get();
    ba_task_requested_abs_time_set(task, virtual_clock_abs_time);

    service_delta_time = ba_task_service_delta_time_get(task);
    user->req_resource_total_usage += service_delta_time;

    user->running_req_resource_total_usage +=
	ba_task_completed_delta_time_get(task);

    user->num_tasks_requested++;

    // modeling complexity, but also the simulator's implementation of
    // the model XXX: under user behaviors in which i might request a
    // task was never disclosed, this minusminus should not happen!
    user->num_tasks_disclosed_only--;

#if BA_SIM_INSPECT
    ba_print_info("user %d needed (requested) task %d (remain: "
		  BA_REAL_LOW_PREC_FMT ") at time "
		  BA_REAL_LOW_PREC_FMT "\n", ba_user_id_get(user),
		  ba_task_id_get(task), ba_task_remaining_delta_time_get(task),
		  ba_events_virtual_clock_abs_time_get());
#endif

    // if it was executed, the following will do the right thing to
    // make the task finished, if not the following just makes it
    // requested.
    add_executed_or_requested_task_helper(user, &user->requested_tasks,
					  &user->executed_tasks, task);
}

extern ba_user_t *
ba_user_new(ba_id_t id)
{
    ba_user_t *user;
    ba_real tasks_per_task_set_real;

    user = ba_mem_malloc(sizeof(ba_user_t));

    user->id = id;

    user->queued_for_destruction = BA_FALSE;
    user->stop_running_tasks = BA_FALSE;
    user->stop_running_tasks_abs_time = 0;

    user->disclosed_tasks = 0;
    user->executed_tasks = 0;
    user->requested_tasks = 0;
    user->finished_tasks_app = ba_glib_fast_slist_append_new();
    user->finished_tasks = 0;
    user->canceled_tasks = 0;

    user->task_set_id_gen = ba_id_gen_new();
    user->task_sets_app = ba_glib_fast_slist_append_new();
    user->task_sets = 0;
    user->active_task_set = 0;

    user->max_req_dis_tasks =
	ba_behavior_max_req_dis_tasks_per_user_get();

    user->task_id_gen = ba_id_gen_new();

    user->task_set_change_range_dist =
	ba_rv_dist_copy(ba_options_task_set_change_range_dist_get());

    user->task_set_change_prob =
	ba_trace_get_and_log_rv(BA_TRACE_ENTRY_TASK_SET_CHANGE_PROB,
				user->id, -1, gen_task_set_change_prob, user);
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(user->task_set_change_prob,
					 0.0);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(user->task_set_change_prob, 1.0);
    user->task_set_change_dist =
	ba_rv_dist_new("bernoulli", 0.0, 0.0, 0.0, user->task_set_change_prob);

    user->tasks_per_task_set_dist =
	ba_rv_dist_copy(ba_options_tasks_per_task_set_dist_get());
    tasks_per_task_set_real =
	ba_trace_get_and_log_rv(BA_TRACE_ENTRY_TASKS_PER_TASK_SET, user->id, -1,
				gen_tasks_per_task_set, user);
    user->tasks_per_task_set = rint(tasks_per_task_set_real);

    user->service_delta_time_dist =
	ba_rv_dist_copy(ba_options_service_delta_time_dist_get());

    user->think_delta_time_dist =
	ba_rv_dist_copy(ba_options_think_delta_time_dist_get());

    user->resource_total_usage = 0.0;
    user->warmup_resource_total_usage = 0.0;
    user->req_resource_total_usage = 0.0;
    user->running_req_resource_total_usage = 0.0;
    user->warmup_req_resource_total_usage = 0.0;
    user->correct_warmup_req_resource_total_usage = 0.0;
    user->warmup_dis_resource_total_usage = 0.0;
    user->warmup_scaled_resource_total_usage = 0.0;

    user->num_task_sets = 0;
    user->num_tasks_disclosed = 0;
    user->num_tasks_disclosed_only = 0;
    user->num_tasks_executed = 0;
    user->num_tasks_requested = 0;
    user->num_tasks_finished = 0;
    user->num_tasks_canceled = 0;
    user->service_delta_time_stats = ba_stats_new();
    user->think_delta_time_stats = ba_stats_new();

    user->total_vrt_delta_time = 0.0;
    user->total_vsd = 0.0;
    user->deadlines_met = 0;

    return user;
}

static ba_real
gen_task_set_change_prob(void *generic_data)
{
    BA_ASSERT(generic_data);
    ba_user_t *user;
    ba_real rv;

    user = (ba_user_t *)generic_data;

    rv = ba_rv_dist(user->task_set_change_range_dist);

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(rv, 0.0);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(rv, 1.0);

    return rv;
}

static ba_real
gen_tasks_per_task_set(void *generic_data)
{
    BA_ASSERT(generic_data);
    ba_user_t *user;
    ba_real rv;
    uint32_t rv_int;

    user = (ba_user_t *)generic_data;

 retry:
    rv = ba_rv_dist(user->tasks_per_task_set_dist);
    rv_int = rint(rv);
    if(rv_int == 0) {
	goto retry;
    }
    rv = rv_int;

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(rv, 1.0);

    return rv;
}

static void
report_info(ba_user_t *user)
{
    BA_ASSERT(user);

    // calculate somethings first...

    // resource stuff
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(user->warmup_resource_total_usage,
			    user->correct_warmup_req_resource_total_usage);
    user->warmup_dis_resource_total_usage = user->warmup_resource_total_usage -
	user->correct_warmup_req_resource_total_usage;
    if(ba_types_real_equal(user->correct_warmup_req_resource_total_usage,
			   0.0)) {
	// really should be infinite.  this number is only used for
	// batch to see how bad batch is, and i show it's plenty bad, even if i
	// underestimate in this rare case.
	user->warmup_scaled_resource_total_usage = 1.0;
    } else {
	user->warmup_scaled_resource_total_usage =
	    user->warmup_resource_total_usage /
	    user->correct_warmup_req_resource_total_usage;
    }
    /*
    user->warmup_dis_resource_total_usage = user->warmup_resource_total_usage -
	user->warmup_req_resource_total_usage;
    if(ba_types_real_equal(user->warmup_req_resource_total_usage, 0.0)) {
	user->warmup_scaled_resource_total_usage = 1.0;
    } else {
	user->warmup_scaled_resource_total_usage =
	    user->warmup_resource_total_usage /
	    user->warmup_req_resource_total_usage;
    }
    */

    // tasks and task sets stuff
    // the following two are unused
    user->mean_service_delta_time =
	ba_stats_sample_mean_get(user->service_delta_time_stats);
    if(ba_stats_non_empty(user->think_delta_time_stats)) {
	user->mean_think_delta_time =
	    ba_stats_sample_mean_get(user->think_delta_time_stats);
    }

    // visible response time stuff
    if(user->req_resource_total_usage == 0.0) {
	user->total_vsd = 0.0;
    } else {
	user->total_vsd = user->total_vrt_delta_time /
	    user->req_resource_total_usage;
    }

    // now report...

    // resource stuff
    ba_report_user_resource_usage(user->warmup_resource_total_usage);
    ba_report_user_req_resource_usage(user->
				      correct_warmup_req_resource_total_usage);
    /*
    ba_report_user_req_resource_usage(user->warmup_req_resource_total_usage);
    */
    ba_report_user_dis_resource_usage(user->warmup_dis_resource_total_usage);
    ba_report_user_scaled_resource_usage(user->
					 warmup_scaled_resource_total_usage);

    // tasks and task sets stuff
    ba_report_user_task_sets(user->num_task_sets);
    ba_report_user_disclosed_tasks(user->num_tasks_disclosed);
    ba_report_user_disclosed_only_tasks(user->num_tasks_disclosed_only);
    ba_report_user_executed_tasks(user->num_tasks_executed);
    ba_report_user_requested_tasks(user->num_tasks_requested);
    ba_report_user_canceled_tasks(user->num_tasks_canceled);
    ba_report_user_finished_tasks(user->num_tasks_finished);
    ba_report_user_service_delta_time(user->mean_service_delta_time);
    if(ba_stats_non_empty(user->think_delta_time_stats)) {
	ba_report_user_think_delta_time(user->mean_think_delta_time);
    }
    
    // visible response time stuff
    ba_report_user_user_vrt_delta_time(user->total_vrt_delta_time);
    ba_report_user_user_vsd(user->total_vsd);
    ba_report_user_deadlines_met(user->deadlines_met);
}

static void
assert_info(ba_user_t *user)
{
    BA_ASSERT(user);

    // resources

    // fix up if i ever do `no virtual clock terminate' stuff XXX

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(user->resource_total_usage, 0.0);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(user->resource_total_usage,
				      ba_options_virtual_clock_terminate_get());

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(user->warmup_resource_total_usage,
					 0.0);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(user->warmup_resource_total_usage,
				      ba_options_virtual_clock_terminate_get() -
				      ba_report_warmup_abs_time_get());
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(user->warmup_resource_total_usage,
				      user->resource_total_usage);

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(user->req_resource_total_usage, 0.0);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(user->req_resource_total_usage,
				      user->resource_total_usage);
    if(ba_options_dis_sched_behavior_get() == BA_SCHED_BEHAVIOR_NOOP) {
	BA_ASSERT_REAL_EQUAL(user->resource_total_usage,
			     user->req_resource_total_usage);
    }

    BA_ASSERT_REAL_EQUAL(user->running_req_resource_total_usage,
			 user->req_resource_total_usage);

    //
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(user->warmup_req_resource_total_usage,
					 0.0);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(user->warmup_req_resource_total_usage,
				      user->warmup_resource_total_usage);
    if(ba_options_dis_sched_behavior_get() == BA_SCHED_BEHAVIOR_NOOP) {
	BA_ASSERT_REAL_EQUAL(user->warmup_resource_total_usage,
			     user->warmup_req_resource_total_usage);
    }
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(user->
				 correct_warmup_req_resource_total_usage,
					 0.0);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(user->
				      correct_warmup_req_resource_total_usage,
				      user->warmup_resource_total_usage);
    if(ba_options_dis_sched_behavior_get() == BA_SCHED_BEHAVIOR_NOOP) {
	BA_ASSERT_REAL_EQUAL(user->warmup_resource_total_usage,
			     user->correct_warmup_req_resource_total_usage);
    }
    //
    
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(user->warmup_dis_resource_total_usage,
					 0.0);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(user->warmup_dis_resource_total_usage,
				      user->resource_total_usage);
    
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(user->
					 warmup_scaled_resource_total_usage,
					 1.0);

    // tasks and task sets
    BA_ASSERT(user->num_task_sets >= 1);

    // modeling complexity: these num_tasks_* asserts will probably
    // change on more complex user behavior.

    BA_ASSERT(user->num_tasks_disclosed >= 1);
    if(user->max_req_dis_tasks) {
	BA_ASSERT(user->num_tasks_disclosed <= user->max_req_dis_tasks);
    }

    BA_ASSERT(user->num_tasks_disclosed_only >= 0);
    if(user->max_req_dis_tasks) {
	BA_ASSERT(user->num_tasks_disclosed_only <= user->max_req_dis_tasks);
    }

    if(!ba_options_virtual_clock_terminate_get()) {
	BA_ASSERT(user->num_tasks_executed >= 1);
    }
    BA_ASSERT(user->num_tasks_executed <=
	      user->num_tasks_disclosed_only + user->num_tasks_requested);

    BA_ASSERT(user->num_tasks_requested >= 1);
    if(user->max_req_dis_tasks) {
	BA_ASSERT(user->num_tasks_requested <= user->max_req_dis_tasks);
	BA_ASSERT(user->num_tasks_requested + user->num_tasks_disclosed_only <=
		  user->max_req_dis_tasks);
    }

    if(!ba_options_virtual_clock_terminate_get()) {
       BA_ASSERT(user->num_tasks_finished >= 1);
    }
    BA_ASSERT(user->num_tasks_finished <= user->num_tasks_executed);
    BA_ASSERT(user->num_tasks_finished <= user->num_tasks_requested);
    if(user->max_req_dis_tasks) {
	BA_ASSERT(user->num_tasks_finished <= user->max_req_dis_tasks);
	if(user->tasks_per_task_set > user->max_req_dis_tasks) {
	    BA_ASSERT(user->num_tasks_requested +
		      user->num_tasks_disclosed_only ==
		      user->max_req_dis_tasks);
	}
    }
    if(ba_report_warmup_abs_time_get() == 0.0) {
	BA_ASSERT(user->num_tasks_finished >= user->deadlines_met);
    }

    BA_ASSERT(user->num_tasks_canceled >= 0);

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_min_get(user->service_delta_time_stats), 0.0);
    BA_ASSERT(ba_stats_count_get(user->service_delta_time_stats) >=
	      user->num_tasks_finished);
    if(ba_stats_non_empty(user->think_delta_time_stats)) {
	BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
          ba_stats_min_get(user->think_delta_time_stats), 0.0);
	if(!ba_options_virtual_clock_terminate_get()) {
	    BA_ASSERT(ba_stats_count_get(user->think_delta_time_stats) ==
		      user->num_tasks_finished);
	}
    }

    // visible response time
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(user->total_vrt_delta_time, 0.0);

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(user->total_vsd, 0.0);
    
    BA_ASSERT(user->deadlines_met >= 0);
    if(ba_report_warmup_abs_time_get() == 0.0) {
	BA_ASSERT(user->deadlines_met <= user->num_tasks_finished);
	if(user->num_tasks_finished == 1) {
	    BA_ASSERT(user->deadlines_met == 0);
	}
    }
}

extern void
ba_user_delete(ba_user_t *user)
{
    GSList *node;

    BA_ASSERT(user);

    BA_ASSERT(user->queued_for_destruction);
    BA_ASSERT(user->stop_running_tasks);

    BA_ASSERT(user->active_task_set == 0);

    // new code begin
    for(node = user->finished_tasks; node; node = g_slist_next(node)) {
	ba_task_t *task;
	
	task = (ba_task_t *)node->data;
	BA_ASSERT(task);
	user->correct_warmup_req_resource_total_usage += 
	    ba_task_completed_warmup_delta_time_get(task);	
    }
    // new code end

    report_info(user);
    assert_info(user);

    ba_id_gen_delete(user->task_set_id_gen);

    g_slist_foreach(user->task_sets, ba_task_set_delete_from_list, 0);
    g_slist_free(user->task_sets);
    ba_glib_fast_slist_append_delete(user->task_sets_app);

    BA_ASSERT(user->disclosed_tasks == 0);
    BA_ASSERT(user->executed_tasks == 0);
    BA_ASSERT(user->requested_tasks == 0);

    g_slist_free(user->disclosed_tasks); // moot, since 0
    g_slist_free(user->executed_tasks); // moot, since 0
    g_slist_free(user->requested_tasks); // moot, since 0

    // unlike executed and requested tasks, the following will likely
    // have tasks on them.
    g_slist_free(user->finished_tasks);
    ba_glib_fast_slist_append_delete(user->finished_tasks_app);
    g_slist_free(user->canceled_tasks);

    ba_rv_dist_delete(user->task_set_change_range_dist);
    ba_rv_dist_delete(user->task_set_change_dist);
    ba_rv_dist_delete(user->tasks_per_task_set_dist);
    ba_rv_dist_delete(user->service_delta_time_dist);
    ba_rv_dist_delete(user->think_delta_time_dist);

    ba_stats_delete(user->service_delta_time_stats);
    ba_stats_delete(user->think_delta_time_stats);

    ba_id_gen_delete(user->task_id_gen);

    ba_mem_free(user);
}

extern void
ba_user_delete_from_list(void *user_in,
			 void *unused)
{
    BA_ASSERT(user_in);
    ba_user_t *user;

    user = (ba_user_t *)user_in;

    ba_user_delete(user);
}

extern ba_status
ba_user_init(void)
{
    return BA_STATUS_SUCCESS;
}

extern ba_status
ba_user_destroy(void)
{

    return BA_STATUS_SUCCESS;
}

extern void
ba_user_destroy_from_hook(void *unused)
{
    ba_status status;

    status = ba_user_destroy();
    BA_ASSERT(status == BA_STATUS_SUCCESS);
}
